Package: treestats
Type: Package
Title: Phylogenetic Tree Statistics
Version: 1.71.12
Authors@R: c(person(given = "Thijs",
        family = "Janzen",
        email = "thijsjanzen@gmail.com",
        role = c("cre", "aut"),
        comment = c(ORCID = "0000-0002-4162-1140")))
Maintainer: Thijs Janzen <thijsjanzen@gmail.com>
Description: Collection of phylogenetic tree statistics,
             collected throughout the literature. All functions have been 
             written to maximize computation speed. The package includes 
             umbrella functions to calculate all statistics, all balance 
             associated statistics, or all branching time related statistics. 
             Furthermore, the 'treestats' package supports summary statistic 
             calculations on Ltables, provides speed-improved coding of 
             branching times, Ltable conversion and includes algorithms to
             create intermediately balanced trees. Full description can be
             found in Janzen (2024) <doi:10.1016/j.ympev.2024.108168>.
License: GPL-3
Encoding: UTF-8
Depends: Rcpp
LinkingTo: Rcpp, RcppArmadillo, nloptr
Imports: ape, nloptr, treebalance, DDD, RSpectra
Suggests: phytools, phyloTop, testthat, geiger, nLTT, castor, adephylo,
        ggplot2, tibble, dplyr, tidyr, picante, lintr, rmarkdown,
        knitr, igraph, Matrix, pheatmap, ggdendro, dendextend, Matrix,
        TreeSim, nlme
BugReports: https://github.com/thijsjanzen/treestats/issues
URL: https://thijsjanzen.github.io/treestats/
VignetteBuilder: knitr
SystemRequirements: C++17
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-06-16 09:55:09 UTC; thijsjanzen
Author: Thijs Janzen [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-4162-1140>)
Repository: CRAN
Date/Publication: 2026-06-16 11:50:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-16 14:08:16 UTC; unix
Archs: treestats.so.dSYM

# Generated by roxygen2: do not edit by hand

export(ILnumber)
export(area_per_pair)
export(average_leaf_depth)
export(avg_ladder)
export(avg_vert_depth)
export(b1)
export(b2)
export(beta_statistic)
export(blum)
export(branching_times)
export(calc_all_stats)
export(calc_brts_stats)
export(calc_topology_stats)
export(check_binary)
export(cherries)
export(colless)
export(colless_corr)
export(colless_quad)
export(create_fully_balanced_tree)
export(create_fully_unbalanced_tree)
export(crown_age)
export(diameter)
export(double_cherries)
export(eigen_centrality)
export(entropy_j)
export(ew_colless)
export(four_prong)
export(gamma_statistic)
export(i_stat)
export(imbalance_steps)
export(inv_branch_dist)
export(j_one)
export(l_to_phylo)
export(laplacian_spectrum)
export(list_statistics)
export(ltable_to_newick)
export(make_unbalanced_tree)
export(max_betweenness)
export(max_closeness)
export(max_del_width)
export(max_depth)
export(max_ladder)
export(max_width)
export(mean_branch_length)
export(mean_branch_length_ext)
export(mean_branch_length_int)
export(mean_i)
export(mean_pair_dist)
export(minmax_adj)
export(minmax_laplace)
export(mntd)
export(mw_over_md)
export(nLTT)
export(nLTT_base)
export(number_of_lineages)
export(phylo_to_l)
export(phylogenetic_diversity)
export(pigot_rho)
export(pitchforks)
export(psv)
export(rebase_ltable)
export(rogers)
export(root_imbalance)
export(rquartet)
export(sackin)
export(sshape)
export(stairs)
export(stairs2)
export(sym_nodes)
export(tot_coph)
export(tot_internal_path)
export(tot_path_length)
export(tree_height)
export(treeness)
export(var_branch_length)
export(var_branch_length_ext)
export(var_branch_length_int)
export(var_leaf_depth)
export(var_pair_dist)
export(wiener)
import(Rcpp)
import(nloptr)
useDynLib(treestats)

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/treestats)](https://cran.r-project.org/package=treestats)
[![](https://cranlogs.r-pkg.org/badges/grand-total/treestats)](https://cran.r-project.org/package=treestats)
[![](https://cranlogs.r-pkg.org/badges/treestats)](https://cran.r-project.org/package=treestats)
[![R-CMD-check](https://github.com/thijsjanzen/treestats/workflows/R-CMD-check/badge.svg)](https://github.com/thijsjanzen/treestats/actions)
[![codecov.io](https://codecov.io/gh/thijsjanzen/treestats/branch/main/graph/badge.svg)](https://app.codecov.io/gh/thijsjanzen/treestats) 


<img src="https://github.com/thijsjanzen/treestats/blob/main/layout/hex_treestats.png?raw=true" align="right" width="180" />

# Treestats 
#### Measuring properties of phylogenetic trees

The **treestats** R package contains rapid, C++ based, functions to
calculate summary statistics on phylogenies. For some functions (but not all, see below), the
phylogenies are required to be ultrametric and/or binary.

## Getting started

### Installation

To get started, you can either install from CRAN or use the latest
version from GitHub:

```         
install.packages("treestats") # install from CRAN

# use the devtools package to install latest version from GitHub:
install.packages("devtools")
devtools::install_github("thijsjanzen/treestats")
```

### Basic usage

Given a tree (for example a simulated tree, as in the code example), you
can either access individual statistics, or calculate all currently
implemented statistics:

```         
focal_tree   <- ape::rphylo(n = 10, birth = 1, death = 0)
colless_stat <- treestats::colless(focal_tree)
all_stats    <- treestats::calc_all_stats(focal_tree)
```

## Rcpp

For all of the implemented statistics, the package provides Rcpp versions that are
much, much faster than their R sister functions. Furthermore, some
additional functions have been improved as well: 
* ape::branching.times
* DDD::phylo2L
* DDD::L2phylo

![](https://github.com/thijsjanzen/treestats/blob/cd3649833740eb7cdb23f722a2738cfd23bc4b10/layout/Figure_S3.png?raw=true)

### List of statistics

The following summary statistics are included:

<table>
    <tr>
        <td>Statistic</td>
        <td>Information</td>
        <td>Norma lization</td>
        <td>Req Binary?</td>
        <td>Req Ultrametric?</td>
        <td>Req Rooted?</td>
        <td>Sensitive to root position</td>
        <td>Reference</td>
    </tr>
    <tr>
        <td>area_per_pair</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Lima et al., 2020</td>
    </tr>
    <tr>
        <td>average_leaf_depth</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Shao &amp; Sokal, 1990</td>
    </tr>
    <tr>
        <td>avg_ladder</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Kendall et al., 2018</td>
    </tr>
    <tr>
        <td>avg_vert_depth</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colijn &amp; Gardy, 2014</td>
    </tr>
    <tr>
        <td>b1</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Shao &amp; Sokal, 1990</td>
    </tr>
    <tr>
        <td>b2</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Shao &amp; Sokal, 1990</td>
    </tr>
    <tr>
        <td>beta</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Aldous, 1996</td>
    </tr>
    <tr>
        <td>blum / sshape</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Blum &amp; François, 2006</td>
    </tr>
    <tr>
        <td>cherries</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>McKenzie et al., 1999</td>
    </tr>
    <tr>
        <td>colless</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colless, 1982</td>
    </tr>
    <tr>
        <td>colless_corr</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Heard, 1992</td>
    </tr>
    <tr>
        <td>colless_quad</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Bartoszek et al., 2021</td>
    </tr>
    <tr>
        <td>crown_age</td>
        <td>Br times</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td></td>
    </tr>
    <tr>
        <td>diameter</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>double_cherries</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>eigen_centrality</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>eigen_centralityW</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>ew_colless</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Mooers &amp; S. B. Heard, 1997</td>
    </tr>
    <tr>
        <td>four_prong</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>gamma</td>
        <td>Br times</td>
        <td>None</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Pybus &amp; Harvey, 2000</td>
    </tr>
    <tr>
        <td>i_stat</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Fusco &amp; Cronk, 1995</td>
    </tr>
    <tr>
        <td>il_number</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Kendall et al., 2018</td>
    </tr>
    <tr>
        <td>imbalance_steps</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Janzen &amp; Etienne, 2024</td>
    </tr>
    <tr>
        <td>inv_branch_dist</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Williams et al. 2026</td>
    </tr>
    <tr>
        <td>j_one</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Lemant et al., 2022</td>
    </tr>
    <tr>
        <td>j_stat</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td>Izsák &amp; Papp, 2000</td>
    </tr>
    <tr>
        <td>laplace_spectrum_a</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Lewitus &amp; Morlon, 2016</td>
    </tr>
    <tr>
        <td>laplace_spectrum_e</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Lewitus &amp; Morlon, 2016</td>
    </tr>
    <tr>
        <td>laplace_spectrum_g</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Lewitus &amp; Morlon, 2016</td>
    </tr>
    <tr>
        <td>laplace_spectrum_p</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Lewitus &amp; Morlon, 2016</td>
    </tr>
    <tr>
        <td>max_adj</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>max_betweenness</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>max_closeness</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>Small</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>max_closenessW</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>Small</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>max_del_width</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colijn &amp; Gardy, 2014</td>
    </tr>
    <tr>
        <td>max_depth</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colijn &amp; Gardy, 2014</td>
    </tr>
    <tr>
        <td>max_ladder</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Kendall et al., 2018</td>
    </tr>
    <tr>
        <td>max_laplace</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>max_width</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colijn &amp; Gardy, 2014</td>
    </tr>
    <tr>
        <td>mean_branch_length</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Janzen &amp; Etienne, 2017</td>
    </tr>
    <tr>
        <td>mean_branch_length_ext</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td>Saulnier et al., 2017</td>
    </tr>
    <tr>
        <td>mean_branch_length_int</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Saulnier et al., 2017</td>
    </tr>
    <tr>
        <td>min_adj</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>min_laplace</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
    <tr>
        <td>mntd</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td>Webb et al., 2002</td>
    </tr>
    <tr>
        <td>mpd</td>
        <td>Topo + br len</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td>Webb et al., 2002</td>
    </tr>
    <tr>
        <td>mw_over_md</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colijn &amp; Gardy, 2014</td>
    </tr>
    <tr>
        <td>nltt_base</td>
        <td>Br times</td>
        <td>None</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Janzen et al., 2015</td>
    </tr>
    <tr>
        <td>number_of_lineages</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td></td>
    </tr>
    <tr>
        <td>phylogenetic_div</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Faith, 1992</td>
    </tr>
    <tr>
        <td>pigot_rho</td>
        <td>Br times</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Pigot et al., 2010</td>
    </tr>
    <tr>
        <td>pitchforks</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Kendall et al., 2018</td>
    </tr>
    <tr>
        <td>psv</td>
        <td>Topo + br len</td>
        <td>Tips</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td>Helmus et al., 2007</td>
    </tr>
    <tr>
        <td>rogers</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Rogers, 1996</td>
    </tr>
    <tr>
        <td>root_imbalance</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Guyer et al., 1993</td>
    </tr>
    <tr>
        <td>rquartet</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Coronado et al., 2019</td>
    </tr>
    <tr>
        <td>sackin</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Sackin, 1972</td>
    </tr>
    <tr>
        <td>stairs</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Norström et al., 2012</td>
    </tr>
    <tr>
        <td>stairs2</td>
        <td>Topology</td>
        <td>None</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Norström et al., 2012</td>
    </tr>
    <tr>
        <td>symmetry_nodes</td>
        <td>Topology</td>
        <td>Tips</td>
        <td>TRUE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Kersting &amp; Fischer, 2021</td>
    </tr>
    <tr>
        <td>tot_coph</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Mir et al., 2013</td>
    </tr>
    <tr>
        <td>tot_internal_path</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Knuth, 1997</td>
    </tr>
    <tr>
        <td>tot_path</td>
        <td>Topology</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Colijn &amp; Gardy, 2014</td>
    </tr>
    <tr>
        <td>tree_height</td>
        <td>Br times</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td></td>
    </tr>
    <tr>
        <td>treeness</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Astolfi &amp; Zonta-Sgaramella, 1984</td>
    </tr>
    <tr>
        <td>var_branch_length</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Saulnier et al., 2017</td>
    </tr>
    <tr>
        <td>var_branch_length_ext</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Saulnier et al., 2017</td>
    </tr>
    <tr>
        <td>var_branch_length_int</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Saulnier et al., 2017</td>
    </tr>
    <tr>
        <td>var_depth</td>
        <td>Topology</td>
        <td>Yule</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>TRUE</td>
        <td>YES</td>
        <td>Coronado et al., 2020</td>
    </tr>
    <tr>
        <td>vpd</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>NO</td>
        <td>Webb et al., 2002</td>
    </tr>
    <tr>
        <td>wiener</td>
        <td>Topo + br len</td>
        <td>None</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>FALSE</td>
        <td>YES</td>
        <td>Chindelevitch et al., 2021</td>
    </tr>
</table>

## C++ Library

For the Rcpp improved summary statistics (excluding statistics that rely
on the calculation of eigen values, as these rely on the Rcpp
independent Eigen code), R independent C++ code is provided in the
inst/include folder. These can be independently linked by adding the
treestats package in the DESCRIPTION in both the LinkingTo and Depends
fields. Then, in your package, you can also calculate these functions.

Please note that for all functions, there are two versions available: 1)
based on input of a phylo object, which is typically one 2-column matrix
containing all edges, and a vector containing the edge lengths
(depending on which information is required to calculate the statistic).
2) based on input of an Ltable (Lineage table), which is a 4-column
matrix containing information on each species, being 1) birth time, 2)
parent species, 3) species label and 4) death time (or -1 if extant).

Ltable input can be useful when summary statistics are required for more
complicated simulation models.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = FALSE)
knitr::opts_chunk$set(fig.width = 6)
knitr::opts_chunk$set(fig.height = 6)
speed_data <- read.table("https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/Figure_S3_S4/timings.txt", header = TRUE)  # nolint


## ----out.width="100%", echo=FALSE---------------------------------------------
knitr::include_graphics("https://github.com/thijsjanzen/treestats/blob/main/layout/Figure_S3.png?raw=true") # nolint

## ----summarise_data, echo=FALSE-----------------------------------------------
res <- c()
for (x in unique(speed_data$treestatsfunction)) {
  a <- subset(speed_data, speed_data$treestatsfunction == x &
                          speed_data$ntips == 1000 &
                          speed_data$method == "treestats")
  res <- rbind(res, c(x, mean(a$time)))
}
res <- data.frame("treestats_function" = res[, 1],
                     "time" = res[, 2])
res$time <- as.numeric(res$time)

res2 <- res[order(res$time), ]


## ----out.width="100%", echo = FALSE-------------------------------------------
opar <- par(no.readonly = TRUE)
par(mar = c(8, 4, 4, 4))
barplot(res2$time, names.arg = res2$treestats_function, las = 2,
        cex.names = 0.4, ylab = "Time", log = "y")
par(opar)

## ----out.width="100%", echo=FALSE---------------------------------------------
knitr::include_graphics("https://github.com/thijsjanzen/treestats/blob/main/layout/Figure_S4.png?raw=true") # nolint

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
require(TreeSim)
require(ggplot2)
require(tibble)
require(tidyr)
require(dplyr)

## ----simulate_trees-----------------------------------------------------------
found <- c()
found_min <- c()
found_max <- c()
for (tree_size in ceiling(10^seq(1, 3, length.out = 10))) {
  cat(tree_size, "\n")
  stats <- c()

  # more replicates for smaller trees:
  num_repl <- min(100, ceiling(5 + 1000 / tree_size))
  speciation <- 0.5

  # calculate the expected crown age for the relevant tree size:
  max_t <- (1 / speciation) * log(tree_size / 2)

  for (r in 1:num_repl) {
    tree <- TreeSim::sim.bd.taxa.age(n = tree_size,
                                     numbsim = 1,
                                     lambda = speciation,
                                     mu = 0,
                                     age = max_t,
                                     mrca = TRUE)[[1]]
    stats <- rbind(stats, treestats::calc_all_stats(tree, normalize = TRUE))
  }
  stats2 <- apply(stats, 2, mean)

  min_stats <- apply(stats, 2, function(x) return(quantile(x, 0.025)))
  max_stats <- apply(stats, 2, function(x) return(quantile(x, 0.975)))

  found <- rbind(found, stats2)
  found_min <- rbind(found_min, min_stats)
  found_max <- rbind(found_max, max_stats)
}

found <- as_tibble(found)
found_min <- as_tibble(found_min)
found_max <- as_tibble(found_max)

## ----collect_information------------------------------------------------------
norm_info <- read.table("https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/datasets/normalize.txt", header = TRUE) # nolint

found2 <- found |>
  tidyr::gather(key = "statistic", value = "mean", -number_of_lineages)



found_min2 <- found_min |>
  tidyr::gather(key = "statistic", value = "min", -number_of_lineages)
found_max2 <- found_max |>
  tidyr::gather(key = "statistic", value = "max", -number_of_lineages)

found3 <- left_join(found2, norm_info,
                    by = join_by(statistic))
found3 <- left_join(found3, found_min2,
                    by = join_by(statistic, number_of_lineages))
found3 <- left_join(found3, found_max2,
                    by = join_by(statistic, number_of_lineages))

## ----yule_plots---------------------------------------------------------------
found3 |>
  filter(normalize == TRUE & type == "Yule") |>
  ggplot(aes(x = number_of_lineages, y = mean)) +
  geom_ribbon(aes(ymin = min, ymax = max), fill = "yellow", alpha = 0.2) +
  geom_point(col = "blue", alpha = 0.3) +
  geom_line(col = "gold") +
  scale_x_log10() +
  facet_wrap(~statistic, scales = "free_y", ncol = 3) +
  theme_minimal() +
  xlab("Extant tree size") +
  ggtitle("Statistics normalized by the Yule expectation")

## ----size_plots---------------------------------------------------------------
found3 |>
  filter(normalize == TRUE & type == "Tips") |>
  ggplot(aes(x = number_of_lineages, y = mean)) +
  geom_ribbon(aes(ymin = min, ymax = max), fill = "yellow", alpha = 0.2) +
  geom_point(col = "blue", alpha = 0.3) +
  geom_line(col = "gold") +
  scale_x_log10() +
  facet_wrap(~statistic, scales = "free_y", ncol = 4) +
  theme_minimal() +
  xlab("Extant tree size") +
  ggtitle("Statistics normalized by tree size")

## ----no_norm------------------------------------------------------------------
found3 |>
  filter(normalize == FALSE) |>
  ggplot(aes(x = number_of_lineages, y = mean)) +
  geom_ribbon(aes(ymin = min, ymax = max), fill = "yellow", alpha = 0.2) +
  geom_point(col = "blue", alpha = 0.3) +
  geom_line(col = "gold") +
  scale_x_log10() +
  facet_wrap(~statistic, scales = "free_y", ncol = 5) +
  theme_minimal() +
  xlab("Extant tree size") +
  ggtitle("Statistics not normalized")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
library(ggplot2)
library(tibble)
library(dendextend)
library(ggdendro)
library(pheatmap)

## ----load_data----------------------------------------------------------------
begin <- "https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/datasets/phylogenies/fracced/" # nolint

tree_collection_files <- c("amphibia_fracced.rds",
                           "birds_fracced.rds",
                           "ferns_fracced.rds",
                           "mammals_fracced.rds",
                           "ray_finned_fish_fracced.rds",
                           "sharks_fracced.rds",
                           "vascular_plants_fracced.rds")

taxa_names <- c("Amphibians", "Birds", "Ferns", "Mammals",
                "Ray finned Fish", "Cartaliginous Fish", "Vascular Plants")

found_stats <- c()
for (i in seq_along(tree_collection_files)) {
  raw_url <- paste0(begin, tree_collection_files[i])
  con <- url(raw_url, open = "rb")
  on.exit(close(con))

  tree_collection <- readRDS(con)

  families <- names(tree_collection)
  cat(i, "\n")
  for (j in seq_along(tree_collection)) {
    focal_tree <- tree_collection[[j]]
    if (length(focal_tree$tip.label) >= 10) {
      all_stats <- treestats::calc_all_stats(focal_tree, FALSE)
      to_add <- unlist(all_stats)
      to_add <- c(families[j], to_add)
      found_stats <- rbind(found_stats, to_add)
    }
  }
}

## ----add_random---------------------------------------------------------------
num_stats <- length(names(all_stats))
colnames(found_stats) <- c("Family", names(all_stats))
found_stats <- tibble::as_tibble(found_stats)
found_stats$random <- runif(n = length(found_stats$Family))

# remove failed calculations:
to_remove <- which(is.na(found_stats$sackin))
found_stats <- found_stats[-to_remove, ]

found_stats2 <- found_stats |>
  dplyr::mutate_at(2:(num_stats + 2), as.numeric)

df <- as.data.frame(found_stats2[, 2:(num_stats + 2)])

## ----correct_for_tree_size----------------------------------------------------
get_cor <- function(local_stats) {

  local_stats <- local_stats[!is.na(local_stats$gamma), ]
  local_stats <- local_stats[!is.na(local_stats$sackin), ]

  local_stats2 <- local_stats |>
    dplyr::mutate_at(2:(num_stats + 2), as.numeric)

  local_stats2 <- as.data.frame(local_stats2[, 2:(num_stats + 2)])

  res_cor <- stats::cor(as.data.frame(local_stats2), method = "pearson")

  res_cor2 <- res_cor

  for (i in seq_along(res_cor[1, ])) {
    for (j in seq_along(res_cor[1, ])) {
      stat1 <- colnames(res_cor)[i]
      stat2 <- colnames(res_cor)[j]

      if (stat1 != stat2) {
        if (stat1 != "number_of_lineages" && stat2 != "number_of_lineages") {
          x <- unlist(as.vector(local_stats2[stat1]))  # nolint
          y <- unlist(as.vector(local_stats2[stat2]))  # nolint
          z <- unlist(as.vector(local_stats2["number_of_lineages"])) # nolint

          a1 <- nlme::gls(y ~ z)
          a2 <- nlme::gls(x ~ z)

          found_cor <- cor(a1$residuals, a2$residuals)

          res_cor2[i, j] <- found_cor
        }
      }
    }
  }
  return(res_cor2)
}

master_cor <- get_cor(found_stats2)

## ----plot_dendrogram, fig.width=25, fig.height=25-----------------------------

cor1 <- master_cor

cor1 <- as.matrix(cor1)
to_remove <- which(colnames(cor1) == "number_of_lineages")

cor1 <- cor1[-to_remove, -to_remove]
cor1 <- as.data.frame(cor1)
cor1 <- tibble::as_tibble(cor1)
cor1 <- cor1 |>
  dplyr::mutate_at(seq_len(ncol(cor1)), as.numeric)

res_dist <- stats::as.dist(1 - abs(as.matrix(cor1)))

hc <- hclust(res_dist, method = "average")
dend0 <- stats::as.dendrogram(hc)
ddata <- ggdendro::dendro_data(hc, type = "rectangle")

xmax <- 0.8

clust_ref <- dendextend::cutree(dend0, h = xmax)
xmin <- 0
all_rect <- c() # xmin, xmax, ymin, ymax

for (a in unique(clust_ref)) {
  b <- clust_ref[clust_ref == a]
  in_plot <- subset(ddata$labels, ddata$labels$label %in% names(b))
  ymin <- min(in_plot$x) - 0.5
  ymax <- max(in_plot$x) + 0.5
  to_add <- c(xmin, xmax, ymin, ymax)
  all_rect <- rbind(all_rect, to_add)
}

rect_plot <- data.frame(xmin = all_rect[, 1],
                        xmax = all_rect[, 2],
                        ymin = all_rect[, 3],
                        ymax = all_rect[, 4],
                        categ = seq_along(all_rect[, 1]))

lvls <- sort(unique(as.numeric(rect_plot$categ)))
rect_plot$categ <- factor(rect_plot$categ, levels = lvls)

ggplot() +
  geom_rect(data = rect_plot,
            aes(xmin = ymin, xmax = ymax, ymin = xmin, ymax = xmax,
                fill = categ), alpha = 1) +
  scale_fill_brewer(type = "qual", palette = 3) +
  geom_segment(data = ggdendro::segment(ddata),
               aes(x = x, y = y, xend = xend, yend = yend)
  ) +
  geom_text(data = ggdendro::label(ddata),
            aes(x = x, y = y, label = label, hjust = 0),
            size = 10
  ) +
  coord_flip() +
  ylim(1, -0.3) +
  theme_classic() +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        legend.position = "none") +
  ylab("1 - abs(correlation)") +
  xlab("")

## ----heatmap, , fig.width=25, fig.height=25-----------------------------------
breakz <- seq(-1, 1, length.out = 99)
cor2 <- as.matrix(cor1)
hm1 <- pheatmap::pheatmap(mat = cor2,
                          breaks = breakz,
                          clustering_method = "average",
                          fontsize_col = 18,
                          fontsize_row = 18)

#' Mean Nearest Taxon distance
#' @description Per tip, evaluates the shortest distance to another tip,
#' then takes the average across all tips.
#' @param phy phylo object or ltable
#' @return Mean Nearest Taxon Distance.
#' @references  Webb, C., D. Ackerly, M. McPeek, and M. Donoghue. 2002.
#' Phylogenies and community ecology. Annual Review of Ecology and
#' Systematics 33:475-505.
#' @export
mntd <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    if (sum(phy[, 4] != -1)) {
      # we now have to do it brute force:
      phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
      return(brute_mntd(phy))
    }
    return(calc_mntd_ltable_cpp(phy))
  }
  if (inherits(phy, "phylo")) {
    if (!ape::is.ultrametric(phy, option = 2)) {
      return(brute_mntd(phy)) # slower though
    }
    return(calc_mntd_cpp(phy))
  }
  stop("input object has to be phylo or ltable")
}

#' @keywords internal
brute_mntd <- function(phy) {
  dist_mat <- ape::cophenetic.phylo(phy)
  diag(dist_mat) <- NA
  ntd <- apply(dist_mat, 1,
               function(x) {
                 return(min(x, na.rm = TRUE))
               }
              )
  return(mean(ntd))
}

#' Root imbalance
#' @description Measures the distribution of tips over the two crown lineages,
#' e.g. \eqn{n1 / (n1 + n2)}, where n1 is the number of tips connected to crown
#' lineage 1 and n2 is the number of tips connected to crown lineage 2. We
#' always take n1 > n2, thus root imbalance is always in [0.5, 1]. Root
#' imbalance is also calculated for polytomous trees, as long as there is no
#' polytomy at the root.
#' @param phy phylo object or ltable
#' @return Root imbalance
#' @references  Guyer, Craig, and Joseph B. Slowinski. "Adaptive radiation and
#' the topology of large phylogenies." Evolution 47.1 (1993): 253-263.
#' @export
root_imbalance <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  # although the tree need not be binary, the root has to be, luckily the test
  # for rootedness catches this, as ape::is.rooted checks for the number of
  # lineages connected to the crown, which has to be 2.


  if (inherits(phy, "matrix")) {
    return(calc_root_imbalance_ltable_cpp(phy))
  }
  if (inherits(phy, "phylo")) {
    return(calc_root_imbalance_cpp(as.vector(t(phy$edge))))
  }
  stop("input object has to be phylo or ltable")
}

#' Intensive quadratic entropy statistic J.
#' @description The intensive quadratic entropy statistic J is given by the
#' average distance between two randomly chosen species, thus given by the
#' sum of all pairwise distances, divided by \eqn{S^2}, where S is the number
#' of tips of the tree.
#' @param phy phylo object or ltable
#' @return intensive quadratic entropy statistic J
#' @references  Izsák, János, and Laszlo Papp. "A link between ecological
#' diversity indices and measures of biodiversity."
#' Ecological Modelling 130.1-3 (2000): 151-156.
#' @export
entropy_j <- function(phy) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
      n <- length(phy$tip.label)
      return(treestats::mean_pair_dist(phy) / n)
      # notice that the statistic divides by S^2, but since the mean is
      # already dividing by S, this is fine.
  }
  stop("input object has to be phylo or ltable")
}

context("entropy j")

test_that("usage", {
  if (requireNamespace("picante")) {
    a1 <- treestats::entropy_j(extant_tree)

    n <- length(extant_tree$tip.label)
    sample_mat <- matrix(data = 1, nrow = n, ncol = n)
    colnames(sample_mat) <- extant_tree$tip.label

    a2 <- picante::mpd(sample_mat, stats::cophenetic(extant_tree),
                       abundance.weighted = FALSE)[[1]]
    testthat::expect_equal(a1, a2 / n)

    ltab <- treestats::phylo_to_l(extant_tree)
    testthat::expect_equal(treestats::entropy_j(extant_tree),
                           treestats::entropy_j(ltab))
  }
})

test_that("polytomies", {
  if (requireNamespace("picante")) {

    test_func <- function(phy) {
      n <- length(phy$tip.label)
      sample_mat <- matrix(data = 1, nrow = n, ncol = n)
      colnames(sample_mat) <- phy$tip.label

      a2 <- picante::mpd(sample_mat, stats::cophenetic(phy),
                         abundance.weighted = FALSE)[[1]]
      return(a2 / n)
    }

    test_polytomies(treestats::entropy_j, test_func)
  }
})

test_that("unrooted", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
  a1 <- treestats::entropy_j(focal_tree)
  a2 <- treestats::entropy_j(ape::unroot(focal_tree))
  testthat::expect_equal(a1, a2)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::entropy_j(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::entropy_j(list()),
    "input object has to be phylo or ltable"
  )
})

#' Number of tips of a tree, including extinct tips.
#' @param phy phylo object
#' @return number of lineages
#' @export
number_of_lineages <- function(phy) {
  if (inherits(phy, "matrix")) {
    return(length(phy[, 1]))
  }

  if (inherits(phy, "phylo")) {
      return(length(phy$tip.label))
  }
  stop("input object has to be phylo or ltable")
}

context("number_of_lineages")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)

  num_lin <- treestats::number_of_lineages(focal_tree)

  testthat::expect_equal(num_lin, 100)

  focal_ltab <- treestats::phylo_to_l(focal_tree)
  num_lin2 <- treestats::number_of_lineages(focal_ltab)
  testthat::expect_equal(num_lin, 100)

  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0.2,
                            fossils = TRUE)

  num_lin <- treestats::number_of_lineages(focal_tree)

  testthat::expect_gte(num_lin, 100)

  focal_ltab <- treestats::phylo_to_l(focal_tree)
  num_lin2 <- treestats::number_of_lineages(focal_ltab)
  testthat::expect_gte(num_lin, 100)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::number_of_lineages(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::number_of_lineages(list()),
    "input object has to be phylo or ltable"
  )
})

#' Pigot's rho
#' @param phy phylo object
#' @return rho
#' @description Calculates the change in rate between the first half and the
#' second half of the extant phylogeny. \eqn{\rho = (r2 - r1) / (r1 + r2)},
#' where r reflects the rate in either the first or second half.
#' The rate within a half is given by \eqn{(log(n2) - log(n1) / t}, where n2 is
#' the number of lineages at the end of the half, and n1 the number of lineages
#' at the start of the half. Rho varies between -1 and 1, with a 0 indicating a
#' constant rate across the phylogeny, a \eqn{\rho < 0} indicating a slow down
#' and a \eqn{rho > 0} indicating a speed up of speciation. In contrast to the
#' Gamma statistic, Pigot's rho is not sensitive to tree size.
#' @references Alex L. Pigot, Albert B. Phillimore, Ian P. F. Owens,
#' C. David L. Orme, The Shape and Temporal Dynamics of Phylogenetic Trees
#' Arising from Geographic Speciation, Systematic Biology, Volume 59, Issue 6,
#' December 2010, Pages 660–673, https://doi.org/10.1093/sysbio/syq058
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' pigot_rho(simulated_tree) # should be around 0.
#' ddd_tree <- DDD::dd_sim(pars = c(1, 0, 10), age = 7)$tes
#' pigot_rho(ddd_tree) # because of diversity dependence, should be < 0
pigot_rho <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)


  if (inherits(phy, "matrix")) {
    return(calc_rho_ltable_cpp(phy))
  }

  if (inherits(phy, "phylo")) {
    if (!ape::is.ultrametric(phy)) {
      return(calc_rho_complete_cpp(phy))
    }
    return(calc_rho_cpp(phy))
  }

  stop("input object has to be phylo or ltable")
}

context("pigot_rho")

test_that("usage", {
  set.seed(42)

  if (requireNamespace("DDD")) {
    # DDD tree expected to slow down diversification
    focal_tree <- DDD::dd_sim(pars = c(1, 0, 10), age = 7)$tes
    rho <- treestats::pigot_rho(focal_tree)
    testthat::expect_lt(rho, 0)
  }

  set.seed(42)
  # BD tree, expected increase in diversification
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
  rho <- treestats::pigot_rho(focal_tree)
  testthat::expect_gt(rho, 0)

  ltab <- treestats::phylo_to_l(focal_tree)
  rho2 <- treestats::pigot_rho(ltab)
  testthat::expect_equal(rho, rho2)

  if (requireNamespace("geiger")) {
    # and we can do with extinct trees as well
    focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0.2,
                              fossils = TRUE)
    rho4 <- treestats::pigot_rho(focal_tree)
    extant_tree <- geiger::drop.extinct(focal_tree)
    rho5 <- treestats::pigot_rho(extant_tree)
    testthat::expect_false(rho4 == rho5)
  }

  # use very small tree to trigger use of complete method:
  focal_tree <- ape::rphylo(n = 6, birth = 1, death = 0)
  # compare to ltab, as this is always the same method
  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::pigot_rho(focal_tree),
                         treestats::pigot_rho(ltab))

  # and very large tree to trigger other method:
  focal_tree <- ape::rphylo(n = 300, birth = 1, death = 0)
  # compare to ltab, as this is always the same method
  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::pigot_rho(focal_tree),
                         treestats::pigot_rho(ltab))

})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    res <- try_stat(focal_tree, treestats::pigot_rho)
    testthat::expect_true(!is.na(res))
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::pigot_rho(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::pigot_rho(list()),
    "input object has to be phylo or ltable"
  )
})

#' Cherry index
#' @description Calculate the number of cherries, from the phyloTop package.
#' A cherry is a pair of sister tips.
#' @param input_obj phylo object or ltable
#' @param normalization "none", "yule", or "pda", the found number of
#' cherries is divided by the expected number, following
#' McKenzie & Steel 2000.
#' @return number of cherries
#' @references McKenzie, Andy, and Mike Steel. "Distributions of cherries for
#' two models of trees." Mathematical biosciences 164.1 (2000): 81-92.
#' @export
cherries <- function(input_obj, normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(input_obj,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  num_cherries <- 0

  if (inherits(input_obj, "matrix")) {
    num_cherries <- cherries_ltable_cpp(input_obj)
  } else if (inherits(input_obj, "phylo")) {
    num_cherries <- cherries_cpp(as.vector(t(input_obj$edge)))
  } else {
    stop("input object has to be phylo or ltable")
  }

  if (normalization == "yule") {
    if (inherits(input_obj, "matrix")) {
      n <- length(input_obj[, 1])
    }
    if (inherits(input_obj, "phylo")) {
      n <- length(input_obj$tip.label)
    }
    expected_num_cherries <- n / 3
    num_cherries <- num_cherries / expected_num_cherries
  }

  if (normalization == "pda") {
    if (inherits(input_obj, "matrix")) {
      n <- length(input_obj[, 1])
    }
    if (inherits(input_obj, "phylo")) {
      n <- length(input_obj$tip.label)
    }
    expected_num_cherries <- n / 4
    num_cherries <- num_cherries / expected_num_cherries
  }

  return(num_cherries)
}

context("cherries")

test_that("usage", {
  if (requireNamespace("phyloTop")) {
    c1 <- treestats::cherries(extant_tree)
    c2 <- phyloTop::cherries(extant_tree)
    testthat::expect_equal(c1, c2)

    c3 <- treestats::cherries(treestats::phylo_to_l(extant_tree))
    testthat::expect_equal(c1, c3)

    c1 <- treestats::cherries(complete_tree)
    c2 <- phyloTop::cherries(complete_tree)
    testthat::expect_equal(c1, c2)

    c3 <- treestats::cherries(treestats::phylo_to_l(complete_tree))
    testthat::expect_equal(c1, c3)
  }
})

test_that("normalisation", {
  c1 <- treestats::cherries(extant_tree)
  c2 <- treestats::cherries(extant_tree, normalization = "yule")
  c3 <- treestats::cherries(extant_tree, normalization = "pda")
  testthat::expect_lt(c2, c1)
  testthat::expect_lt(c3, c1)

  focal_ltab <- treestats::phylo_to_l(extant_tree)

  c4 <- treestats::cherries(focal_ltab)
  c5 <- treestats::cherries(focal_ltab, normalization = "yule")
  c6 <- treestats::cherries(focal_ltab, normalization = "pda")

  testthat::expect_equal(c1, c4)
  testthat::expect_equal(c2, c5)
  testthat::expect_equal(c3, c6)

  stats1 <- c()
  stats2 <- c()
  stats3 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::cherries(focal_tree))
    stats2 <- c(stats2, treestats::cherries(focal_tree,
                                            normalization = "yule"))
    stats3 <- c(stats3, treestats::cherries(focal_tree,
                                            normalization = "pda"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))
  a3 <- cor(stats3, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_equal(a2, a3)
})

test_that("polytomies", {
  # phyloTop does not support non-binary trees, so no comparison possible.
  for (focal_tree in poly_trees) {
   local_stat <- try_stat(focal_tree, treestats::cherries)
   testthat::expect_true(!is.na(local_stat))
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::cherries(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::cherries(list()),
    "input object has to be phylo or ltable"
  )
})

#' Normalized LTT statistic
#' @description The nLTT statistic calculates the sum of
#' absolute differences in the number of lineages over time, where both the
#' number of lineages and the time are normalized. The number of lineages is
#' normalized by the number of extant tips, whereas the time is normalized by
#' the crown age. The nLTT can only be calculated for reconstructed trees.
#' Only use the treestats version if you are very certain
#' about the input data, and are certain that performing nLTT is valid (e.g.
#' your tree is ultrametric etc). If you are less certain, use the nLTT function
#' from the nLTT package.
#' @param phy phylo object or ltable
#' @param ref_tree reference tree to compare with (should be same type as phy)
#' @references Janzen, T., Höhna, S. and Etienne, R.S. (2015), Approximate
#' Bayesian Computation of diversification rates from molecular phylogenies:
#' introducing a new efficient summary statistic, the nLTT. Methods Ecol Evol,
#' 6: 566-575. https://doi.org/10.1111/2041-210X.12350
#' @return number of lineages
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' reference_tree <- ape::rphylo(n = 10, birth = 0.2, death = 0)
#' nLTT(simulated_tree, reference_tree)
#' nLTT(simulated_tree, simulated_tree) # should be zero.
nLTT <- function(phy, # nolint
                 ref_tree) {

  if (inherits(phy, "phylo")) {
    if (inherits(ref_tree, "matrix")) {
      return(nLTT(ref_tree, phy))
    }

    return(calc_nltt_cpp(phy, ref_tree))
  }

  if (inherits(phy, "matrix")) {
    if (inherits(ref_tree, "phylo")) {
       ref_tree <- treestats::phylo_to_l(ref_tree)
    }
    return(calc_nltt_ltable_cpp(phy, ref_tree))
  }

  stop("input needs to be phylo or ltable object")
}

#' Reference nLTT statistic
#' @description The base nLTT statistic can be used as a semi stand-alone
#' statistic for phylogenetic trees. However, please note that although this
#' provides a nice way of checking the power of the nLTT statistic without
#' directly comparing two trees, the nLTT_base statistic is not a substitute
#' for directly comparing two phylogenetic trees. E.g. one would perhaps
#' naively assume that \eqn{nLTT(A, B) = |nLTT(A, base) - nLTT(B, base)}.
#' Indeed, in some cases this may hold true (when, for instance, all normalized
#' lineages of A are less than all normalized lineages of B), but once the
#' nLTT curve of A intersects the nLTT curve of B, this no longer applies.
#' @param phy phylo object
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' nLTT_base(simulated_tree)
#' @return number of lineages
#' @export
nLTT_base <- function(phy) {  # nolint

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = TRUE,
             require_rooted = TRUE)

  empty_tree <- ape::read.tree(text = "(1:4,2:4):0;")
  return(nLTT(phy, empty_tree))
}

context("nltt")

test_that("usage", {
  if (requireNamespace("nLTT")) {
    set.seed(42)
    tree1 <- ape::rphylo(n = 100, birth = 1, death = 0)
    tree2 <- ape::rphylo(n = 100, birth = 0.5, death = 0)

    nltt <- treestats::nLTT(tree1, tree2)
    nltt_check <- nLTT::nLTTstat(tree1, tree2)
    testthat::expect_equal(nltt, nltt_check)


    empty_tree <- ape::rphylo(n = 2, birth = 1, death = 0)
    nltt_base <- treestats::nLTT_base(tree1)
    nltt_base2 <- treestats::nLTT(tree1, empty_tree)
    nltt_base3 <-  nLTT::nLTTstat(tree1, empty_tree)

    testthat::expect_equal(nltt_base, nltt_base2, tolerance = 1e-3)
    testthat::expect_equal(nltt_base, nltt_base3, tolerance = 1e-3)

    empty_tree <- ape::read.tree(text = "(1:4,2:4):0;")
    nltt_stat <- nLTT::nltt_diff(tree1, empty_tree)
    testthat::expect_equal(nltt_stat, nltt_base, tolerance = 1e-3)
    testthat::expect_equal(nltt_stat, nltt_base2, tolerance = 1e-3)

    # test ltable functionality
    ltab1 <- treestats::phylo_to_l(tree1)
    ltab2 <- treestats::phylo_to_l(tree2)
    nltt3 <- treestats::nLTT(ltab1, ltab2)
    testthat::expect_equal(nltt3, nltt)

    # mixed ltable:
    nltt4 <- treestats::nLTT(ltab1, tree2)
    testthat::expect_equal(nltt4, nltt)

    nltt5 <- treestats::nLTT(tree1, ltab2)
    testthat::expect_equal(nltt5, nltt)
  }
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    if (ape::is.ultrametric(focal_tree)) {
      res <- try_stat(focal_tree, treestats::nLTT_base)
     testthat::expect_true(!is.na(res))
    }
  }
})

test_that("wrong_object", {
  set.seed(42)
  tree1 <- ape::rphylo(n = 10, birth = 1, death = 0)
  testthat::expect_error(
    treestats::nLTT(10, tree1),
    "input needs to be phylo or ltable object"
  )

  testthat::expect_error(
    treestats::nLTT(list(), tree1),
    "input needs to be phylo or ltable object"
  )
})

#' Rquartet index.
#' @description The rquartet index counts the number of potential fully
#' balanced rooted subtrees of 4 tips in the tree. The function in treestats
#' assumes a bifurcating tree. For trees with polytomies, we use
#' treebalance::rquartedI, which can also take polytomies into account.
#' @param phy phylo object or ltable
#' @param normalization The index can be normalized by the expectation under
#' the Yule ("yule") or PDA model ("pda").
#' @return rquartet index
#' @references  T. M. Coronado, A. Mir, F. Rosselló, and G. Valiente. A balance
#' index for phylogenetic trees based on rooted quartets. Journal of
#' Mathematical Biology, 79(3):1105-1148, 2019. doi: 10.1007/s00285-019-01377-w.
#' @export
rquartet <- function(phy, normalization = "none") {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  normalization <- check_normalization_key(normalization)

  if (!inherits(phy, "matrix") && !inherits(phy, "phylo")) {
    stop("input object has to be phylo or ltable")
  }

  if (inherits(phy, "matrix")) {
    answ <- calc_rquartet_ltable_cpp(phy)
  }
  if (inherits(phy, "phylo")) {
    if (!check_binary(phy)) {
      return(treebalance::rQuartetI(phy))
    }

    answ <- calc_rquartet_cpp(as.vector(t(phy$edge)))
  }

  if (normalization == "yule") {
    ntips <- ifelse(inherits(phy, "matrix"),
                    length(phy[, 1]),
                    length(phy$tip.label))
    answ <- answ / (choose(ntips, 4))
  }
  if (normalization == "pda") {
    ntips <- ifelse(inherits(phy, "matrix"),
                    length(phy[, 1]),
                    length(phy$tip.label))
    answ <- answ / (5 * choose(ntips, 4))
  }

  return(answ)
}

context("rquartet")

test_that("usage", {
  if (requireNamespace("treebalance")) {

    standard_test(treestats::rquartet, treebalance::rQuartetI)

    # normalization
    ltab <- treestats::phylo_to_l(extant_tree)
    a3 <- treestats::rquartet(extant_tree, normalization = "yule")
    a4 <- treestats::rquartet(ltab, normalization = "yule")
    testthat::expect_equal(a3, a4)
    testthat::expect_lt(a3, 2)

    a3 <- treestats::rquartet(extant_tree, normalization = "pda")
    a4 <- treestats::rquartet(ltab, normalization = "pda")
    testthat::expect_equal(a3, a4)
    testthat::expect_lt(a3, 2)
  }
})

test_that("polytomy", {
  if (requireNamespace("treebalance")) {
    focal_tree <- ape::read.tree(text = "(t1:1.5,(t2:1,t3:1,t4:1):0.5);")
    res <- treestats::rquartet(focal_tree)
    ref <- treebalance::rQuartetI(focal_tree)
    testthat::expect_equal(res, ref)

    test_polytomies(treestats::rquartet, treebalance::rQuartetI)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::rquartet(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::rquartet(list()),
    "input object has to be phylo or ltable"
  )
})

#' Diameter statistic
#' @description The Diameter of a tree is defined as the maximum length of a
#' shortest path. When taking branch lengths into account, this is equal to
#' twice the crown age.
#' When the tree is unrooted, we add 1 to the unweighted diameter, to reflect
#' traversing the (virtual) root.
#' @param phy phylo object or ltable
#' @param weight if TRUE, uses branch lengths.
#' @return Diameter
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." PloS one 16.12 (2021): e0259877.
#' @export
diameter <- function(phy,
                     weight = FALSE) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    diam_stat <- calc_diameter_ltable_cpp(phy, weight)
    return(diam_stat)
  }
  if (inherits(phy, "phylo")) {
    diam_stat <- NA
    if (ape::is.rooted(phy)) {
      diam_stat <- calc_diameter_cpp(as.vector(t(phy$edge)),
                                     phy$edge.length,
                                     weight)
    } else {
      if (!weight) {
        phy$edge.length <- rep(1, length(phy$edge.length))
      }
      dist_mat <- ape::cophenetic.phylo(phy)
      diam_stat <- max(dist_mat, na.rm = TRUE)
    }
    return(diam_stat)
  }
  stop("input object has to be phylo or ltable")
}

context("diameter")

test_that("usage", {
  a1_1 <- treestats::diameter(extant_tree, weight = TRUE)
  a1_2 <- treestats::diameter(extant_tree, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:

  a2_1 <- 8.750583  # treeCentrality::computeDiameter(focal_tree,
  #                                                   weight = TRUE))
  a2_2 <- 21        # treeCentrality::computeDiameter(focal_tree,
  #                                                      weight = FALSE))

  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)
  testthat::expect_equal(a1_2, a2_2)

  ltab <- treestats::phylo_to_l(extant_tree)
  testthat::expect_equal(treestats::diameter(extant_tree, weight = TRUE),
                         treestats::diameter(ltab, weight = TRUE))

  testthat::expect_equal(treestats::diameter(extant_tree, weight = FALSE),
                         treestats::diameter(ltab, weight = FALSE))

  a1_1 <- treestats::diameter(complete_tree, weight = TRUE)
  a1_2 <- treestats::diameter(complete_tree, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:
  a2_1 <- 11.21469 # treeCentrality::computeDiameter(focal_tree,
  #                                                  weight = TRUE))
  a2_2 <- 21       # treeCentrality::computeDiameter(focal_tree,
  #                                                  weight = FALSE))


  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)
  testthat::expect_equal(a1_2, a2_2)

  ltab <- treestats::phylo_to_l(complete_tree)
  testthat::expect_equal(treestats::diameter(complete_tree, weight = TRUE),
                         treestats::diameter(ltab,       weight = TRUE))

  testthat::expect_equal(treestats::diameter(complete_tree, weight = FALSE),
                         treestats::diameter(ltab,       weight = FALSE))
})

test_that("polytomies", {
  if (requireNamespace("igraph")) {

    test_func <- function(phy) {
      df <- as.data.frame(cbind(phy$edge,
                                weight = phy$edge.length))
      g <- igraph::graph_from_data_frame(df, directed = FALSE)
      # diameter by default works without weights
      return(igraph::diameter(g, weights = rep(1, length(phy$edge.length))))
    }
    test_polytomies(treestats::diameter, test_func)
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::diameter(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::diameter(list()),
    "input object has to be phylo or ltable"
  )
})

#' @keywords internal
check_normalization_key <- function(normalization) {
  output <- normalization
  if (output == "Yule") output <- "yule"
  if (output == "PDA")  output <- "pda"
  if (output == "Tips") output <- "tips"

  return(output)
}

#' Faster than the ape version
#' @description Fast Rcpp function to check if a tree is binary
#' @param phy phylo object
#' @export
check_binary <- function(phy) {
  return(check_is_binary_rcpp(as.vector(t(phy$edge))))
}

#' @keywords internal
check_tree <- function(phy,
                       require_binary = FALSE,
                       require_ultrametric = FALSE,
                       require_rooted = FALSE) {

  # early exit
  if (!require_binary &&
      !require_ultrametric &&
      !require_rooted) return()


  if (inherits(phy, "phylo")) {
    if (require_binary) {
      valid <- check_binary(phy)
      if (!valid) {
        stop("Tree is non-binary, statistic not applicable")
      }
    }
    if (require_ultrametric) {
      valid <- ape::is.ultrametric(phy, tol = 1e-7, option = 1)

      if (!valid) {
        stop("Tree is not ultrametric, statistic not applicable")
      }
    }

    if (require_rooted) {
      valid <- ape::is.rooted(phy)

      if (!valid) {
        stop("Tree is not rooted, statistic not applicable")
      }
    }

  }
  if (inherits(phy, "matrix")) {
    if (require_ultrametric) {
      valid <- sum(phy[, 4] != -1)
      if (valid > 0) {
        stop("Tree is not ultrametric, statistic not applicable")
      }
    }

    if (require_binary) {
        max_num_branch_events <- max(table(phy[, 1]))
        if (max_num_branch_events > 2) {
          stop("Tree is non-binary, statistic not applicable")
        }
    }
  }
}

#' Average leaf depth statistic.
#' @description The average leaf depth statistic
#' is a normalized version of the Sackin index, normalized by the number of
#' tips.
#' @param phy phylo object or ltable
#' @param normalization "none" or "yule", in which case the statistic is
#' divided by the expectation under the yule model, following Remark 1 in
#' Coronado et al. 2020.
#' @return average leaf depth statistic
#' @export
#' @references M. Coronado, T., Mir, A., Rosselló, F. et al. On Sackin’s
#' original proposal: the variance of the leaves’ depths as a phylogenetic
#' balance index. BMC Bioinformatics 21, 154 (2020).
#' https://doi.org/10.1186/s12859-020-3405-1
#' K.-T. Shao and R. R. Sokal. Tree balance. Systematic Zoology,
#' 39(3):266, 1990. doi: 10.2307/2992186.
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' average_leaf_depth(simulated_tree)
average_leaf_depth <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)


  if (inherits(phy, "phylo")) {
    n <- length(phy$tip.label)
    ald <- calc_sackin_cpp(as.vector(t(phy$edge)), "none") / n
    if (normalization == "yule" || normalization == TRUE) {
      expectation <- 2 * log(n)
      ald <- ald / expectation
    }
    return(ald)
  }

  if (inherits(phy, "matrix")) {
    n <- length(phy[, 1])
    ald <- calc_sackin_ltable_cpp(phy, "none") / n
    if (normalization == "yule" || normalization == TRUE) {
      expectation <- 2 * log(n)
      ald <- ald / expectation
    }
    return(ald)
  }

  stop("input object has to be phylo or ltable")
}

context("average leaf depth")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    ald <- treestats::average_leaf_depth(extant_tree)
    ald_check <- treebalance::avgLeafDepI(extant_tree)
    testthat::expect_equal(ald, ald_check)

    ltab <- treestats::phylo_to_l(extant_tree)
    testthat::expect_equal(treestats::average_leaf_depth(extant_tree),
                           treestats::average_leaf_depth(ltab))


    # with extinct species:
    ald <- treestats::average_leaf_depth(complete_tree)
    ald_check <- treebalance::avgLeafDepI(complete_tree)
    testthat::expect_equal(ald, ald_check)

    ltab <- treestats::phylo_to_l(complete_tree)
    testthat::expect_equal(treestats::average_leaf_depth(complete_tree),
                           treestats::average_leaf_depth(ltab))
  }
})

test_that("polytomies", {
  if (requireNamespace("treebalance")) {
    test_polytomies(treestats::average_leaf_depth, treebalance::avgLeafDepI)
  }
})

test_that("normalization", {
  c1 <- treestats::average_leaf_depth(extant_tree)
  c2 <- treestats::average_leaf_depth(extant_tree, normalization = "yule")
  testthat::expect_lt(c2, c1)
  c3 <- treestats::average_leaf_depth(treestats::phylo_to_l(extant_tree),
                            normalization = "yule")
  testthat::expect_equal(c2, c3)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::average_leaf_depth(focal_tree))
    stats2 <- c(stats2, treestats::average_leaf_depth(focal_tree,
                                                      normalization = "yule"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_lt(a2, 0.2)
})



test_that("wrong_object", {
  testthat::expect_error(
    treestats::average_leaf_depth(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::average_leaf_depth(list()),
    "input object has to be phylo or ltable"
  )
})

#' Mean Pairwise distance
#' @description Fast function using C++ to calculate the mean pairwise distance,
#' using the fast algorithm by Constantinos, Sandel & Cheliotis (2012).
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", in which case the obtained mean
#' pairwise distance is normalized by the factor 2log(n),
#' where n is the number of tips.
#' @return Mean pairwise distance
#' @references  Webb, C., D. Ackerly, M. McPeek, and M. Donoghue. 2002.
#' Phylogenies and community ecology. Annual Review of Ecology and
#' Systematics 33:475-505.
#'
#' Tsirogiannis, Constantinos, Brody Sandel, and Dimitris Cheliotis.
#' "Efficient computation of popular phylogenetic tree measures." Algorithms in
#' Bioinformatics: 12th International Workshop, WABI 2012, Ljubljana, Slovenia,
#' September 10-12, 2012. Proceedings 12. Springer Berlin Heidelberg, 2012.
#' @export
mean_pair_dist <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }

  if (inherits(phy, "phylo")) {
    if (check_binary(phy) && ape::is.rooted(phy)) {
      mpd <- calc_mpd_cpp(as.vector(t(phy$edge)),
                          phy$edge.length)
    } else {
      dist_mat <- ape::cophenetic.phylo(phy)
      diag(dist_mat) <- NA
      mpd <- mean(dist_mat, na.rm = TRUE)
    }
    if (normalization == "tips" || normalization == TRUE) {
      n <- length(phy$tip.label)
      mpd <- mpd / (2 * log(n))
    }
    return(mpd)
  }
  stop("input object has to be phylo or ltable")
}

context("mean pair distance")

test_that("usage", {
  if (requireNamespace("picante")) {
    a1 <- treestats::mean_pair_dist(extant_tree)

    n <- length(extant_tree$tip.label)
    sample_mat <- matrix(data = 1, nrow = n, ncol = n)
    colnames(sample_mat) <- extant_tree$tip.label

    a2 <- picante::mpd(sample_mat, cophenetic(extant_tree),
                       abundance.weighted = FALSE)[[1]]
    testthat::expect_equal(a1, a2)

    ltab <- treestats::phylo_to_l(extant_tree)
    testthat::expect_equal(treestats::mean_pair_dist(extant_tree),
                           treestats::mean_pair_dist(ltab))
  }
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    res <- treestats::mean_pair_dist(focal_tree)
    testthat::expect_true(!is.na(res))
  }
})

test_that("unrooted", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
  a1 <- treestats::mean_pair_dist(focal_tree)
  a2 <- treestats::mean_pair_dist(ape::unroot(focal_tree))
  testthat::expect_equal(a1, a2)
})

test_that("normalization", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 30, birth = 1, death = 0)

  c1 <- treestats::mean_pair_dist(focal_tree)
  c2 <- treestats::mean_pair_dist(focal_tree, normalization = "tips")
  testthat::expect_lt(c2, c1)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::mean_pair_dist(focal_tree))
    stats2 <- c(stats2, treestats::mean_pair_dist(focal_tree,
                                                  normalization = "tips"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_lt(a2, 0.3)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::mean_pair_dist(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_pair_dist(list()),
    "input object has to be phylo or ltable"
  )
})

#' \eqn{J^1} index.
#' @description The \eqn{J^1} index calculates the Shannon Entropy of a tree,
#' where at each node with two children, the Shannon Entropy is the sum of
#' \eqn{p_i log_2(p_i)} over the two children \eqn{i}, and \eqn{p_i} is
#' \eqn{L / (L + R)},
#' where L and R represent the number of tips connected to the two daughter
#' branches.
#' @param input_obj phylo object or ltable
#' @return \eqn{j^1} index
#' @references Jeanne Lemant, Cécile Le Sueur, Veselin Manojlović, Robert Noble,
#' Robust, Universal Tree Balance Indices, Systematic Biology, Volume 71,
#' Issue 5, September 2022, Pages 1210–1224
#'
#' https://doi.org/10.1093/sysbio/syac027
#' @export
j_one <- function(input_obj) {

  check_tree(input_obj,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(input_obj, "matrix")) {
    return(calc_j_one_ltable_cpp(input_obj))
  }
  if (inherits(input_obj, "phylo")) {
    return(calc_j_one_cpp(as.vector(t(input_obj$edge))))
  }

  stop("input object has to be phylo or ltable")
}

#' Total path length
#' @description The total path length describes the sums of the depths
#' of all vertices of the tree. This can also be calculated as the sum of the
#' total internal path and the sackin index.
#' @param phy phylo object or ltable
#' @return Total path length
#' @references  C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
#' transmission patterns. Evolution, Medicine, and Public Health,
#' 2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
#' @export
tot_path_length <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {
    t_p <- sackin(phy) + tot_internal_path(phy)
    return(t_p)
  }
  stop("input object has to be phylo or ltable")
}

context("tot_path")

test_that("usage", {

  if (requireNamespace("treebalance")) {
    standard_test(treestats::tot_path_length, treebalance::totPathLen)

    test_polytomies(treestats::tot_path_length, treebalance::totPathLen)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::tot_path_length(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::tot_path_length(list()),
    "input object has to be phylo or ltable"
  )
})

#' Apply all tree statistics related to branching times to a single tree.
#' @param phylo phylo object
#' @return list with statistics
#' @export
#' @description this function applies all tree statistics based on
#' branching times to a single tree (more or less ignoring topology), being:
#' \itemize{
#'   \item{gamma}
#'   \item{pigot's rho}
#'   \item{mean branch length}
#'   \item{nLTT with empty tree}
#'   \item{treeness}
#'   \item{var branch length}
#'   \item{mean internal branch length}
#'   \item{mean external branch length}
#'   \item{var internal branch length}
#'   \item{var external branch length}
#'   \item{mean of the sum of inverse branch lengths}
#' }
#'
calc_brts_stats <- function(phylo) {

  stats <- list()

  stats$gamma                  <- try_stat(phylo, treestats::gamma_statistic)
  stats$pigot_rho              <- try_stat(phylo, treestats::pigot_rho)
  stats$nltt_base              <- try_stat(phylo, treestats::nLTT_base)

  stats$treeness               <- try_stat(phylo, treestats::treeness)

  stats$mean_branch_length     <- try_stat(phylo, treestats::mean_branch_length)
  stats$var_branch_length      <- try_stat(phylo, treestats::var_branch_length)

  stats$mean_branch_length_int <- try_stat(phylo,
                                           treestats::mean_branch_length_int)
  stats$mean_branch_length_ext <- try_stat(phylo,
                                           treestats::mean_branch_length_ext)
  stats$var_branch_length_int  <- try_stat(phylo,
                                           treestats::var_branch_length_int)
  stats$var_branch_length_ext  <- try_stat(phylo,
                                           treestats::var_branch_length_ext)

  inv_path_dist <- try_stat(phylo, treestats::inv_branch_dist)
  stats$mean_inv_branch_dist <- mean(inv_path_dist)

  stats <- unlist(stats)
  stats <- stats[order(names(stats))]

  return(stats)
}

context("brts_stats")

test_that("usage", {
  testthat::skip_on_cran() # these tests take very long
  set.seed(42)
  focal_tree <- ape::rphylo(n = 10, birth = 1, death = 0)

  testthat::expect_invisible(
    all_stats <- treestats::calc_brts_stats(focal_tree)
  )
  testthat::expect_equal(length(all_stats), 11)
})

#' Symmetry nodes metric
#' @description Balance metric that returns the total number of internal nodes
#' that are not-symmetric (confusingly enough). A node is considered symmetric
#' when both daughter trees have the same topology, measured as having the
#' same sum of depths, where depth is measured as the distance from the root
#' to the node/tip.
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", in which case the resulting statistic
#' is divided by the number of tips - 2 (e.g. the maximum value of the symmetry
#' nodes index for a tree).
#' @return Maximum depth (in number of edges)
#' @references  S. J. Kersting and M. Fischer. Measuring tree balance using
#' symmetry nodes — A new balance index and its extremal properties.
#' Mathematical Biosciences, page 108690, 2021. ISSN 0025-5564.
#' doi:https://doi.org/10.1016/j.mbs.2021.108690
#' @export
sym_nodes <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {
    sym_nodes_stat <- calc_sym_nodes_cpp(as.vector(t(phy$edge)))
    if (normalization == "tips" || normalization == TRUE) {
      sym_nodes_stat <- sym_nodes_stat / (length(phy$tip.label) - 2)
    }
    return(sym_nodes_stat)
  }
  stop("input object has to be phylo or ltable")
}

context("util-functions")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 5, birth = 1, death = 0)

  c1 <- treestats::colless(focal_tree, "yule")
  c2 <- treestats::colless(focal_tree, "Yule")
  testthat::expect_equal(c1, c2)

  c1 <- treestats::colless(focal_tree, "pda")
  c2 <- treestats::colless(focal_tree, "PDA")
  testthat::expect_equal(c1, c2)

  sn1 <- treestats::sym_nodes(focal_tree, "tips")
  sn2 <- treestats::sym_nodes(focal_tree, "Tips")
  testthat::expect_equal(sn1, sn2)
})

#' Apply all available tree statistics to a single tree
#' @param phylo phylo object
#' @param normalize if set to TRUE, results are normalized (if possible) under
#' either the  Yule expectation (if available), or the number of tips
#' @return List with statistics
#' @export
#' @description this function applies all tree statistics available in
#' this package to a
#' single tree, being:
#' \itemize{
#'   \item gamma
#'   \item Sackin
#'   \item Colless
#'   \item corrected Colless
#'   \item quadratic Colless
#'   \item Aldous' beta statistic
#'   \item Blum
#'   \item crown age
#'   \item tree height
#'   \item Pigot's rho
#'   \item number of lineages
#'   \item nLTT with empty tree
#'   \item phylogenetic diversity
#'   \item avgLadder index
#'   \item cherries
#'   \item double cherries
#'   \item ILnumber
#'   \item pitchforks
#'   \item stairs
#'   \item stairs2
#'   \item laplacian spectrum
#'   \item B1
#'   \item B2
#'   \item area per pair (aPP)
#'   \item average leaf depth (aLD)
#'   \item I statistic
#'   \item ewColless
#'   \item max Delta Width (maxDelW)
#'   \item maximum of Depth
#'   \item variance of Depth
#'   \item maximum Width
#'   \item Rogers
#'   \item total Cophenetic distance
#'   \item symmetry Nodes
#'   \item mean of pairwise distance (mpd)
#'   \item variance of pairwise distance (vpd)
#'   \item Phylogenetic Species Variability (psv)
#'   \item mean nearest taxon distance (mntd)
#'   \item J statistic of entropy
#'   \item rquartet index
#'   \item Wiener index
#'   \item max betweenness
#'   \item max closeness
#'   \item diameter, without branch lenghts
#'   \item maximum eigen vector value
#'   \item mean branch length
#'   \item variance of branch length
#'   \item mean external branch length
#'   \item variance of external branch length
#'   \item mean internal branch length
#'   \item variance of internal branch length
#'   \item number of imbalancing steps
#'   \item j_one statistic
#'   \item treeness statistic
#'   \item branch weighted colless statistic
#'   \item minimum eigenvalue of the Laplacian matrix
#'   \item maximum eigenvalue of the Laplacian matrix
#'   \item minimum eigenvalue of the adjacency matrix
#'   \item maximum eigenvalue of the adjacency matrix
#'   \item mean of the sum of inverse branch lengths
#' }
#'
#' For the Laplacian spectrum properties, four properties of the eigenvalue
#' distribution are returned: 1) asymmetry, 2) peakedness, 3) log(principal
#' eigenvalue) and 4) eigengap.
#' Please notice that for some very small or very large trees, some of the
#' statistics can not be calculated. The function will report an NA for this
#' statistic, but will not break, to facilitate batch analysis of large numbers
#' of trees.
#' @rawNamespace import(Rcpp)
#' @rawNamespace import(nloptr)
#' @rawNamespace useDynLib(treestats)
calc_all_stats <- function(phylo, normalize = FALSE) {

  stats <- list()

  stats$gamma              <- try_stat(phylo, treestats::gamma_statistic)

  stats$sackin             <- try_stat(phylo, treestats::sackin,
                                      normalize, c("yule", "none"))

  stats$colless            <- try_stat(phylo, treestats::colless,
                                         normalize, c("yule", "none"))
  stats$colless_corr       <- try_stat(phylo, treestats::colless_corr,
                                       normalize, c("yule", "none"))
  stats$colless_quad       <- try_stat(phylo, treestats::colless_quad,
                                       normalize, c("yule", "none"))

  stats$beta               <- try_stat(phylo, treestats::beta_statistic)

  #stats$blum               <- try_stat(phylo, treestats::blum,
  #                                     normalize, c(TRUE, FALSE))
  stats$sshape             <- try_stat(phylo, treestats::sshape,
                                       normalize, c(TRUE, FALSE))

  stats$crown_age          <- try_stat(phylo, treestats::crown_age)
  stats$tree_height        <- try_stat(phylo, treestats::tree_height)
  stats$pigot_rho          <- try_stat(phylo, treestats::pigot_rho)

  stats$number_of_lineages <- try_stat(phylo, treestats::number_of_lineages)
  stats$nltt_base          <- try_stat(phylo, treestats::nLTT_base)
  stats$phylogenetic_div   <- try_stat(phylo, treestats::phylogenetic_diversity)

  stats$avg_ladder         <- try_stat(phylo, treestats::avg_ladder)
  stats$max_ladder         <- try_stat(phylo, treestats::max_ladder)

  stats$cherries           <- try_stat(phylo, treestats::cherries,
                                     normalize, c("yule", "none"))

  stats$double_cherries    <- try_stat(phylo, treestats::double_cherries)
  stats$four_prong         <- try_stat(phylo, treestats::four_prong)

  stats$il_number          <- try_stat(phylo, treestats::ILnumber,
                                       normalize, c("tips", "none"))

  stats$pitchforks         <- try_stat(phylo, treestats::pitchforks,
                                       normalize, c("tips", "none"))

  stats$stairs              <- try_stat(phylo, treestats::stairs)

  temp_stats <- tryCatch(expr = {treestats::laplacian_spectrum(phylo) }, #nolint
                         error = function(e) {return(NA) }) #nolint

  if (length(temp_stats) == 5) {
    stats$laplace_spectrum_a  <- temp_stats$asymmetry
    stats$laplace_spectrum_p  <- temp_stats$peakedness
    stats$laplace_spectrum_e  <- log(temp_stats$principal_eigenvalue)
    stats$laplace_spectrum_g  <- temp_stats$eigengap[[1]]
  } else {
    stats$laplace_spectrum_a  <- NA
    stats$laplace_spectrum_p  <- NA
    stats$laplace_spectrum_e  <- NA
    stats$laplace_spectrum_g  <- NA
  }

  temp_stats <- try_stat(phylo,
                      function(x) {
                        return(treestats::minmax_laplace(x, TRUE))
                       })
  stats$max_laplace <- NA
  stats$min_laplace <- NA

  obj_not_na <- !is.null(temp_stats) &&
    !(length(temp_stats) == 1 && is.na(temp_stats))

  max_not_na <- obj_not_na && !is.na(temp_stats$max)
  min_not_na <- obj_not_na && !is.null(temp_stats$min)

  if (max_not_na) stats$max_laplace <- temp_stats$max
  if (min_not_na) stats$min_laplace <- temp_stats$min

  temp_stats <- try_stat(phylo,
                         function(x) {
                           return(treestats::minmax_adj(x, TRUE))
                         })

  stats$min_adj <- NA
  stats$max_adj <- NA

  obj_not_na <- !is.null(temp_stats) &&
                !(length(temp_stats) == 1 && is.na(temp_stats))

  max_not_na <- obj_not_na && !is.na(temp_stats$max)
  min_not_na <- obj_not_na && !is.null(temp_stats$min)

  if (max_not_na) stats$max_adj <- temp_stats$max
  if (min_not_na) stats$min_adj <- temp_stats$min

  stats$imbalance_steps  <- try_stat(phylo, treestats::imbalance_steps,
                                     normalize)

  stats$j_one              <- try_stat(phylo, treestats::j_one)

  stats$b1                 <- try_stat(phylo, treestats::b1,
                                       normalize, c("tips", "none"))

  stats$b2                 <- try_stat(phylo, treestats::b2,
                                       normalize, c("yule", "none"))

  stats$area_per_pair      <- try_stat(phylo, treestats::area_per_pair,
                                       normalize, c("yule", "none"))

  stats$average_leaf_depth <- try_stat(phylo, treestats::average_leaf_depth,
                                       normalize, c("yule", "none"))

  stats$i_stat             <- try_stat(phylo, treestats::mean_i)

  stats$ew_colless         <- try_stat(phylo, treestats::ew_colless)

  stats$max_del_width      <- try_stat(phylo, treestats::max_del_width,
                                       normalize, c("tips", "none"))

  stats$max_depth          <- try_stat(phylo, treestats::max_depth,
                                       normalize, c("tips", "none"))

  stats$avg_vert_depth     <- try_stat(phylo, treestats::avg_vert_depth)

  stats$max_width          <- try_stat(phylo, treestats::max_width,
                                       normalize, c("tips", "none"))

  stats$mw_over_md         <- try_stat(phylo, treestats::mw_over_md)

  stats$tot_path           <- try_stat(phylo, treestats::tot_path_length)

  stats$tot_internal_path  <- try_stat(phylo, treestats::tot_internal_path)

  stats$rogers             <- try_stat(phylo, treestats::rogers,
                                       normalize, c("tips", "none"))


  stats$stairs2            <- try_stat(phylo, treestats::stairs2)

  stats$tot_coph           <- try_stat(phylo, treestats::tot_coph,
                                       normalize, c("yule", "none"))

  stats$var_depth          <- try_stat(phylo, treestats::var_leaf_depth,
                                       normalize, c("yule", "none"))

  stats$symmetry_nodes     <- try_stat(phylo, treestats::sym_nodes,
                                       normalize, c("tips", "none"))

  stats$mpd                <- try_stat(phylo, treestats::mean_pair_dist,
                                       normalize, c("tips", "none"))

  stats$psv                <- try_stat(phylo, treestats::psv,
                                       normalize, c("tips", "none"))

  stats$vpd                <- try_stat(phylo, treestats::var_pair_dist)

  stats$mntd               <- try_stat(phylo, treestats::mntd)

  stats$j_stat             <- try_stat(phylo, treestats::entropy_j)

  stats$rquartet           <- try_stat(phylo, treestats::rquartet,
                                       normalize, c("yule", "none"))

  stats$wiener             <- try_stat(phylo, treestats::wiener, normalize)

  stats$max_betweenness    <- try_stat(phylo, treestats::max_betweenness,
                                       normalize, c("tips", "none"))

  local_closeness <- function(tree, w, n) {
    return(treestats::max_closeness(tree, weight = w,
                                    normalization =
                                      ifelse(n == TRUE, "tips", "none")))
  }

  stats$max_closeness      <- try_stat(phylo, function(x) {
                                        local_closeness(x, FALSE, normalize)})

  stats$max_closenessW     <- try_stat(phylo, function(x) {
                                        local_closeness(x, TRUE, normalize)})

  stats$diameter           <- try_stat(phylo, treestats::diameter)

  stats$eigen_centrality   <- try_stat(phylo,
                                       function(x) {
                          return(max(treestats::eigen_centrality(x,
                                       weight = FALSE)$eigenvector))}) #nolint

  stats$eigen_centralityW  <- try_stat(phylo,
                                       function(x) {
                                      return(max(treestats::eigen_centrality(x,
                                           weight = TRUE)$eigenvector))}) #nolint


  stats$mean_branch_length <- try_stat(phylo, treestats::mean_branch_length)
  stats$var_branch_length  <- try_stat(phylo, treestats::var_branch_length)

  stats$mean_branch_length_int <- try_stat(phylo,
                                           treestats::mean_branch_length_int)
  stats$mean_branch_length_ext <- try_stat(phylo,
                                           treestats::mean_branch_length_ext)
  stats$var_branch_length_int  <- try_stat(phylo,
                                           treestats::var_branch_length_int)
  stats$var_branch_length_ext  <- try_stat(phylo,
                                           treestats::var_branch_length_ext)

  stats$treeness <- try_stat(phylo, treestats::treeness)

  stats$root_imbalance <- try_stat(phylo, treestats::root_imbalance)

  inv_path_dist <- try_stat(phylo, treestats::inv_branch_dist)
  stats$mean_inv_branch_dist <- mean(inv_path_dist)

  stats <- unlist(stats)
  stats <- stats[order(names(stats))]

  return(stats)
}

#' @keywords internal
try_stat <- function(phylo,
                     func,
                     normalize = FALSE,
                     norm_res = c(TRUE, FALSE)) {
  res <- NA
  if (normalize) {
    res <- tryCatch(expr = {
                            func(phylo,
                                 normalization =
                                   ifelse(normalize,
                                          norm_res[1], norm_res[2]))
                           },
                    error = function(e) {return(NA_real_)})     #nolint
  } else {
    res <- tryCatch(expr = {func(phylo)},  #nolint
                    error = function(e) {return(NA_real_)})     #nolint
  }
  return(res)
}

context("list_statistics")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 5, birth = 1, death = 0)
  all_stats <- treestats::calc_all_stats(focal_tree)
  names_all_stats <- names(all_stats)
  all_names <- treestats::list_statistics(only_balance_stats = FALSE)
  testthat::expect_equal(sum(all_names %in% names_all_stats),
                         length(all_names))

  testthat::expect_equal(length(names_all_stats),
                         length(all_names))

  all_stats <- treestats::calc_topology_stats(focal_tree)
  names_all_stats <- names(all_stats)
  all_names <- treestats::list_statistics(only_balance_stats = TRUE)
  testthat::expect_equal(sum(all_names %in% names_all_stats),
                         length(all_names))
})

context("all_statistics")

test_that("usage", {
  testthat::skip_on_cran() # these tests take very long

  set.seed(42)
  focal_tree <- ape::rphylo(n = 10, birth = 1, death = 0)

  testthat::expect_invisible(
   all_stats <- treestats::calc_all_stats(focal_tree)
  )
  testthat::expect_equal(length(all_stats), 71)

  focal_tree <- ape::rphylo(n = 23175, birth = 1, death = 0)
  testthat::expect_warning(
    all_stats <- treestats::calc_all_stats(focal_tree)
  )

  testthat::expect_true(
    is.na(all_stats[names(all_stats) == "laplace_spectrum_a"]))
  testthat::expect_true(
    is.na(all_stats[names(all_stats) == "laplace_spectrum_p"]))
  testthat::expect_true(
    is.na(all_stats[names(all_stats) == "laplace_spectrum_e"]))
  testthat::expect_true(
    is.na(all_stats[names(all_stats) == "laplace_spectrum_g"]))

  testthat::expect_true(is.na(all_stats[names(all_stats) == "min_laplace"]))
  testthat::expect_true(is.na(all_stats[names(all_stats) == "min_adj"]))

  testthat::expect_true(is.na(all_stats[names(all_stats) == "vpd"]))
})

context("IL number")

test_that("usage", {
  if (requireNamespace("phyloTop")) {
     c1 <- treestats::ILnumber(extant_tree)
    c2 <- phyloTop::ILnumber(extant_tree)
    testthat::expect_equal(c1, c2)

    c3 <- treestats::ILnumber(treestats::phylo_to_l(extant_tree))
    testthat::expect_equal(c1, c3)

    c1 <- treestats::ILnumber(complete_tree)
    c2 <- phyloTop::ILnumber(complete_tree)
    testthat::expect_equal(c1, c2)

    c3 <- treestats::ILnumber(treestats::phylo_to_l(complete_tree))
    testthat::expect_equal(c1, c3)
  }
})

test_that("normalization", {
  c1 <- treestats::ILnumber(extant_tree)
  c2 <- treestats::ILnumber(extant_tree, normalization = "tips")
  testthat::expect_lt(c2, c1)
  c3 <- treestats::ILnumber(treestats::phylo_to_l(extant_tree),
                            normalization = "tips")
  testthat::expect_equal(c2, c3)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::ILnumber(focal_tree))
    stats2 <- c(stats2, treestats::ILnumber(focal_tree, normalization = "tips"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_lt(a2, 0.2)
})

test_that("polytomies", {
  # phyloTop does not apply to non-binary trees, treestats does!
  for (focal_tree in poly_trees) {
    local_stat <- try_stat(focal_tree, treestats::ILnumber)
    testthat::expect_true(!is.na(local_stat))
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::ILnumber(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::ILnumber(list()),
    "input object has to be phylo or ltable"
  )
})

context("avgLadder")

test_that("usage", {

  if (requireNamespace("phyloTop")) {
    standard_test(treestats::avg_ladder, phyloTop::avgLadder)
  }
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    res <- try_stat(focal_tree, treestats::avg_ladder)
    testthat::expect_true(is.na(res))
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::avg_ladder(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::avg_ladder(list()),
    "input object has to be phylo or ltable"
  )
})

context("treeness")

test_that("usage", {
  set.seed(42)
  # DDD tree has all speciation events at the beginning
  # expect very low stemminess
  dd_tree1 <- DDD::dd_sim(pars = c(1, 0, 10), age = 20)$tas
  treeness_1 <- treestats::treeness(dd_tree1)
  testthat::expect_lt(treeness_1, 0.5)

  dd_tree1_ltab <- treestats::phylo_to_l(dd_tree1)
  treeness1_ltab <- treestats::treeness(dd_tree1_ltab)
  testthat::expect_equal(treeness_1, treeness1_ltab)


  dd_tree2 <- DDD::dd_sim(pars = c(0.3, 0, 100), age = 20)$tas
  treeness_2 <- treestats::treeness(dd_tree2)
  testthat::expect_gt(treeness_2, treeness_1)

  dd_tree2_ltab <- treestats::phylo_to_l(dd_tree2)
  treeness_2_ltab <- treestats::treeness(dd_tree2_ltab)
  testthat::expect_equal(treeness_2, treeness_2_ltab)
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    res <- try_stat(focal_tree, treestats::treeness)
    testthat::expect_true(!is.na(res))
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::treeness(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::treeness(list()),
    "input object has to be phylo or ltable"
  )
})

context("mean I")

test_that("usage", {
  if (requireNamespace("treebalance")) {

    test_func <- function(phy) {
      # treebalance allows 'few' polytomies
      # but that is too hard to verify
      return(treebalance::IbasedI(phy, method = "mean",
                           correction = "prime", logs = FALSE))
    }

    standard_test(treestats::mean_i, test_func)
  }
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    a1 <- try_stat(focal_tree, treestats::mean_i)
    testthat::expect_true(is.na(a1))
  }
})

test_that("abuse", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 3, birth = 1, death = 0)
  testthat::expect_warning(treestats::mean_i(focal_tree),
              "I statistic is only available for trees with at least 4 tips.")

  focal_ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_warning(treestats::mean_i(focal_ltab),
              "I statistic is only available for trees with at least 4 tips.")
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::mean_i(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_i(list()),
    "input object has to be phylo or ltable"
  )
})

context("phylodiv")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 3, birth = 1, death = 0)

  div1 <- treestats::phylogenetic_diversity(focal_tree)
  div2 <- sum(focal_tree$edge.length)
  testthat::expect_equal(div1, div2, tolerance = 1e-4)

  ca <- max(treestats::branching_times(focal_tree))
  pds <- treestats::phylogenetic_diversity(focal_tree,
                                           t = seq(ca, 0, length.out =  100))
  testthat::expect_equal(length(pds), 100)
  for (i in 2:length(pds)) {
    testthat::expect_gt(pds[i], pds[i - 1])
  }

  # now check sub time
  brts <- ape::branching.times(focal_tree)
  tt <- (brts[1] + brts[2]) / 2
  div1 <- treestats::phylogenetic_diversity(focal_tree, tt)
  div2 <- (brts[1] - tt) * 2
  testthat::expect_equal(div1[[1]], div2[[1]])

  # now with extinct lineages
  focal_tree <- ape::read.tree(text =
                          "((t1:2.0, t2:2.0):1.0, (t3:1.0, t4:2.0):1.0):1.0;")

  div1 <- treestats::phylogenetic_diversity(focal_tree)
  testthat::expect_equal(div1, 8)

  div1 <- treestats::phylogenetic_diversity(focal_tree, 2.0)
  div2 <- treestats::phylogenetic_diversity(focal_tree, 2.02)
  testthat::expect_gt(div1, div2)

  # now with double extinct lineages
  focal_tree <- ape::read.tree(
    text = "((t1:2,t2:2):2,(t5:3,(t3:1,t4:0.5):1):1);")
  div1 <- treestats::phylogenetic_diversity(focal_tree)
  testthat::expect_equal(div1, 10)
  div1 <- treestats::phylogenetic_diversity(focal_tree, 2)
  div2 <- treestats::phylogenetic_diversity(focal_tree, 2.01)
  testthat::expect_gt(div1, div2)
})

test_that("usage 2", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 50, birth = 1, death = 0)

  div1 <- treestats::phylogenetic_diversity(focal_tree)
  div2 <- sum(focal_tree$edge.length)
  testthat::expect_equal(div1, div2, tolerance = 1e-4)

  ltab <- treestats::phylo_to_l(focal_tree)
  div3 <- treestats::phylogenetic_diversity(ltab)
  testthat::expect_equal(div3, div2)
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
      res <- try_stat(focal_tree, treestats::phylogenetic_diversity)
      testthat::expect_true(!is.na(res))
  }
})

test_that("ltab", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 3, birth = 1, death = 0)

  focal_ltab <- treestats::phylo_to_l(focal_tree)

  div1 <- treestats::phylogenetic_diversity(focal_ltab)

  testthat::expect_error(treestats::phylogenetic_diversity(focal_ltab,
                                                           t = 1))

  testthat::expect_error(treestats::phylogenetic_diversity(focal_ltab,
                                                           t = c(0.0, 1.0)))

  focal_tree <- ape::rphylo(n = 30, birth = 1, death = 0.3, fossils = TRUE)
  focal_ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_error(treestats::phylogenetic_diversity(focal_ltab))
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::phylogenetic_diversity(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::phylogenetic_diversity(list()),
    "input object has to be phylo or ltable"
  )
})

context("polytomies")

test_that("usage", {

  # Thanks to Fien Strijthaegen for providing a testing tree

  poly_trees <- c("(151:34,(152:2,((((153:10,154:16):1,155:24):2,156:2):6,(157:5,((158:0,(((159:0,160:5):14,161:1):0,((((162:4,163:0):6,164:0):0,(165:22,166:4):3):1,167:13):4):9):3,(168:25,169:3):29):2):0):2):0,170:11,171:4):0;", #nolint
                  "((t6:1,t2:1)1:0.5,(((t15:1,t8:1,t11:1)1:1,t12:1,t13:1,t10:1)1:1,((t9:1,(t4:1,t14:1)1:1)1:1,t1:1)1:1,(t3:1,t7:1,t5:1)1:1)1:0.5);", #nolint
                  "(((t9:1,t14:1)1:1,t4:1,t12:1)1:0.5,(((t6:1,t19:1,t10:1,t2:1)1:1,t8:1,(t5:1,t16:1)1:1)1:1,(((((t11:1,t17:1)1:1,(t7:1,t3:1)1:1)1:1,(t1:1,t18:1)1:1)1:1,t15:1)1:1,t13:1)1:1)1:0.5);", #nolint
                  "((t1:2,t2:2,t3:2):1,((t4:1,t5:1):1,(t6:1,t7:1,t8:1,(t9:0.5,t10:0.5):0.5):1):1);", #nolint  ultrametric tree with polytomies, from chatgpt
                  "(t1:3,(t2:1,t3:1):2,(t4:1,(t5:1,t6:1):1):1,(t7:1,t8:1,t9:1,t10:1):2);")   #nolint     tree with polytomy at the root


                  # nolint
  for (poly_text in poly_trees) {
    poly_tree <- ape::read.tree(text = poly_text)
    all_stats <- treestats::calc_all_stats(poly_tree)

    na_stats <- names(all_stats)[which(is.na(all_stats))]
    non_na_stats <- names(all_stats)[which(!is.na(all_stats))]

    names_na_stats <- c("avg_ladder", "beta",
                        "colless", "colless_corr", "colless_quad",
                        "diameter", "double_cherries",
                        "eigen_centrality", "eigen_centralityW",
                        "ew_colless", "four_prong", "gamma",
                        "i_stat", "imbalance_steps", "j_one",
                        "max_adj", "max_betweenness",
                        "max_ladder", "max_laplace", "min_adj",
                        "min_laplace", "rogers",
                        "stairs", "stairs2", "symmetry_nodes",
                        "wiener",
                        "root_imbalance",
                        "mean_inv_branch_dist", # TODO: make this doable
                        "nltt_base") # TODO: nLTT is NA because of ultrametric!

    a <- na_stats %in% names_na_stats
    b <- names_na_stats %in% na_stats
 # disable tests for now (29-06-2026)
    #   testthat::expect_all_true(a)
 #    testthat::expect_all_true(b)

    # test phylo_to_l and rebase_ltable on polytomies:
    testthat::expect_silent(
      ltab  <- treestats::phylo_to_l(poly_tree))
    testthat::expect_silent(
      ltab2 <- treestats::rebase_ltable(ltab)
    )
  }
})

test_that("manual trees", {
  # these trees were provided by Luise Haeuser.

  poly_tree <- ape::read.tree(text = "(((t9:1,t14:1)1:1,t4:1,t12:1)1:0.5,(((t6:1,t19:1,t10:1,t2:1)1:1,t8:1,(t5:1,t16:1)1:1)1:1,(((((t11:1,t17:1)1:1,(t7:1,t3:1)1:1)1:1,(t1:1,t18:1)1:1)1:1,t15:1)1:1,t13:1)1:1)1:0.5);") #nolint
  il_val <- treestats::ILnumber(poly_tree)
  testthat::expect_equal(il_val, 3) # reference value through solving by eye

  b1_val <- treestats::b1(poly_tree)
  b1_ref <- treebalance::B1I(poly_tree)
  testthat::expect_equal(b1_val, b1_ref)

  poly_tree <- ape::read.tree(text = "((t6:1,t2:1)1:0.5,(((t15:1,t8:1,t11:1)1:1,t12:1,t13:1,t10:1)1:1,((t9:1,(t4:1,t14:1)1:1)1:1,t1:1)1:1,(t3:1,t7:1,t5:1)1:1)1:0.5);") #nolint
  tip1 <- treestats::tot_internal_path(poly_tree)
  tip2 <- treebalance::totIntPathLen(poly_tree)
  testthat::expect_equal(tip1, tip2)

  b1_val <- treestats::b1(poly_tree)
  b1_ref <- treebalance::B1I(poly_tree)
  testthat::expect_equal(b1_val, b1_ref)
})

context("imbal_steps")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 32, birth = 1, death = 0)
  focal_tree <- treestats::create_fully_balanced_tree(focal_tree)
  a1 <- treestats::imbalance_steps(focal_tree)
  a2 <- 32 - log(32, 2) - 1
  testthat::expect_equal(a1, a2)

  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::imbalance_steps(focal_tree),
                         treestats::imbalance_steps(ltab))

  # normalization
  a1 <- treestats::imbalance_steps(focal_tree,
                                   normalization = TRUE)
  testthat::expect_equal(a1, 1.0)
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    local_stat <- try_stat(focal_tree, treestats::imbalance_steps)
    testthat::expect_true(is.na(local_stat))
  }
})

test_that("abuse", {
  tree1 <- ape::rphylo(n = 2, birth = 1, death = 0)
  testthat::expect_message(
    treestats::imbalance_steps(tree1, normalization = TRUE)
  )

  testthat::expect_message(
    treestats::imbalance_steps(treestats::phylo_to_l(tree1),
                               normalization = TRUE)
  )
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::imbalance_steps(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::imbalance_steps(list()),
    "input object has to be phylo or ltable"
  )
})

context("double cherries")

test_that("usage", {
    set.seed(42)
    focal_tree <- ape::rphylo(n = 4, birth = 1, death = 0)
    bal_tree <- treestats::create_fully_balanced_tree(focal_tree)

    c1 <- treestats::double_cherries(bal_tree)
    testthat::expect_equal(c1, 1)

    ltab <- treestats::phylo_to_l(bal_tree)
    c2 <- treestats::double_cherries(ltab)
    testthat::expect_equal(c2, 1)

    focal_tree <- ape::rphylo(n = 8, birth = 1, death = 0)
    bal_tree <- treestats::create_fully_balanced_tree(focal_tree)

    c1 <- treestats::double_cherries(bal_tree)
    testthat::expect_equal(c1, 2)

    ltab <- treestats::phylo_to_l(bal_tree)
    c2 <- treestats::double_cherries(ltab)
    testthat::expect_equal(c2, 2)
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    res <- try_stat(focal_tree, treestats::double_cherries)
    testthat::expect_true(!is.na(res))
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::double_cherries(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::double_cherries(list()),
    "input object has to be phylo or ltable"
  )
})

#' Four prong index
#' @description Calculate the number of 4-tip caterpillars. The function does
#' accept non-binary trees, but four-prongs are only calculated on
#' non-polytomous sub-trees with 4 tips, e.g. the 4-tip part has to be a
#' caterpillar.
#' @param input_obj phylo object or ltable
#' @return number of 4-tip caterpillars
#' @references Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." PloS one 16.12 (2021): e0259877.
#'
#' Rosenberg, Noah A. "The mean and variance of the numbers of r-pronged nodes
#' and r-caterpillars in Yule-generated genealogical trees."
#' Annals of Combinatorics 10 (2006): 129-146.
#' @export
four_prong <- function(input_obj) {
  check_tree(input_obj,
             require_binary = FALSE,
             require_ultrametric = FALSE)

  if (inherits(input_obj, "matrix")) {
    return(calc_four_prong_ltable_cpp(input_obj))
  }
  if (inherits(input_obj, "phylo")) {
    return(calc_four_prong_cpp(as.vector(t(input_obj$edge))))
  }
  stop("input object has to be phylo or ltable")
}

#' Calculate all topology based statistics for a single tree
#' @param phylo phylo object
#' @param normalize if set to TRUE, results are normalized (if possible) under
#' either the  Yule expectation (if available), or the number of tips
#' @return list with statistics
#' @export
#' @description this function calculates all tree statistics based on topology
#' available in this package for a single tree, being:
#' \itemize{
#'   \item{area_per_pair}
#'   \item{average_leaf_depth}
#'   \item{avg_ladder}
#'   \item{avg_vert_depth}
#'   \item{b1}
#'   \item{b2}
#'   \item{beta}
#'   \item{blum}
#'   \item{cherries}
#'   \item{colless}
#'   \item{colless_corr}
#'   \item{colless_quad}
#'   \item{diameter}
#'   \item{double_cherries}
#'   \item{eigen_centrality}
#'   \item{ew_colless}
#'   \item{four_prong}
#'   \item{i_stat}
#'   \item{il_number}
#'   \item{imbalance_steps}
#'   \item{j_one}
#'   \item{max_betweenness}
#'   \item{max_closeness}
#'   \item{max_del_width}
#'   \item{max_depth}
#'   \item{max_ladder}
#'   \item{max_width}
#'   \item{mw_over_md}
#'   \item{pitchforks}
#'   \item{rogers}
#'   \item{root_imbalance}
#'   \item{rquartet}
#'   \item{sackin}
#'   \item{stairs}
#'   \item{stairs2}
#'   \item{symmetry_nodes}
#'   \item{tot_coph}
#'   \item{tot_internal_path}
#'   \item{tot_path_length}
#'   \item{var_depth}
#' }
#'
calc_topology_stats <- function(phylo, normalize = FALSE) {

  stats <- list()

  stats$sackin             <- try_stat(phylo, treestats::sackin,
                                       normalize, c("yule", "none"))
  stats$colless            <- try_stat(phylo, treestats::colless,
                                       normalize, c("yule", "none"))

  stats$colless_corr       <- try_stat(phylo, treestats::colless_corr)

  stats$colless_quad       <- try_stat(phylo, treestats::colless_quad)

  stats$beta               <- try_stat(phylo, treestats::beta_statistic)

  stats$sshape             <- try_stat(phylo, treestats::sshape,
                                       normalize, c(TRUE, FALSE))

  stats$avg_ladder         <- try_stat(phylo, treestats::avg_ladder)
  stats$max_ladder         <- try_stat(phylo, treestats::max_ladder)
  stats$cherries           <- try_stat(phylo, treestats::cherries,
                                       normalize, c("yule", "none"))

  stats$double_cherries    <- try_stat(phylo, treestats::double_cherries)

  stats$il_number          <- try_stat(phylo, treestats::ILnumber,
                                       normalize, c("tips", "none"))

  stats$pitchforks         <- try_stat(phylo, treestats::pitchforks,
                                       normalize, c("tips", "none"))

  stats$four_prong         <- try_stat(phylo, treestats::four_prong)

  stats$stairs             <- try_stat(phylo, treestats::stairs)

  stats$b1                 <- try_stat(phylo, treestats::b1,
                                       normalize, c("tips", "none"))

  stats$b2                 <- try_stat(phylo, treestats::b2,
                                       normalize, c("yule", "none"))

  stats$area_per_pair      <- try_stat(phylo, treestats::area_per_pair,
                                       normalize, c("yule", "none"))

  stats$average_leaf_depth <- try_stat(phylo, treestats::average_leaf_depth,
                                       normalize, c("yule", "none"))

  stats$i_stat             <- try_stat(phylo, treestats::mean_i)

  stats$ew_colless         <- try_stat(phylo, treestats::ew_colless)

  stats$max_del_width      <- try_stat(phylo, treestats::max_del_width,
                                       normalize, c("tips", "none"))

  stats$max_depth          <- try_stat(phylo, treestats::max_depth,
                                       normalize, c("tips", "none"))

  stats$max_width          <- try_stat(phylo, treestats::max_width,
                                       normalize, c("tips", "none"))

  stats$mw_over_md         <- try_stat(phylo, treestats::mw_over_md)

  stats$rogers             <- try_stat(phylo, treestats::rogers,
                                       normalize, c("tips", "none"))


  stats$stairs2            <- try_stat(phylo, treestats::stairs2)

  stats$tot_coph           <- try_stat(phylo, treestats::tot_coph,
                                       normalize, c("yule", "none"))

  stats$var_depth          <- try_stat(phylo, treestats::var_leaf_depth,
                                       normalize, c("yule", "none"))

  stats$symmetry_nodes     <- try_stat(phylo, treestats::sym_nodes,
                                       normalize, c("tips", "none"))

  stats$rquartet           <- try_stat(phylo, treestats::rquartet,
                                       normalize, c("yule", "none"))

  stats$imbalance_steps    <- try_stat(phylo, treestats::imbalance_steps,
                                       normalize)

  stats$j_one              <- try_stat(phylo, treestats::j_one)

  stats$diameter           <- try_stat(phylo, treestats::diameter)

  stats$avg_vert_depth     <- try_stat(phylo, treestats::avg_vert_depth)

  calc_local_eigen_centrality <- function(x) {
    res <- treestats::eigen_centrality(x, weight = FALSE, scale = FALSE)
    return(max(res$eigenvector))
  }

  stats$eigen_centrality   <- try_stat(phylo, calc_local_eigen_centrality)

  stats$max_betweenness    <- try_stat(phylo, treestats::max_betweenness)

  calc_max_closeness <- function(x, n) {
    return(treestats::max_closeness(x, weight = FALSE, normalization = n))
  }

  stats$max_closeness      <- try_stat(phylo, calc_max_closeness,
                                     normalize, c("tips", "none"))

  stats$root_imbalance     <- try_stat(phylo, treestats::root_imbalance)

  stats$tot_internal_path  <- try_stat(phylo, treestats::tot_internal_path)

  stats$tot_path_length    <- try_stat(phylo, treestats::tot_path_length)

  stats <- unlist(stats)
  stats <- stats[order(names(stats))]

  return(stats)
}

#' Stairs index
#' @description Calculates the staircase-ness measure, from the phyloTop
#' package. The staircase-ness reflects the number of subtrees that are
#' imbalanced, e.g. subtrees where the left child has more extant tips than the
#' right child, or vice versa.
#' @param input_obj phylo object or ltable
#' @return number of stairs
#' @references Norström, Melissa M., et al. "Phylotempo: a set of r scripts for
#' assessing and visualizing temporal clustering in genealogies inferred from
#' serially sampled viral sequences." Evolutionary Bioinformatics 8 (2012):
#' EBO-S9738.
#' @export
stairs <- function(input_obj) {
  check_tree(input_obj,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(input_obj, "matrix")) {
    return(stairs_ltable_cpp(input_obj))
  }
  if (inherits(input_obj, "phylo")) {
    return(stairs_cpp(as.vector(t(input_obj$edge))))
  }
  stop("input object has to be phylo or ltable")
}

#' Stairs2 index
#' @description Calculates the stairs2 measure, from the phyloTop package. The
#' stairs2 reflects the imbalance at each node, where it represents the average
#' across measure at each node, the measure being \eqn{min(L, R) / max(L, R)},
#' where L and R reflect the number of tips connected at the left (L) and right
#' (R) daughter.
#' @param input_obj phylo object or ltable
#' @return number of stairs
#' @references Norström, Melissa M., et al. "Phylotempo: a set of r scripts for
#' assessing and visualizing temporal clustering in genealogies inferred from
#' serially sampled viral sequences." Evolutionary Bioinformatics 8 (2012):
#' EBO-S9738.
#' @export
stairs2 <- function(input_obj) {
  check_tree(input_obj,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(input_obj, "matrix")) {
    return(stairs2_ltable_cpp(input_obj))
  }
  if (inherits(input_obj, "phylo")) {
    return(stairs2_cpp(as.vector(t(input_obj$edge))))
  }
  stop("input object has to be phylo or ltable")
}

context("stairs")

test_that("usage stairs1", {
  if (requireNamespace("phyloTop")) {
    test_func <- function(phy) {
      phyloTop::stairs(phy)[[1]]
    }

    standard_test(treestats::stairs, test_func)
  }
})

test_that("usage stairs2", {
  if (requireNamespace("phyloTop")) {

    test_func <- function(phy) {
      phyloTop::stairs(phy)[[2]]    # second index is stairs2
    }

    standard_test(treestats::stairs2, test_func)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::stairs(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::stairs(list()),
    "input object has to be phylo or ltable"
  )
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    a1 <- try_stat(focal_tree, treestats::stairs)
    testthat::expect_true(is.na(a1))
    a1 <- try_stat(focal_tree, treestats::stairs2)
    testthat::expect_true(is.na(a1))
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::stairs2(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::stairs2(list()),
    "input object has to be phylo or ltable"
  )
})

#' Wiener index
#' @description The Wiener index is defined as the sum of all shortest path
#' lengths between pairs of nodes in a tree.
#' @param phy phylo object or ltable
#' @param normalization if TRUE, the Wiener index is normalized by the number of
#' nodes, e.g. by choose(n, 2), where n is the number of nodes.
#' @param weight if TRUE, branch lenghts are used.
#' @return Wiener index
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." Plos one 16.12 (2021): e0259877.
#' Mohar, B., Pisanski, T. How to compute the Wiener index of a graph.
#' J Math Chem 2, 267–277 (1988)
#' @export
wiener <- function(phy, normalization = FALSE, weight = TRUE) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  normalization <- check_normalization_key(normalization)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {
    if (ape::is.rooted(phy)) {
      return(calc_wiener_cpp(as.vector(t(phy$edge)),
                             phy$edge.length,
                             normalization,
                             weight))
    } else {
      if (!weight) phy$edge.length <- rep(1, length(phy$edge.length))
      dist_n <- ape::dist.nodes(phy)
      return(sum(dist_n[lower.tri(dist_n)]))
    }
  }
  stop("input object has to be phylo or ltable")
}

context("wiener")

test_that("usage", {

  calc_using_ape <- function(phy) {
    av2 <- ape::dist.nodes(phy)
    return(sum(av2[lower.tri(av2)]))
  }

  standard_test(treestats::wiener, calc_using_ape)

  test_polytomies(treestats::wiener, calc_using_ape)


  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)

  a1_1 <- treestats::wiener(focal_tree, weight = TRUE)
  a1_2 <- treestats::wiener(focal_tree, weight = FALSE)
  a1_3 <- treestats::wiener(focal_tree, normalization = TRUE, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:
  a2_1 <- 120552.3 # treeCentrality::computeWienerIndex(focal_tree, #noLint
  #                                    weight = TRUE)  #noLint
  a2_2 <- 217028   # treeCentrality::computeWienerIndex(focal_tree,  #noLint
  #                                    weight = FALSE) #noLint
  a2_3 <- 11.01609 # treeCentrality::computeWienerIndex(focal_tree, #noLint
  #                                    norm = TRUE, #noLint
  #                                    weight = FALSE) #noLint

  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)
  testthat::expect_equal(a1_2, a2_2)
  testthat::expect_equal(a1_3, a2_3, tolerance = 0.01)

  testthat::expect_equal(a1_1, calc_using_ape(focal_tree))

  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::wiener(focal_tree, weight = TRUE),
                         treestats::wiener(ltab, weight = TRUE))

  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0.2, fossils = TRUE)

  a1_1 <- treestats::wiener(focal_tree, weight = TRUE)
  a1_2 <- treestats::wiener(focal_tree, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:
  a2_1 <- 216717.3 # treeCentrality::computeWienerIndex(focal_tree,
  #                                    weight = TRUE)
  a2_2 <- 362732   # treeCentrality::computeWienerIndex(focal_tree,
  #                                    weight = FALSE)

  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)
  testthat::expect_equal(a1_2, a2_2)

  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::wiener(focal_tree, weight = TRUE),
                         treestats::wiener(ltab,       weight = TRUE))
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::wiener(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::wiener(list()),
    "input object has to be phylo or ltable"
  )
})

#' B2 metric
#' @description Balance metric that uses the Shannon-Wiener statistic of
#' information content. The b2 measure is given by the sum over the depths of
#' all tips, divided by \eqn{2^{d}}, e.g. \eqn{\sum d_i / 2^{d_i}}
#' @param phy phylo object or ltable
#' @param normalization "none" or "yule", when "yule" is chosen, the statistic
#' is divided by the Yule expectation, following from theorem 3.7 in Bienvenu
#' 2020.
#' @return b2 statistic
#' @references  K.-T. Shao and R. R. Sokal. Tree Balance.
#' Systematic Zoology, 39(3):266, 1990. doi: 10.2307/2992186.
#'
#' Bienvenu, François, Gabriel Cardona, and Celine Scornavacca.
#' "Revisiting Shao and Sokal’s $$ B_2 $$ B 2 index of phylogenetic balance."
#' Journal of Mathematical Biology 83.5 (2021): 1-43.
#' @export
b2 <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    b2_stat <- calc_b2_ltable_cpp(phy)
    if (normalization == "yule" || normalization == TRUE) {
      n <- length(phy[, 1])
      expectation <- sum(1 / (1:n))
      b2_stat <- b2_stat / expectation
    }
    return(b2_stat)
  }
  if (inherits(phy, "phylo")) {
    b2_stat <- calc_b2_cpp(as.vector(t(phy$edge)))
    if (normalization == "yule" || normalization == TRUE) {
      n <- length(phy$tip.label)
      expectation <- sum(1 / (1:n))
      b2_stat <- b2_stat / expectation
    }
    return(b2_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' Height of a tree.
#' @description In a reconstructed tree, obtaining the tree height is fairly
#' straightforward, and the function
#' \link[beautier:get_crown_age]{beautier::get_crown_age} does a great
#' job at it. However, in a non-ultrametric tree, that function no longer works.
#' Alternatively, taking the maximum value of
#' \link[adephylo:distRoot]{adephylo::distRoot} will
#' also yield the tree height (including the root branch), but will typically
#' perform many superfluous calculations and thus be slow.
#' @param phy phylo object
#' @return crown age
#' @export
tree_height <- function(phy) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(phy[1, 1])
  }

  if (inherits(phy, "phylo")) {
    crown_age_of_tree <- crown_age(phy)
    root_length <- 0
    if (!is.null(phy$root.edge)) {
      root_length <- phy$root.edge
    }
    tree_height <- crown_age_of_tree + root_length
    return(tree_height)
  }
  stop("input object has to be of class phylo")
}

#' Crown age of a tree.
#' @description In a reconstructed tree, obtaining the crown age is
#' fairly straightforward, and the function
#' \link[beautier:get_crown_age]{beautier::get_crown_age}  does
#' a great job at it. However, in a non-ultrametric tree, that function no
#' longer works. This function provides a functioning alternative.
#' @param phy phylo object or ltable
#' @return crown age
#' @export
crown_age <- function(phy) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(phy[1, 1])
  }

  if (inherits(phy, "phylo")) {
    return(calc_crown_age_cpp(phy))
  }
  stop("input object has to be phylo or ltable")
}

context("crown_age")

test_that("usage", {
  if (requireNamespace("adephylo")) {
    age1 <- treestats::tree_height(extant_tree)
    age2 <- max(adephylo::distRoot(extant_tree))
    testthat::expect_equal(age1, age2)

    # now with ltable:
    age3 <- treestats::tree_height(treestats::phylo_to_l(extant_tree))
    testthat::expect_equal(age3, age2)


    # now, more challenging, with extinct lineages:
    age1 <- treestats::tree_height(complete_tree)
    age2 <- max(adephylo::distRoot(complete_tree))
    testthat::expect_equal(age1, age2)

    # and now with crown age:
    focal_tree <- ape::rbdtree(birth = 1, death = 0.3, Tmax = 3)

    age1 <- treestats::crown_age(focal_tree)
    testthat::expect_equal(age1, 3)

    # and now without root.edge
    focal_tree <- ape::rphylo(100, 1, 0)
    age1 <- treestats::crown_age(focal_tree)
    age2 <- max(adephylo::distRoot(focal_tree))
    testthat::expect_equal(age1, age2)

    age3 <- treestats::crown_age(treestats::phylo_to_l(focal_tree))
    testthat::expect_equal(age3, age1)
  }
})

test_that("polytomies", {
  if (requireNamespace("adephylo")) {
    test_func <- function(phy) {
        max(adephylo::distRoot(phy))
    }
    testthat::expect_warning(
      # these are warnings from adephylo.
      # adding this testthat catcher inside the test_func
      # breaks the test.
      test_polytomies(treestats::crown_age, test_func)
    )
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::crown_age(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::crown_age(list()),
    "input object has to be phylo or ltable"
  )
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::tree_height(10),
    "input object has to be of class phylo"
  )

  testthat::expect_error(
    treestats::tree_height(list()),
    "input object has to be of class phylo"
  )
})

#' Treeness statistic
#' @description Calculates the fraction of tree length on internal branches,
#' also known as treeness or stemminess
#' @param phy phylo object or Ltable
#' @return sum of all internal branch lengths (e.g. branches not leading
#' to a tip) divided by the sum over all branch lengths.
#' @export
treeness <- function(phy) {


  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    # in an unrooted tree, this number still indicates the first non-tip
    root_no <- min(phy$edge[, 1])

    int_branch_length <- sum(phy$edge.length[phy$edge[, 2] >= root_no])

    return(int_branch_length / (sum(phy$edge.length)))
  }

  stop("input object has to be phylo or ltable")
}

#' @keywords internal
integr <- function(x, f) {
  # get lengths of var and integrand
  n <- length(x)

  # trapezoidal integration
  integral <- 0.5 * sum((x[2:n] - x[1:(n - 1)]) * (f[2:n] + f[1:(n - 1)]))

  # print definite integral
  return(integral)
}


#' @keywords internal
skew_ness <- function(x) {
  n <- length(x)
  return((sum((x - mean(x)) ^ 3) / n) / (sum((x - mean(x)) ^ 2) / n) ^ (3 / 2))
}

#' Laplacian spectrum statistics, from RPANDA
#' @description Computes the distribution of eigenvalues for the modified graph
#' Laplacian of a phylogenetic tree, and several summary statistics of this
#' distribution. The modified graph Laplacian of a phylogeny is given by the
#' difference between its' distance matrix (e.g. all pairwise distances between
#' all nodes), and the degree matrix (e.g. the diagonal matrix where each
#' diagonal element represents the sum of branch lengths to all other nodes).
#' Each row of the modified graph Laplacian sums to zero. For a tree with n
#' tips, there are at most \eqn{N = 2n-1} nodes, and hence the modified graph
#' Laplacian is represented by a \eqn{N x N} matrix. Where RPANDA relies on the
#' package igraph to calculate the modified graph Laplacian, the treestats
#' package uses C++ to directly calculate the different entries in the matrix.
#' This makes the treestats implementation slightly faster, although the bulk of
#' computation occurs in estimating the eigen values, where RcppArmadillo
#' generates a bit of speed up.
#' @param phy phy
#' @return list with five components: 1) eigenvalues the vector of eigen
#' values, 2) principal_eigenvalue the largest eigenvalue of the spectral
#' density distribution 3) asymmetry the skewness of the spectral density
#' distribution 4) peak_height the largest y-axis value of the spectral
#' density distribution and 5) eigengap the position of the largest
#' difference between eigenvalues, giving the number of modalities in the tree.
#' @references Eric Lewitus, Helene Morlon, Characterizing and Comparing
#' Phylogenies from their Laplacian Spectrum, Systematic Biology, Volume 65,
#' Issue 3, May 2016, Pages 495–507, https://doi.org/10.1093/sysbio/syv116
#' @export
laplacian_spectrum <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (!inherits(phy, "phylo")) {
    stop("input object has to be phylo or ltable")
  }
  n <- length(phy$tip.label)
  m <- n - 1
  nm <- n + m
  if (nm > 46340) { # sqrt(2^31 - 1) #nolint
    stop("tree too big")
  }

  dens_rpanda <- function(x,
                          bw = stats::bw.nrd0,
                          n = 4096,
                          from = min(x) - 3 * sd, to = max(x) + 3 * sd,
                          adjust = 1,
                          ...) {
    if (has_na <- any(is.na(x))) {
      x <- stats::na.omit(x)
      if (length(x) == 0)
        stop("no finite or non-missing data!")
    }
    sd <- (if (is.numeric(bw))
      bw[1]
      else bw(x)) * adjust
    xx <- seq(from, to, len = n)

    mat <-  outer_cpp(xx, x, sd)

    structure(list(x = xx,
                   y = rowMeans(mat),
                   bw = sd,
                   call = match.call(),
                   n = length(x),
                   data.name = deparse(substitute(x)),
                   has.na = has_na),
              class = "density")
  }

  phy <- ape::reorder.phylo(phy)

  x <- get_eigen_values_arma_cpp(phy)
  x <- sort(x, decreasing = TRUE)

  d <- dens_rpanda(log(x))
  dsc <- d$y / (integr(d$x, d$y))
  principal_eigenvalue <- max(x)
  skewness <- skew_ness(x)
  peak_height <- max(dsc)
  gaps <- abs(diff(x))
  gap_mat <- as.matrix(gaps)  # nolint
  modalities <- seq_along(gap_mat)
  gap_mat_col <- cbind(modalities, gap_mat)  # nolint
  eigen_gap <- subset(gap_mat_col, gap_mat_col[, 2] == max(gap_mat_col[, 2]))
  res <- list(eigenvalues = x,
              principal_eigenvalue = principal_eigenvalue,
              asymmetry = skewness,
              peakedness = peak_height,
              eigengap = eigen_gap[, 1])
  return(res)
}

context("rpanda")

test_that("usage", {
  testthat::skip_on_cran() # RPANDA uses iGraph for the Laplacian spectrum
                           # this can cause BLAS issues
                           # also, RPANDA is no longer on CRAN

  if (requireNamespace("RPANDA")) {
    set.seed(42)
    focal_tree <- ape::rphylo(n = 5, birth = 1, death = 0)

    ref <- RPANDA::spectR(focal_tree)
    stat <- treestats::laplacian_spectrum(focal_tree)

    diff_eig <- sum(ref$eigenvalues - stat$eigenvalues)
    testthat::expect_equal(0, diff_eig)
    testthat::expect_equal(ref$asymmetry, stat$asymmetry)
    testthat::expect_equal(ref$peakedness, stat$peakedness, tolerance = 0.01)
    testthat::expect_equal(ref$principal_eigenvalue, stat$principal_eigenvalue)
    testthat::expect_equal(ref$eigengap, stat$eigengap)


    stat2 <- treestats::laplacian_spectrum(
      phy = treestats::phylo_to_l(focal_tree)
    )
    testthat::expect_true(all.equal(stat, stat2))
  }
})

test_that("polytomies", {
  if (requireNamespace("RPANDA")) {
    for (focal_tree in poly_trees) {
      stat <- treestats::laplacian_spectrum(focal_tree)
      ref <- RPANDA::spectR(focal_tree)

      diff_eig <- sum(ref$eigenvalues - stat$eigenvalues)
      testthat::expect_equal(0, diff_eig)
      testthat::expect_equal(ref$asymmetry, stat$asymmetry)
      testthat::expect_equal(ref$peakedness, stat$peakedness, tolerance = 0.01)
      testthat::expect_equal(ref$principal_eigenvalue,
                             stat$principal_eigenvalue)
      testthat::expect_equal(ref$eigengap, stat$eigengap)
    }
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::laplacian_spectrum(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::laplacian_spectrum(list()),
    "input object has to be phylo or ltable"
  )
})

#' Aldous' beta statistic.
#' @description The Beta statistic fits a beta splitting model to each node,
#' assuming that the number of extant descendants of each daughter branch is
#' split following a beta distribution, such that the number of extant
#' descendants x and y at a node follows \eqn{q(x, y) = s_n(\beta)^{-1}
#' \frac{(\gamma(x + 1 + \beta)\gamma(y + 1 + \beta))}{\gamma(x+1)\gamma(y+1)}},
#' where \eqn{s_n(\beta)^{-1}} is a normalizing constant. When this model is fit
#' to a tree,
#' different values of beta correspond to the expectation following from
#' different diversification models, such that a beta of 0 corresponds to a
#' Yule tree, a beta of \eqn{-3/2} to a tree following from a PDA model.
#' In general,
#' negative beta values correspond to trees more unbalanced than Yule trees, and
#' beta values larger than zero indicate trees more balanced than Yule trees.
#' The lower bound of the beta splitting parameter is -2.
#' @param phy phylogeny or ltable
#' @param upper_lim Upper limit for beta parameter, default = 10.
#' @param algorithm optimization algorithm used, default is "COBYLA"
#' (Constrained Optimization BY Linear Approximations), also available are
#' "subplex" and "simplex". Subplex and Simplex seem to have difficulties with
#' unbalanced trees, e.g. if beta < 0.
#' @param abs_tol absolute stopping criterion of optimization. Default is 1e-4.
#' @param rel_tol relative stopping criterion of optimization. Default is 1e-6.
#' @return Beta value
#' @references Aldous, David. "Probability distributions on cladograms." Random
#' discrete structures. Springer, New York, NY, 1996. 1-18.
#' Jones, Graham R. "Tree models for macroevolution and phylogenetic analysis."
#' Systematic biology 60.6 (2011): 735-746.
#' @export
#' @examples

#' simulated_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' beta_statistic(balanced_tree) # should be approximately 10
#' beta_statistic(simulated_tree) # should be near 0
#' beta_statistic(unbalanced_tree) # should be approximately -2
beta_statistic <- function(phy,
                           upper_lim = 10,
                           algorithm = "COBYLA",
                           abs_tol = 1e-4,
                           rel_tol = 1e-6) {
  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    beta_stat <- calc_beta_ltable_cpp(phy, upper_lim,
                                      algorithm, abs_tol, rel_tol)
    return(beta_stat)
  }
  if (inherits(phy, "phylo")) {
    beta_stat <- calc_beta_cpp(phy, upper_lim,
                               algorithm, abs_tol, rel_tol)
    return(beta_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' Mean I statistic.
#' @description The mean I value is defined for all nodes with at least 4 tips
#' connected, such that different topologies can be formed. Then, for each node,
#' \eqn{I = (nm - nt/2) / (nt - 1 - nt/2)}, where nt is the total number of tips
#' descending from that node, nm is the daughter branch leading to most tips,
#' and \eqn{nt/2} is the minimum size of the maximum branch, rounded up.
#' Following Purvis et al 2002, we perform a correction on I, where we correct
#' I for odd nt, such that \eqn{I' = I * (nt - 1) / nt}. This correction ensures
#' that I is independent of nt. We report the mean value across all I' (again,
#' following Purvis et al. 2002).
#' @param phy phylo object or ltable
#' @return average I value across all nodes
#' @references  G. Fusco and Q. C. Cronk. A new method for evaluating the shape
#' of large phylogenies. Journal of Theoretical Biology, 1995.
#' doi: 10.1006/jtbi.1995.0136.
#'
#' A. Purvis, A. Katzourakis, and P.-M. Agapow. Evaluating Phylogenetic Tree
#' Shape: Two Modifications to Fusco & Cronks Method. Journal of Theoretical
#' Biology, 2002. doi: 10.1006/jtbi.2001.2443.
#' @export
mean_i <- function(phy) {

  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    if (length(phy[, 1]) < 4) {
      warning("I statistic is only available for trees with at least 4 tips.")
      return(NA)
    }
    return(calc_Ibased_ltable_cpp(phy))
  }
  if (inherits(phy, "phylo")) {

    if (length(phy$tip.label) < 4) {
      warning("I statistic is only available for trees with at least 4 tips.")
      return(NA)
    }

    return(calc_Ibased_cpp(as.vector(t(phy$edge))))
  }
  stop("input object has to be phylo or ltable")
}


#' Mean I statistic.
#' @description The mean I value is defined for all nodes with at least 4 tips
#' connected, such that different topologies can be formed. Then, for each node,
#' \eqn{I = (nm - nt/2) / (nt - 1 - nt/2)}, where nt is the total number of tips
#' descending from that node, nm is the daughter branch leading to most tips,
#' and \eqn{nt/2} is the minimum size of the maximum branch, rounded up.
#' Following Purvis et al 2002, we perform a correction on I, where we correct
#' I for odd nt, such that \eqn{I' = I * (nt - 1) / nt}. This correction ensures
#' that I is independent of nt. We report the mean value across all I' (again,
#' following Purvis et al. 2002).
#' @param phy phylo object or ltable
#' @return average I value across all nodes
#' @references  G. Fusco and Q. C. Cronk. A new method for evaluating the shape
#' of large phylogenies. Journal of Theoretical Biology, 1995.
#' doi: 10.1006/jtbi.1995.0136.
#'
#' A. Purvis, A. Katzourakis, and P.-M. Agapow. Evaluating Phylogenetic Tree
#' Shape: Two Modifications to Fusco & Cronks Method. Journal of Theoretical
#' Biology, 2002. doi: 10.1006/jtbi.2001.2443.
#' @export
i_stat <- function(phy) {
  return(mean_i(phy))
}

#' Rogers J index of (im)balance.
#' @description The Rogers index is calculated as the total number of internal
#' nodes that are unbalanced, e.g. for which both daughter nodes lead to a
#' different number of extant tips. in other words, the number of nodes where
#' \eqn{L  !=  R} (where L(R) is the number of extant tips of the Left (Right)
#' daughter node).
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", in which case the resulting statistic
#' is divided by the number of tips - 2 (e.g. the maximum value of the rogers
#' index for a tree).
#' @return Rogers index
#' @references  J. S. Rogers. Central Moments and Probability Distributions of
#' Three Measures of Phylogenetic Tree Imbalance. Systematic Biology,
#' 45(1):99-110, 1996. doi: 10.1093/sysbio/45.1.99.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' rogers(balanced_tree)
#' rogers(unbalanced_tree) # should be higher
rogers <- function(phy, normalization = "none") {

  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  normalization <- check_normalization_key(normalization)

  if (inherits(phy, "matrix")) {
    rogers_stat <- calc_rogers_ltable_cpp(phy)
    if (normalization == "tips" || normalization == TRUE) {
      rogers_stat <- rogers_stat / (length(phy[, 1]) - 2)
    }
    return(rogers_stat)
  }
  if (inherits(phy, "phylo")) {
    rogers_stat <- calc_rogers_cpp(as.vector(t(phy$edge)))
    if (normalization == "tips" || normalization == TRUE) {
      rogers_stat <- rogers_stat / (length(phy$tip.label) - 2)
    }
    return(rogers_stat)
  }
  stop("input object has to be phylo or ltable")
}

context("rogers")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    standard_test(treestats::rogers, treebalance::rogersI)

    test_polytomies(treestats::rogers, treebalance::rogersI)
  }
})


test_that("normalization", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 30, birth = 1, death = 0)

  c1 <- treestats::rogers(focal_tree)
  c2 <- treestats::rogers(focal_tree, normalization = "tips")
  testthat::expect_lt(c2, c1)
  c3 <- treestats::rogers(treestats::phylo_to_l(focal_tree),
                          normalization = "tips")
  testthat::expect_equal(c2, c3)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::rogers(focal_tree))
    stats2 <- c(stats2, treestats::rogers(focal_tree, normalization = "tips"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_lt(a2, 0.2)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::rogers(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::rogers(list()),
    "input object has to be phylo or ltable"
  )
})

#' Average vertex depth metric
#' @description The average vertex depth metric, measures the average path
#' (in edges), between the tips and the root.
#' @param phy phylo object or ltable
#' @return Average depth (in number of edges)
#' @references  C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
#' transmission patterns. Evolution, Medicine, and Public Health,
#' 2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
#' @export
avg_vert_depth <- function(phy) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)


  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {
    # polytomies are checked at the Rcpp side
    max_d_stat <-  calc_avg_vert_depth_cpp(as.vector(t(phy$edge)))
    return(max_d_stat)
  }
  stop("input object has to be phylo or ltable")
}

context("avg vertex depth")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    a1 <- treestats::avg_vert_depth(extant_tree)
    a2 <- treebalance::avgVertDep(extant_tree)
    testthat::expect_equal(a1, a2)

    ltab <- treestats::phylo_to_l(extant_tree)
    testthat::expect_equal(treestats::avg_vert_depth(extant_tree),
                           treestats::avg_vert_depth(ltab))

    a1 <- treestats::avg_vert_depth(complete_tree)
    a2 <- treebalance::avgVertDep(complete_tree)
    testthat::expect_equal(a1, a2)

    ltab <- treestats::phylo_to_l(complete_tree)
    testthat::expect_equal(treestats::avg_vert_depth(complete_tree),
                           treestats::avg_vert_depth(ltab))
  }
})

test_that("polytomies", {
  if (requireNamespace("treebalance")) {
    test_polytomies(treestats::avg_vert_depth, treebalance::avgVertDep)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::avg_vert_depth(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::avg_vert_depth(list()),
    "input object has to be phylo or ltable"
  )
})

#' Mean branch length of a tree
#' @description Mean branch length of a tree, including extinct branches.
#' @param phy phylo object or Ltable
#' @return mean branch length
#' @export
mean_branch_length <- function(phy) {

  if (inherits(phy, "matrix")) {
    return(calc_mean_branch_length_ltable_cpp(phy))
  }
  if (inherits(phy, "phylo")) {
    return(calc_mean_branch_length_cpp(as.vector(phy$edge.length)))
  }

  stop("input object has to be phylo or ltable")
}

#' Variance of branch lengths of a tree
#' @description Variance of branch lengths of a tree,
#' including extinct branches.
#' @param phy phylo object or Ltable
#' @return variance of branch lengths
#' @export
var_branch_length <- function(phy) {
  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    return(stats::var(phy$edge.length))
  }

  stop("input object has to be phylo or ltable")
}

#' Variance of internal branch lengths of a tree
#' @description Variance of internal branch lengths of a tree,
#' e.g. of branches not leading to a tip.
#' @param phy phylo object or Ltable
#' @return variance of internal branch lengths
#' @export
var_branch_length_int <- function(phy) {
  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    root_no <- min(phy$edge[, 1])

    int_branches <- phy$edge.length[phy$edge[, 2] >= root_no]

    return(stats::var(int_branches))
  }

  stop("input object has to be phylo or ltable")
}

#' Mean length of internal branches of a tree
#' @description Mean length of internal branches of a tree, e.g. of
#' branches not leading to a tip.
#' @param phy phylo object or Ltable
#' @return mean of internal branch lengths
#' @export
mean_branch_length_int <- function(phy) {
  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    root_no <- min(phy$edge[, 1])

    int_branches <- phy$edge.length[phy$edge[, 2] >= root_no]

    return(mean(int_branches))
  }

  stop("input object has to be phylo or ltable")
}

#' Variance of external branch lengths of a tree
#' @description Variance of external branch lengths of a tree, e.g. of
#' branches leading to a tip.
#' @param phy phylo object or Ltable
#' @return variance of external branch lengths
#' @export
var_branch_length_ext <- function(phy) {
  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    root_no <- min(phy$edge[, 1])

    ext_branches <- phy$edge.length[phy$edge[, 2] < root_no]

    return(stats::var(ext_branches))
  }

  stop("input object has to be phylo or ltable")
}

#' Mean length of external branch lengths of a tree
#' @description Mean length of external branch lengths of a tree, e.g. of
#' branches leading to a tip.
#' @param phy phylo object or Ltable
#' @return mean of external branch lengths
#' @export
mean_branch_length_ext <- function(phy) {
  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    root_no <- min(phy$edge[, 1])

    ext_branches <- phy$edge.length[phy$edge[, 2] < root_no]

    return(mean(ext_branches))
  }

  stop("input object has to be phylo or ltable")
}

context("branch_length_stats")

test_that("usage", {
  focal_tree <- ape::read.tree(text = "((t2:1,t1:1):2,(t4:2,t3:2):1);")

  m_br <- treestats::mean_branch_length(focal_tree)
  testthat::expect_equal(m_br, 1.5)

  v_br <- treestats::var_branch_length(focal_tree)
  testthat::expect_equal(v_br, 0.3)

  m_br_ext <- treestats::mean_branch_length_ext(focal_tree)
  testthat::expect_equal(m_br_ext, 1.5)

  m_br_int <- treestats::mean_branch_length_int(focal_tree)
  testthat::expect_equal(m_br_ext, 1.5)

  v_br_ext <-   treestats::var_branch_length_ext(focal_tree)
  testthat::expect_equal(v_br_ext, 1 / 3)

  v_br_int <-   treestats::var_branch_length_int(focal_tree)
  testthat::expect_equal(v_br_int, 1 / 2)


  ltab <- treestats::phylo_to_l(focal_tree)

  m_br <- treestats::mean_branch_length(ltab)
  testthat::expect_equal(m_br, 1.5)

  v_br <- treestats::var_branch_length(ltab)
  testthat::expect_equal(v_br, 0.3)

  m_br_ext <- treestats::mean_branch_length_ext(ltab)
  testthat::expect_equal(m_br_ext, 1.5)

  m_br_int <- treestats::mean_branch_length_int(ltab)
  testthat::expect_equal(m_br_ext, 1.5)

  v_br_ext <-   treestats::var_branch_length_ext(ltab)
  testthat::expect_equal(v_br_ext, 1 / 3)

  v_br_int <-   treestats::var_branch_length_int(ltab)
  testthat::expect_equal(v_br_int, 1 / 2)
})

test_that("unrooted", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
  a1 <- treestats::mean_branch_length_ext(focal_tree)
  a2 <- treestats::mean_branch_length_ext(ape::unroot(focal_tree))
  testthat::expect_equal(a1, a2)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::var_branch_length(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::var_branch_length(list()),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::var_branch_length_int(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::var_branch_length_int(list()),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::var_branch_length_ext(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::var_branch_length_ext(list()),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_branch_length_ext(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_branch_length_ext(list()),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_branch_length_int(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_branch_length_int(list()),
    "input object has to be phylo or ltable"
  )
})

context("mean_branch_length")

test_that("usage", {
  calc_mean_br_r <- function(focal_tree) {
    return(mean(focal_tree$edge.length, na.rm = TRUE))
  }

  focal_tree <- ape::read.tree(text = "(1:4,2:4):0;")

  a1 <- treestats::mean_branch_length(focal_tree)
  a2 <- calc_mean_br_r(focal_tree)
  testthat::expect_equal(a1, a2)

  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::mean_branch_length(focal_tree),
                         treestats::mean_branch_length(ltab))

  a1 <- treestats::mean_branch_length(extant_tree)
  a2 <- calc_mean_br_r(extant_tree)
  testthat::expect_equal(a1, a2)

  ltab <- treestats::phylo_to_l(extant_tree)
  testthat::expect_equal(a2,
                         treestats::mean_branch_length(ltab))


  # with extinct species:
  a1 <- treestats::mean_branch_length(complete_tree)
  a2 <- calc_mean_br_r(complete_tree)
  testthat::expect_equal(a1, a2)

  ltab <- treestats::phylo_to_l(complete_tree)
  testthat::expect_equal(a2,
                         treestats::mean_branch_length(ltab))
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::mean_branch_length(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mean_branch_length(list()),
    "input object has to be phylo or ltable"
  )
})

#' Laplacian Matrix properties
#' @description Calculates the eigenvalues of the Laplacian Matrix, where the
#' Laplacian matrix is the matrix representation of a graph, in this case a
#' phylogeny.
#' When the R package \pkg{RSpectra} is available, a faster
#' calculation can be used, which does not calculate all eigenvalues, but only
#' the maximum and minimum. As such, when using this option, the vector of all
#' eigenvalues is not returned.
#' @param phy phylo object or ltable
#' @param use_rspectra boolean to indicate whether the helping package RSpectra
#' should be used, in which case only the minimum and maximum values are
#' returned, and not the full list of eigenvalues, which makes the computation
#' much faster.
#' @param force_min boolean to indicate whether or not to also calculate
#' the minimum positive eigenvalue if the tree is larger than 23170 tips.
#' Default is FALSE, as this can be very slow, but in some cases might be
#' of interest.
#' @return List with the minimum and maximum eigenvalues
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." Plos one 16.12 (2021): e0259877.
#' @export
minmax_laplace <- function(phy,
                           use_rspectra = TRUE,
                           force_min = FALSE) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    if (sum(phy[, 4] > -1))
      warning("minmax_laplace may give incorrect results for the minimal value due to conversion to a phylo object, when extinct lineages are present")  # nolint
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {

    if (!use_rspectra || !requireNamespace("RSpectra")) {
      return(slow_calc_minmax_laplace(phy))
    }

    edges   <- phy$edge
    lengths <- phy$edge.length
    n_nodes <- max(edges)

    matvec_fun <- function(x, args) {
      Lx_tree_weighted(edges, lengths, x, n_nodes)
    }

    mat_size <- max(phy$edge)
    max_val <- RSpectra::eigs_sym(matvec_fun,
                                  n = mat_size,
                                  k = 1,
                                  which = "LA",
                                  opts = list(retvec = FALSE))$values

    if (mat_size > 46340) {  # floor(sqrt(2^31 - 1)))
      if (force_min == TRUE) {
        warning("calculating the minimum eigenvalue
                   is very costly for large trees")
        res <- RSpectra::eigs_sym(matvec_fun,
                                  n = n_nodes,
                                  # we don't need negative eigenvalues:
                                  k = n_nodes / 2,
                                  which = "SM",
                                  opts = list(retvec = FALSE))$values
        res <- res[round(res, 10) > 0]
        res_min <- min(res)
        return(list("max" = max_val,
                    "min" = res_min))
      } else {
        warning("skipped calculating the minimum eigenvalue of the adjacency
                  matrix, because the tree is too large")
        return(list("max" = max_val))
      }
    } else {
      # if the tree is not very large, then using the matrix is faster
      lap_mat <- matrix(0, mat_size, mat_size)
      lap_mat[phy$edge] <- phy$edge.length
      lap_mat <- lap_mat + t(lap_mat)
      degrees <- rowSums(lap_mat)
      lap_mat <- diag(degrees) - lap_mat

      # we can make use of RSpectra
      min_val <- RSpectra::eigs_sym(lap_mat,
                                    k = 1,
                                    which = "BE",
                                    sigma = 1e-9,
                                    opts = list(retvec = FALSE))$values

      return(list(min = min_val,
                  max = max_val))
    }
  }
  stop("input object has to be phylo or ltable")
}


#' @keywords internal
slow_calc_minmax_laplace <- function(phy) {

  mat_size <- max(phy$edge)
  if (mat_size > 46340) {  # floor(sqrt(2^31 - 1)))
    stop("tree too big, memory allocation fail")
  }

  lap_mat <- matrix(0, mat_size, mat_size)
  lap_mat[phy$edge] <- ifelse(rep(TRUE, mat_size - 1), # weight is always true
                              phy$edge.length, 1)

  lap_mat <- lap_mat + t(lap_mat)
  degrees <- rowSums(lap_mat)
  lap_mat <- diag(degrees) - lap_mat
  eigen_vals <- eigen(lap_mat,
                      symmetric = TRUE,
                      only.values = TRUE)$values

  eigen_vals <- round(eigen_vals, digits = 10)

  max_val <- max(eigen_vals)
  min_val <- min(eigen_vals[eigen_vals > 0])
  return(list("max" = max_val,
              "min" = min_val,
              "values" = eigen_vals))
}

context("minmax_laplace")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 5, birth = 1, death = 0)

  a1_1 <- treestats::minmax_laplace(focal_tree)

  # because treeCentrality is not available on CRAN, we precompute reference
  # values:
  # ref <-
  #    treeCentrality::computeSpectrum(focal_tree, weight = TRUE, dist = FALSE,
  #                                    full = TRUE, lap = TRUE, norm = FALSE)


  ref <- c(5.684185e-01, 4.343028e-01, 1.754782e-01, 1.248247e-01, 7.433148e-02,
           3.966736e-02, 3.384060e-02, 4.262393e-03, 3.699463e-17)
  ref <- round(ref, digits = 10)
  testthat::expect_equal(a1_1$min, min(ref[ref > 0]), tolerance = 0.01)

  testthat::expect_equal(a1_1$max, max(ref), tolerance = 0.01)

  ltab <- treestats::phylo_to_l(focal_tree)

  testthat::expect_equal(
    treestats::minmax_laplace(focal_tree),
    treestats::minmax_laplace(ltab))

  focal_tree <- ape::rphylo(n = 5, birth = 1, death = 0.2, fossils = TRUE)

  ref <- c(3.609466e+00,  2.997761e+00,  2.771082e+00,  1.951827e+00,
           1.704470e+00,  1.384831e+00,  9.340857e-01,  7.710821e-01,
           7.093589e-01,  3.099265e-01,  2.548943e-01,  1.853926e-01,
           1.459424e-01,  3.611505e-02,  1.862803e-02,  6.809262e-03,
           -1.202488e-16)
  ref <- round(ref, digits = 10)
  # ref <-
  #     treeCentrality::computeSpectrum(focal_tree, weight = TRUE, dist = FALSE,
  #                                      full = TRUE, lap = TRUE, norm = FALSE)
  a1_1 <- treestats::minmax_laplace(focal_tree)

  testthat::expect_equal(a1_1$min, min(ref[ref > 0]), tolerance = 0.001)

  testthat::expect_equal(a1_1$max, max(ref), tolerance = 0.001)

  ltab <- treestats::phylo_to_l(focal_tree)

  testthat::expect_warning(
    a2_1 <- treestats::minmax_laplace(ltab)
  )

  testthat::expect_equal(a2_1$min, min(ref[ref > 0]))
  testthat::expect_equal(a2_1$max, max(ref), tolerance = 0.001)

  # test rspectra use
  ref <- treestats::minmax_laplace(focal_tree,
                                   use_rspectra = FALSE)

  a2 <- treestats::minmax_laplace(focal_tree,
                                  use_rspectra = TRUE)
  testthat::expect_equal(ref$min, a2$min)
  testthat::expect_equal(ref$max, a2$max)
})

test_that("igraph", {
  if (requireNamespace("igraph")) {
    set.seed(42)
    focal_tree <- ape::rphylo(n = 10, birth = 1, death = 0)

    a1_1 <- treestats::minmax_laplace(focal_tree)

    df <- as.data.frame(cbind(focal_tree$edge,
                              weight = focal_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)

    lapl_mat <- igraph::laplacian_matrix(g,
                                         normalization = "unnormalized",
                                         sparse = FALSE)
    ref <- eigen(lapl_mat)$values
    ref <- round(ref, digits = 10)

    testthat::expect_equal(a1_1$min, min(ref[ref > 1e-5]))
    testthat::expect_equal(a1_1$max, max(ref))
  }
})

test_that("polytomies", {
  if (requireNamespace("igraph")) {

    test_func <- function(phy) {
      df <- as.data.frame(cbind(phy$edge,
                                weight = phy$edge.length))
      g <- igraph::graph_from_data_frame(df, directed = FALSE)

      lapl_mat <- igraph::laplacian_matrix(g,
                                           normalization = "unnormalized",
                                           sparse = FALSE)
      ref <- eigen(lapl_mat)$values
      ref <- round(ref, digits = 10)
      return(ref)
    }

    min_ref <- function(phy) {
      ref <- test_func(phy)
      return(min(ref[ref > 1e-5]))
    }
    max_ref <- function(phy) {
      return(max(test_func(phy)))
    }

    min_treestats <- function(phy) {
      res <- treestats::minmax_laplace(phy)
      return(res$min)
    }
    max_treestats <- function(phy) {
      res <- treestats::minmax_laplace(phy)
      return(res$max)
    }


    test_polytomies(min_treestats, min_ref, tol = 1e-5)
    test_polytomies(max_treestats, max_ref)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::minmax_laplace(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::minmax_laplace(list()),
    "input object has to be phylo or ltable"
  )
})

#' Variance of all pairwise distances.
#' @description After calculating all pairwise distances between all tips,
#' this function takes the variance across these values.
#' @param phy phylo object or ltable
#' @return Variance in pairwise distance
#' @references  Webb, C., D. Ackerly, M. McPeek, and M. Donoghue. 2002.
#' Phylogenies and community ecology. Annual Review of Ecology and
#' Systematics 33:475-505.
#' @export
var_pair_dist <- function(phy) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {

    if (ape::is.rooted(phy) && check_binary(phy)) {
       return(calc_var_mpd_cpp(phy))
    } else {
      dist_mat <- ape::cophenetic.phylo(phy)
      dist_mat <- dist_mat[lower.tri(dist_mat)]
      n <- length(dist_mat)
      var_mpd <- stats::var(dist_mat, na.rm = TRUE, use = "everything")
      # var uses sample variance, we use population variance
      var_mpd <- var_mpd * (n - 1) / n
      return(var_mpd)
    }
  }
  stop("input object has to be phylo or ltable")
}

context("var pair distance")

test_that("usage", {

  test_func <- function(phy) {
    n <- length(phy$tip.label)

    a2 <- stats::cophenetic(phy)
    a2 <- a2[lower.tri(a2)]
    a3 <- var(a2, na.rm = TRUE, use = "everything")
    # var calculates / (n - 1), we use n
    a3 <- a3 * (length(a2) - 1) / length(a2)
    return(a3)
  }

  standard_test(treestats::var_pair_dist, test_func,
                drop_complete = TRUE)

  test_polytomies(treestats::var_pair_dist, test_func)
})

test_that("unrooted", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
  a1 <- treestats::var_pair_dist(focal_tree)
  a2 <- treestats::var_pair_dist(ape::unroot(focal_tree))
  testthat::expect_equal(a1, a2)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::var_pair_dist(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::var_pair_dist(list()),
    "input object has to be phylo or ltable"
  )
})

#' Convert an L table to phylo object
#' @param ltab ltable
#' @param drop_extinct should extinct species be dropped from the phylogeny?
#' @return phylo object
#' @export
l_to_phylo <- function(ltab, drop_extinct = TRUE) {

  newick_str <- l_to_newick(ltab, drop_extinct)
  phylo_tree <- ape::read.tree(text = newick_str)

  return(phylo_tree)
}

#' Convert an L table to newick string
#' @param ltab ltable
#' @param drop_extinct should extinct species be dropped from the phylogeny?
#' @return phylo object
#' @export
ltable_to_newick <- function(ltab, drop_extinct = TRUE) {
  newick_str <- l_to_newick(ltab, drop_extinct)
  return(newick_str)
}

context("L2phylo")

test_that("usage", {
  if (requireNamespace("geiger") &&
      requireNamespace("DDD")) {
    set.seed(42)

    focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)

    ltable_1 <- treestats::phylo_to_l(focal_tree)

    tree2 <- treestats::l_to_phylo(ltable_1)
    tree3 <- DDD::L2phylo(ltable_1)

    ax <- geiger::is.extinct(tree2)
    ay <- geiger::is.extinct(tree3)

    testthat::expect_true(all.equal(ax, ay))

    diff_edge_length <- sort(tree2$edge.length) - sort(tree3$edge.length)
    testthat::expect_equal(mean(diff_edge_length), 0.0, tolerance = 0.001)

    # and now with a tree with extinct tips
    focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0.3, fossils = TRUE)

    ltable_1 <- treestats::phylo_to_l(focal_tree)

    tree2 <- treestats::l_to_phylo(ltable_1)
    tree3 <- DDD::L2phylo(ltable_1)

    ax <- geiger::is.extinct(tree2)
    ay <- geiger::is.extinct(tree3)

    testthat::expect_true(all.equal(ax, ay))
    diff_edge_length <- sort(tree2$edge.length) - sort(tree3$edge.length)
    testthat::expect_equal(mean(diff_edge_length), 0.0, tolerance = 0.001)

    # and again, but retain extinct lineages
    tree2 <- treestats::l_to_phylo(ltable_1, drop_extinct = FALSE)
    tree3 <- DDD::L2phylo(ltable_1, dropextinct = FALSE)

    ax <- geiger::is.extinct(tree2)
    ay <- geiger::is.extinct(tree3)

    testthat::expect_true(all.equal(ax, ay))
    diff_edge_length <- sort(tree2$edge.length) - sort(tree3$edge.length)
    testthat::expect_equal(mean(diff_edge_length), 0.0, tolerance = 0.001)
  }
})

test_that("newick", {
  if (requireNamespace("geiger")) {
    set.seed(42)
    focal_tree <- ape::rbdtree(birth = 1, death = 0, Tmax = 5)

    ltable_1 <- treestats::phylo_to_l(focal_tree)

    newick_str <- treestats::ltable_to_newick(ltable_1)
    tree_2 <- ape::read.tree(text = newick_str)
    v1 <- as.vector(unlist(treestats::calc_all_stats(focal_tree)))
    v2 <- as.vector(unlist(treestats::calc_all_stats(tree_2)))
    testthat::expect_equal(v1, v2, tolerance = 1e-3)

    focal_tree <- ape::rphylo(n = 100,
                              birth = 1, death = 0.2, T0 = 2, fossils = TRUE)

    ltable_1 <- treestats::phylo_to_l(focal_tree)

    newick_str <- treestats::ltable_to_newick(ltable_1)
    tree_2 <- ape::read.tree(text = newick_str)
    tree_2 <- geiger::drop.extinct(tree_2)
    focal_tree <- geiger::drop.extinct(focal_tree)
    v1 <- as.vector(unlist(treestats::calc_all_stats(focal_tree)))
    v2 <- as.vector(unlist(treestats::calc_all_stats(tree_2)))
    testthat::expect_equal(v1, v2, tolerance = 1e-3)
  }
})

#' Variance of leaf depth statistic
#' @description The variance of leaf depth statistic returns the variance
#' of depths across all tips, where depth of a tip indicates the distance of
#' the tip to the root.
#' @param phy phylo object or ltable
#' @param normalization "none" or "yule", when "yule" is chosen, the statistic
#' is divided by the Yule expectation
#' @return Variance of leaf depths
#' @references  T. M. Coronado, A. Mir, F. Rosselló, and L. Rotger.
#' On Sackin's original proposal: the variance of the leaves' depths as a
#' phylogenetic balance index. BMC Bioinformatics, 21(1), 2020.
#' doi: 10.1186/s12859-020-3405-1.
#' @export
var_leaf_depth <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    var_leaf_depth_stat <- calc_var_leaf_depth_ltable_cpp(phy)
    if (normalization == "yule" || normalization == TRUE) {
      n <- length(phy[, 1])
      yule_expected <-  2 * log(n)
      var_leaf_depth_stat <- var_leaf_depth_stat / yule_expected
    }
    return(var_leaf_depth_stat)
  }
  if (inherits(phy, "phylo")) {
    # polytomies are checked at the Rcpp side
    var_leaf_depth_stat <- calc_var_leaf_depth_cpp(as.vector(t(phy$edge)))

    if (normalization == "yule" || normalization == TRUE) {
      n <- length(phy$tip.label)
      yule_expected <-  2 * log(n)
      var_leaf_depth_stat <- var_leaf_depth_stat / yule_expected
    }
    return(var_leaf_depth_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' @keywords internal
calc_phylogenetic_diversity <- function(phy, t, extinct_tol) {
  if (t == 0) {
    if (!ape::is.ultrametric(phy, option = 2)) {
      if (ape::is.rooted(phy)) {
        return(calc_phylodiv_cpp(phy, t, extinct_tol))
      }
    }
    return(sum(phy$edge.length)) # no need to pass to Rcpp
  } else {
    if (ape::is.rooted(phy)) {
      return(calc_phylodiv_cpp(phy, t, extinct_tol))
    }
  }
}



#' Phylogenetic diversity at time point t
#' @description The phylogenetic diversity at time t is given by the total
#' branch length of the tree reconstructed up until time point t. Time is
#' measured increasingly, with the crown age equal to 0. Thus, the time at
#' the present is equal to the crown age.
#' @param input_obj phylo object or Ltable
#' @param t time point at which to measure phylogenetic diversity, alternatively
#' a vector of time points can also be provided. Time is measured with 0 being
#' the present.
#' @param extinct_tol tolerance to determine if a lineage is extinct at time t.
#' Default is 1/100 * smallest branch length of the tree.
#' @return phylogenetic diversity, or vector of phylogenetic diversity measures
#' if a vector of time points is used as input.
#' @references Faith, Daniel P. "Conservation evaluation and phylogenetic
#' diversity." Biological conservation 61.1 (1992): 1-10.
#' @export
phylogenetic_diversity <- function(input_obj,
                                   t = 0,
                                   extinct_tol = NULL) {

  if (inherits(input_obj, "matrix")) {

    if (length(t) == 1) {
      if (t != 0) {
        stop("Ltable implemenation can only be used for t = present = 0")
      }
      if (sum(input_obj[, 4] > 0)) {
        stop("Ltable implementation only supports extant trees")
      }

      return(calc_phylodiv_ltable_cpp(input_obj))
    } else {
       stop("Ltable implemenation can only be used for a single time point, t = 0") #nolint
    }
  }

  if (inherits(input_obj, "phylo")) {
    if (!ape::is.rooted(input_obj)) {
      stop("phylogenetic diversity is not available for unrooted trees")
    }

    if (is.null(extinct_tol)) {
      extinct_tol <- min(input_obj$edge.length) / 100
    }

    if (length(t) == 1) {
      return(calc_phylogenetic_diversity(input_obj, t, extinct_tol))
    }

    fun_to_apply <- function(focal_time) {
      return(calc_phylogenetic_diversity(input_obj,
                                         focal_time,
                                         extinct_tol))
    }

    out <- lapply(t, fun_to_apply)
    return(unlist(out))
  }

  stop("input object has to be phylo or ltable")
}

#' Total cophenetic index.
#' @description The total cophenetic index is the sum of the depth of the last
#' common ancestor of all pairs of leaves.
#' @param phy phylo object or ltable
#' @param normalization "none" or "yule", when "yule" is chosen, the statistic
#' is divided by the Yule expectation
#' @return Total cophenetic index
#' @references  A. Mir, F. Rosselló, and L. Rotger. A new balance index for
#' phylogenetic trees. Mathematical Bio-sciences, 241(1):125-136, 2013.
#' doi: 10.1016/j.mbs.2012.10.005.
#' @export
tot_coph <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    tot_coph_stat <- calc_tot_coph_ltable_cpp(phy)
    if (normalization == "yule" || normalization == TRUE) {
      n <- length(phy[, 1])
      if (n == 2) warning("normalization not valid for trees of size 2")
      h_n <- sum(1 / (1:n))
      yule_expected <-  n * (n + 1) - 2 * n * h_n
      tot_coph_stat <- tot_coph_stat / yule_expected
    }
    return(tot_coph_stat)
  }
  if (inherits(phy, "phylo")) {
    tot_coph_stat <- calc_tot_coph_cpp(as.vector(t(phy$edge)))
    if (normalization == "yule" || normalization == TRUE) {
      n <- length(phy$tip.label)
      if (n == 2) warning("normalization not valid for trees of size 2")
      h_n <- sum(1 / (1:n))
      yule_expected <-  n * (n + 1) - 2 * n * h_n
      tot_coph_stat <- tot_coph_stat / yule_expected
    }
    return(tot_coph_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' B1 metric
#' @description Balance metric (in the case of a binary tree), which measures
#' the sum across all internal nodes of one over the maximum depth of all
#' attached tips to that node.
#' When calculated on a binary tree, the b1 statistic is a true balance
#' statistic (meaning that the caterpillar tree is the most extreme, unbalanced
#' tree). However, the treestats package also supports calculation on
#' non-binary trees.
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", in which case the resulting
#' statistic is divided by the number of tips in the tree, as a crude way of
#' normalization.
#' @return B1 statistic
#' @references  K.-T. Shao and R. R. Sokal. Tree Balance.
#' Systematic Zoology, 39(3):266, 1990. doi: 10.2307/2992186.
#' @export
b1 <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    b1_stat <- calc_b1_ltable_cpp(phy)
    if (normalization == "tips" || normalization == TRUE) {
      n <- length(phy[, 1])
      b1_stat <- b1_stat / n
    }
    return(b1_stat)
  }
  if (inherits(phy, "phylo")) {
    b1_stat <- calc_b1_cpp(as.vector(t(phy$edge)))
    if (normalization == "tips" || normalization == TRUE) {
      n <- length(phy$tip.label)
      b1_stat <- b1_stat / n
    }
    return(b1_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' Maximum width of branch depths divided by the maximum depth
#' @description Calculates the maximum width divided by the maximum depth.
#' @param phy phylogeny or ltable
#' @references C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
#' transmission patterns. Evolution, Medicine, and Public Health,
#' 2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
#' @export
mw_over_md <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(calc_mw_over_md_ltable_cpp(phy))
  }
  if (inherits(phy, "phylo")) {
    return(calc_mw_over_md_cpp(as.vector(t(phy$edge))))
  }

  stop("input object has to be phylo or ltable")
}

context("mw_over_md")

test_that("usage", {

  if (requireNamespace("treebalance")) {
    standard_test(treestats::mw_over_md, treebalance::mWovermD)
    test_polytomies(treestats::mw_over_md, treebalance::mWovermD)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::mw_over_md(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::mw_over_md(list()),
    "input object has to be phylo or ltable"
  )
})

#' Stepwise increase the imbalance of a tree
#' @description the goal of this function is to increasingly imbalance a tree,
#' by changing the topology, one move at a time. It does so by re-attaching
#' terminal branches to the root lineage, through the ltable. In effect, this
#' causes the tree to become increasingly caterpillarlike. When started with
#' a balanced tree, this allows for exploring the gradient between a fully
#' balanced tree, and a fully unbalanced tree.
#' Please note that the algorithm will try to increase imbalance, until a fully
#' caterpillar like tree is reached, which may occur before unbal_steps is
#' reached.
#' Three methods are available: "youngest", reattaches branches in order of age,
#' starting with the branch originating from the most recent branching event
#' and working itself through the tree. "Random" picks a random branch to
#' reattach. "Terminal" also picks a random branch, but only from terminal
#' branches (e.g. branches that don't have any daughter lineages, which is
#' maximized in a fully imbalanced tree).
#' @param init_tree starting tree to work with
#' @param unbal_steps number of imbalance generating steps
#' @param group_method choice of "any" and "terminal"
#' @param selection_method choice of "random", "youngest" and "oldest"
#' @return phylo object
#' @export
#' @examples
#' simulated_tree <- ape::rphylo(n = 16, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' intermediate_tree <- make_unbalanced_tree(balanced_tree, 8)
#' colless(balanced_tree)
#' colless(intermediate_tree) # should be intermediate value
#' colless(unbalanced_tree) # should be highest colless value
make_unbalanced_tree <- function(init_tree,
                                 unbal_steps,
                                 group_method = "any",
                                 selection_method = "random") {
  if (!group_method %in% c("any", "terminal")) {
    stop("group method unknown, pick from 'any' and 'terminal'")
  }
  if (!selection_method %in% c("youngest", "oldest", "random")) {
    stop("selection method unknown, pick from 'youngest', 'oldest' or 'random")
  }

  ltab_in <- treestats::phylo_to_l(init_tree)
  ltab_in <- rebase_ltable(ltab_in)
  ltab_out <- NULL

  if (group_method == "any") {
    ltab_out <- make_any(ltab_in, unbal_steps, selection_method)
  }

  if (group_method == "terminal") {
    ltab_out <- make_terminal(ltab_in, unbal_steps, selection_method)
  }

  phy_out <- treestats::l_to_phylo(ltab_out)
  return(phy_out)
}

#' @keywords internal
make_any <- function(ltab_in, unbal_steps, selection_method) {
  ltab_out <- c()
  if (selection_method == "youngest") {
    ltab_out <- make_unbalanced_tree_youngest(ltab_in, unbal_steps)
  } else if (selection_method == "oldest") {
    ltab_out <- make_unbalanced_tree_oldest(ltab_in, unbal_steps)
  } else if (selection_method == "random") {
    ltab_out <- make_unbalanced_tree_random(ltab_in, unbal_steps)
  }

  return(ltab_out)
}

#' @keywords internal
make_terminal <- function(ltab_in, unbal_steps, selection_method) {
  ltab_out <- NULL
  if (selection_method == "random") {
    ltab_out <- make_unbalanced_tree_terminal(ltab_in, unbal_steps)
  } else if (selection_method == "youngest") {
    ltab_out <- make_unbalanced_tree_terminal_youngest(ltab_in, unbal_steps)
  } else if (selection_method == "oldest") {
    ltab_out <- make_unbalanced_tree_terminal_oldest(ltab_in, unbal_steps)
  }
  return(ltab_out)
}


#' @keywords internal
get_attractor <- function(ltab) {
  attractor <- 2
  # determine attractor based on number of direct daughters,
  # more daughters = less required movements.
  num_one   <- length(which(ltab[, 2] == -1))
  num_two   <- length(which(ltab[, 2] == 2))

  if (num_one > num_two) {
    attractor <- -1
  }
  return(attractor)
}

#' @keywords internal
make_unbalanced_tree_youngest <- function(ltab,
                                          unbal_steps) {

  attractor <- get_attractor(ltab)

  to_sample_from <- which(ltab[, 2] != attractor &
                          ltab[, 3] != -1 &
                          ltab[, 3] != 2)
  steps_taken <- 0
  while (steps_taken < unbal_steps) {
    ages <- ltab[to_sample_from, 1]
    focal_step <- to_sample_from[which.min(ages)]
    ltab[focal_step, 2] <- attractor
    to_sample_from <- which(ltab[, 2] != attractor &
                            ltab[, 3] != -1 &
                            ltab[, 3] != 2)
    if (length(to_sample_from) < 1) break
    steps_taken <- steps_taken + 1
  }
  return(ltab)
}


#' @keywords internal
make_unbalanced_tree_oldest <- function(ltab,
                                          unbal_steps) {
  attractor <- get_attractor(ltab)

  to_sample_from <- which(ltab[, 2] != attractor &
                          ltab[, 3] != -1 &
                          ltab[, 3] != 2)
  steps_taken <- 0
  while (steps_taken < unbal_steps) {
    ages <- ltab[to_sample_from, 1]
    focal_step <- to_sample_from[which.max(ages)]
    ltab[focal_step, 2] <- attractor
    to_sample_from <- which(ltab[, 2] != attractor &
                            ltab[, 3] != -1 &
                            ltab[, 3] != 2)
    if (length(to_sample_from) < 1) break
    steps_taken <- steps_taken + 1
  }
  return(ltab)
}


#' @keywords internal
make_unbalanced_tree_random <- function(ltab,
                                        unbal_steps) {
  attractor <- get_attractor(ltab)

  to_sample_from <- which(ltab[, 2] != attractor &
                          ltab[, 3] != -1 &
                          ltab[, 3] != 2)

  if (length(to_sample_from) >= 1) {
    for (steps_taken in 1:unbal_steps) {
      focal_step <- DDD::sample2(to_sample_from, 1)
      if (length(to_sample_from) == 1) focal_step <- to_sample_from
      ltab[focal_step, 2] <- attractor
      to_sample_from <- which(ltab[, 2] != attractor &
                              ltab[, 3] != -1 &
                              ltab[, 3] != 2)

      if (length(to_sample_from) < 1) break
    }
  }
  return(ltab)
}


#' @keywords internal
make_unbalanced_tree_terminal <- function(ltab,
                                          unbal_steps) {
  attractor <- get_attractor(ltab)

  ltab <- cbind(ltab, 0)
  for (i in seq_along(ltab[, 1])) {
    ref <- ltab[i, 3]
    ltab[i, 5] <- length(which(ltab[, 2] == ref))
  }

  for (n in 1:unbal_steps) {
    to_sample <- which(ltab[, 5] ==  0 &
                       ltab[, 3] !=  2 &
                       ltab[, 3] != -1 &
                       ltab[, 2] != attractor)
    if (length(to_sample) < 1) {
      break
    }

    focal_spec <- DDD::sample2(to_sample, 1)
    if (length(to_sample) == 1) focal_spec <- to_sample

    parent_spec <- abs(ltab[focal_spec, 2])
    ltab[parent_spec, 5] <- ltab[parent_spec, 5] - 1
    ltab[focal_spec, 2] <- attractor
    ltab[abs(attractor), 5] <- ltab[abs(attractor), 5] + 1
  }

  return(ltab)
}

#' @keywords internal
make_unbalanced_tree_terminal_youngest <- function(ltab,   #nolint
                                          unbal_steps) {

  attractor <- get_attractor(ltab)

  ltab <- cbind(ltab, 0)
  for (i in seq_along(ltab[, 1])) {
    ref <- ltab[i, 3]
    ltab[i, 5] <- length(which(ltab[, 2] == ref))
  }

  for (n in 1:unbal_steps) {
    to_sample <- which(ltab[, 5] ==  0 &
                       ltab[, 3] !=  2 &
                       ltab[, 3] != -1 &
                       ltab[, 2] != attractor)

    ages <- ltab[to_sample, 1]
    focal_spec <- to_sample[which.min(ages)]

    parent_spec <- abs(ltab[focal_spec, 2])
    ltab[parent_spec, 5] <- ltab[parent_spec, 5] - 1
    ltab[focal_spec, 2] <- attractor
    ltab[abs(attractor), 5] <- ltab[abs(attractor), 5] + 1
    if (length(to_sample) < 1) {
      break
    }
  }

  return(ltab)
}

#' @keywords internal
make_unbalanced_tree_terminal_oldest <- function(ltab,   #nolint
                                                   unbal_steps) {

  attractor <- get_attractor(ltab)

  ltab <- cbind(ltab, 0)
  for (i in seq_along(ltab[, 1])) {
    ref <- ltab[i, 3]
    ltab[i, 5] <- length(which(ltab[, 2] == ref))
  }

  for (n in 1:unbal_steps) {
    to_sample <- which(ltab[, 5] ==  0 &
                       ltab[, 3] !=  2 &
                       ltab[, 3] != -1 &
                       ltab[, 2] != attractor)

    ages <- ltab[to_sample, 1]
    focal_spec <- to_sample[which.max(ages)]

    parent_spec <- abs(ltab[focal_spec, 2])
    ltab[parent_spec, 5] <- ltab[parent_spec, 5] - 1
    ltab[focal_spec, 2] <- attractor
    ltab[abs(attractor), 5] <- ltab[abs(attractor), 5] + 1
    if (length(to_sample) < 1) {
      break
    }
  }

  return(ltab)
}

context("make_unbalanced_tree")

test_that("usage", {
  set.seed(42)
  tree_size <- 8
  focal_tree <- ape::rphylo(n = tree_size, birth = 1, death = 0)

  num_steps <- tree_size * 2
  for (g_method in c("any", "terminal")) {
    for (s_method in c("oldest", "youngest", "random")) {
      unbal_tree <- treestats::make_unbalanced_tree(init_tree = focal_tree,
                                                    unbal_steps = num_steps,
                                                    group_method = g_method,
                                                    selection_method = s_method)
      max_d <- treestats::max_depth(unbal_tree)
      testthat::expect_equal(max_d, tree_size - 1)
    }
  }
})

test_that("abuse", {
  set.seed(42)
  tree_size <- 8
  focal_tree <- ape::rphylo(n = tree_size, birth = 1, death = 0)

  testthat::expect_error(
  unbal_tree <- treestats::make_unbalanced_tree(init_tree = focal_tree,
                                                unbal_steps = num_steps,
                                                group_method = "x",
                                                selection_method = "random"),
  "group method unknown, pick from 'any' and 'terminal'")

  testthat::expect_error(
    unbal_tree <- treestats::make_unbalanced_tree(init_tree = focal_tree,
                                                  unbal_steps = num_steps,
                                                  group_method = "any",
                                                  selection_method = "x"),
    "selection method unknown, pick from 'youngest', 'oldest' or 'random"
  )
})

#' Create an unbalanced tree (caterpillar tree)
#' @param phy phylo object
#' @return phylo phylo object
#' @description This function takes an input phylogeny, and returns a phylogeny
#' that is a perfectly imbalanced tree (e.g. a full caterpillar tree), that
#' has the same branching times as the original input tree.
#' @export
#' @examples
#' phy <- ape::rphylo(n = 16, birth = 1, death = 0)
#' bal_tree <- treestats::create_fully_unbalanced_tree(phy)
#' treestats::colless(phy)
#' treestats::colless(bal_tree) # much higher
create_fully_unbalanced_tree <- function(phy) {

  if (!inherits(phy, "phylo")) {
    stop("This function requires a phylogeny as input")
  }

  brts <- as.vector(sort(treestats::branching_times(phy), decreasing = TRUE))

  ltab <- rbind(c(brts[1], 0, -1, -1),
                c(brts[1], -1, 2, -1))

  leftcnt <- 2
  cnt <- 2
  for (i in 2:length(brts)) {
    focal_brt <- brts[i]
    cnt <- cnt + 1
    parent <- leftcnt
    daughter <- cnt * -1
    leftcnt <- daughter
    to_add <- c(focal_brt, parent, daughter, -1)
    ltab <- rbind(ltab, to_add)
  }

  return(treestats::l_to_phylo(ltab))
}


#' Create a fully balanced tree
#' @param phy phylo object
#' @return phylo phylo object
#' @description This function takes an input phylogeny, and returns a phylogeny
#' that is most ideally balanced tree, whilst having the same branching times as
#' the original input tree. Please note that if the number of tips is not even
#' or not a power of two, the tree may not have perfect balance, but the most
#' ideal balance possible.
#' @export
#' @examples phy <- ape::rphylo(n = 16, birth = 1, death = 0)
#' bal_tree <- treestats::create_fully_balanced_tree(phy)
#' treestats::colless(phy)
#' treestats::colless(bal_tree) # much lower
create_fully_balanced_tree <- function(phy) {
  if (!inherits(phy, "phylo")) {
    stop("This function requires a phylogeny as input")
  }

  brts <- as.vector(sort(treestats::branching_times(phy), decreasing = TRUE))

  ltab <- rbind(c(brts[1], 0, -1, -1),
                c(brts[1], -1, 2, -1))

  cnt <- 3
  current_tips_left <- c(2)
  current_tips_right <- c(-1)
  brts <- brts[-1]

  while (length(brts) > 0) {
    num_to_do <- length(current_tips_left)
    for (i in 1:num_to_do) {
      if (i > length(brts)) {
        break
      }
      brt_left <- brts[i]
      parent <- current_tips_left[i]
      daughter <- cnt
      to_add <- c(brt_left, parent, daughter, -1)
      current_tips_left <- c(current_tips_left, daughter)
      ltab <- rbind(ltab, to_add)
      cnt <- cnt + 1
    }
    brts <- brts[-c(1:num_to_do)]

    num_to_do <- length(current_tips_right)

    for (i in 1:num_to_do) {
      if (i > length(brts)) {
        break
      }
      brt_right <- brts[i]
      parent <- current_tips_right[i]
      daughter <- cnt * -1
      to_add <- c(brt_right, parent, daughter, -1)
      current_tips_right <- c(current_tips_right, daughter)
      ltab <- rbind(ltab, to_add)
      cnt <- cnt + 1
    }
    brts <- brts[-c(1:num_to_do)]

  }
  return(treestats::l_to_phylo(ltab))
}

context("beta")

test_that("usage", {
  if (requireNamespace("apTreeshape")) {
    beta_treestats <- treestats::beta_statistic(extant_tree)
    beta_ref       <- apTreeshape::maxlik.betasplit(
      apTreeshape::as.treeshape(extant_tree))

    testthat::expect_equal(beta_treestats, beta_ref$max_lik, tolerance = 0.05)

    bal_tree <- treestats::create_fully_balanced_tree(extant_tree)

    beta_treestats <- treestats::beta_statistic(bal_tree)
    beta_ref       <- apTreeshape::maxlik.betasplit(bal_tree)

    testthat::expect_equal(beta_treestats, beta_ref$max_lik, tolerance = 0.05)


    unbal_tree <- treestats::create_fully_unbalanced_tree(extant_tree)

    beta_treestats <- treestats::beta_statistic(unbal_tree)
    beta_ref       <- apTreeshape::maxlik.betasplit(unbal_tree)

    testthat::expect_equal(beta_ref$max_lik, -2, tolerance = 0.01)
    testthat::expect_equal(beta_treestats, beta_ref$max_lik, tolerance = 0.05)
    testthat::expect_equal(beta_treestats, -2, tolerance = 0.01)


    # test methods
    comp_methods <- function(sim_tree) {
      beta1 <- treestats::beta_statistic(sim_tree, algorithm = "COBYLA")
      beta2 <- treestats::beta_statistic(sim_tree, algorithm = "subplex")
      beta3 <- treestats::beta_statistic(sim_tree, algorithm = "simplex")
      beta4 <- apTreeshape::maxlik.betasplit(sim_tree)$max_lik
      testthat::expect_equal(beta1, beta4, tolerance = 0.1)
      testthat::expect_equal(beta1, beta2, tolerance = 0.1)
      testthat::expect_equal(beta1, beta3, tolerance = 0.1)

      ltab <- treestats::phylo_to_l(sim_tree)
      beta5 <- treestats::beta_statistic(ltab)
      testthat::expect_equal(beta1, beta5, tolerance = 0.1)
    }

    comp_methods(bal_tree)
    comp_methods(extant_tree)

    # ltable
    beta_treestats <- treestats::beta_statistic(extant_tree)
    focal_ltable <- treestats::phylo_to_l(extant_tree)
    beta_ltable  <- treestats::beta_statistic(focal_ltable)
    testthat::expect_equal(beta_ltable, beta_treestats, tolerance = 0.1)
  }
})

test_that("polytomies", {
  if (requireNamespace("apTreeshape")) {

    local_func <- function(phy) {
      res <- apTreeshape::avgLeafDepI(phy)
      return(res$max_lik)
    }

    testthat::expect_warning(
      test_polytomies(treestats::beta, local_func)
    )
  }
})


# abuse
test_that("abuse", {
  focal_tree <- ape::rphylo(n = 10, birth = 1, death = 0)

  testthat::expect_output(
    treestats::beta_statistic(phy = focal_tree, algorithm = "none"),
    "no algorithm chosen")

  focal_tree <- ape::rphylo(n = 2, birth = 1, death = 0)

  testthat::expect_warning(
    treestats::beta_statistic(phy = focal_tree, algorithm = "none"),
    "Trees with only two tips have undefined beta")
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::beta_statistic(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::beta_statistic(list()),
    "input object has to be phylo or ltable"
  )
})

context("j_one")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 128, birth = 1, death = 0)

  bal_tree <- treestats::create_fully_balanced_tree(focal_tree)
  unbal_tree <- treestats::create_fully_unbalanced_tree(focal_tree)

  j_one_1 <- treestats::j_one(bal_tree)
  j_one_2 <- treestats::j_one(unbal_tree)

  testthat::expect_equal(j_one_1, 1)
  testthat::expect_equal(j_one_2, 0.1085403, tolerance = 0.001)

  j_one_1_l <- treestats::j_one(treestats::phylo_to_l(bal_tree))
  j_one_2_l <- treestats::j_one(treestats::phylo_to_l(unbal_tree))

  testthat::expect_equal(j_one_1_l, 1)
  testthat::expect_equal(j_one_2_l, 0.1085403, tolerance = 0.001)
})

test_that("polytomies", {
  # phyloTop does not support non-binary trees, so no comparison possible.
  for (focal_tree in poly_trees) {
    local_stat <- try_stat(focal_tree, treestats::j_one)
    testthat::expect_true(is.na(local_stat))
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::j_one(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::j_one(list()),
    "input object has to be phylo or ltable"
  )
})

context("four prong")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 4, birth = 1, death = 0)

  bal_tree <- treestats::create_fully_balanced_tree(focal_tree)

  c1 <- treestats::four_prong(bal_tree)
  testthat::expect_equal(c1, 0)

  ltab <- treestats::phylo_to_l(bal_tree)
  c2 <- treestats::four_prong(ltab)
  testthat::expect_equal(c2, 0)


  unbal_tree <- treestats::create_fully_unbalanced_tree(focal_tree)

  c1 <- treestats::four_prong(unbal_tree)
  testthat::expect_equal(c1, 1)

  ltab <- treestats::phylo_to_l(unbal_tree)
  c2 <- treestats::four_prong(ltab)
  testthat::expect_equal(c2, 1)
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    res <- treestats::four_prong(focal_tree)
    testthat::expect_true(!is.na(res))
  }

  # now on artificial trees
  focal_tree <- ape::read.tree(text = "((((A:0.1,B:0.1):0.2,C:0.1):0.3,D:0.1):0.4,E:0.1);") # nolint
  res <- treestats::four_prong(focal_tree)
  testthat::expect_equal(res, 1)

  # now we create a poly
  focal_tree <- ape::read.tree(text = " ((((A:0.1,B:0.1,C:0.1):0.2,D:0.1):0.3,E:0.1):0.4);") # nolint
  res <- treestats::four_prong(focal_tree)
  testthat::expect_equal(res, 0)
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::four_prong(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::four_prong(list()),
    "input object has to be phylo or ltable"
  )
})

context("root_imbalance")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 64, birth = 1, death = 0)

  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::root_imbalance(focal_tree),
                         treestats::root_imbalance(ltab))
  bal_tree <- treestats::create_fully_balanced_tree(focal_tree)
  a1 <- treestats::root_imbalance(bal_tree)
  testthat::expect_equal(a1, 0.5)
  a2 <- treestats::root_imbalance(treestats::phylo_to_l(bal_tree))
  testthat::expect_equal(a2, 0.5)

  unbal_tree <- treestats::create_fully_unbalanced_tree(focal_tree)

  a1 <- treestats::root_imbalance(unbal_tree)
  testthat::expect_equal(a1, 63 / 64)
  a2 <- treestats::root_imbalance(treestats::phylo_to_l(unbal_tree))
  testthat::expect_equal(a2, 63 / 64)

  # with extinct species:
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0.2, fossils = TRUE)

  ltab <- treestats::phylo_to_l(focal_tree)
  testthat::expect_equal(treestats::root_imbalance(focal_tree),
                         treestats::root_imbalance(ltab))

})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    if (ape::is.rooted(focal_tree)) {
        res <- try_stat(focal_tree, treestats::root_imbalance)
        testthat::expect_true(!is.na(res))
    }
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::root_imbalance(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::root_imbalance(list()),
    "input object has to be phylo or ltable"
  )
})

context("create_bal_unbal_trees")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 16, birth = 1, death = 0)
  bal_tree <- treestats::create_fully_balanced_tree(focal_tree)
  unbal_tree <- treestats::create_fully_unbalanced_tree(focal_tree)

  a1 <- treestats::cherries(bal_tree)
  testthat::expect_equal(a1, 8)
  a2 <- treestats::pitchforks(unbal_tree)
  testthat::expect_equal(a2, 1)

  a1 <- treestats::colless(bal_tree)
  a2 <- treestats::colless(unbal_tree)
  testthat::expect_gt(a2, a1)

  ax <- treestats::imbalance_steps(bal_tree)
  ay <- treestats::imbalance_steps(unbal_tree)
  testthat::expect_gt(ax, ay)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::create_fully_balanced_tree(10),
    "This function requires a phylogeny as input"
  )

  testthat::expect_error(
    treestats::create_fully_balanced_tree(list()),
    "This function requires a phylogeny as input"
  )

  testthat::expect_error(
    treestats::create_fully_unbalanced_tree(10),
    "This function requires a phylogeny as input"
  )

  testthat::expect_error(
    treestats::create_fully_unbalanced_tree(list()),
    "This function requires a phylogeny as input"
  )
})

#' Imbalance steps index
#' @description Calculates the number of moves required to transform the focal
#' tree into a fully imbalanced (caterpillar) tree. Higher value indicates a
#' more balanced tree.
#' @param input_obj phylo object or ltable
#' @param normalization if true, the number of steps taken is normalized by tree
#' size, by dividing by the maximum number of moves required to move from a
#' fully balanced to a fully imbalanced tree, which is \eqn{N - log2(N) - 1},
#' where N is the number of extant tips.
#' @return required number of moves
#' @export
imbalance_steps <- function(input_obj,
                            normalization = FALSE) {

  check_tree(input_obj,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(input_obj, "phylo")) {
    input_obj <- treestats::phylo_to_l(input_obj)
  }

  if (!inherits(input_obj, "matrix")) {
    stop("input object has to be phylo or ltable")
  }
  if (length(input_obj[, 1]) == 2) {
    message("Balance of a tree with two tips is undefined")
    return(NA)
  }

  steps_taken <- imbalance_steps_cpp(input_obj,
                                     normalization)
  return(steps_taken)
}

#' Maximum ladder index
#' @description Calculate the maximum ladder index, from the phyloTop package.
#' Higher values indicate more unbalanced trees. To calculate the maximum ladder
#' index, first all potential ladders in the tree are calculated. A ladder is
#' defined as a sequence of nodes where one of the daughter branches is a
#' terminal branch, resulting in a 'ladder' like pattern. The maximum ladder
#' index then represents the longest ladder found among all observed ladders
#' in the tree.
#' @param input_obj phylo object or ltable
#' @return longest ladder in the tree
#' @export
max_ladder <- function(input_obj) { # nolint

  check_tree(input_obj,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(input_obj, "matrix")) {
    input_obj <- treestats::l_to_phylo(input_obj, drop_extinct = FALSE)
  }
  if (inherits(input_obj, "phylo")) {
    return(max_ladder_cpp(as.vector(t(input_obj$edge))))
  }
  stop("input object has to be phylo or ltable")
}

context("max_ladder")

test_that("usage", {

  if (requireNamespace("phyloTop")) {
    set.seed(42)

    focal_tree <- ape::rphylo(n = 1000, birth = 1, death = 0)

    c1 <- treestats::max_ladder(focal_tree)
    c2 <- max(phyloTop::ladderSizes(focal_tree)$ladderSizes)
    testthat::expect_equal(c1, c2)

    c3 <- treestats::max_ladder(treestats::phylo_to_l(focal_tree))
    testthat::expect_equal(c1, c3)

    # we use larger trees here to induce larger ladders.
    focal_tree <- ape::rphylo(n = 1000, birth = 1, death = 0.5, fossils = TRUE)

    c1 <- treestats::max_ladder(focal_tree)
    c2 <- max(phyloTop::ladderSizes(focal_tree)$ladderSizes)
    testthat::expect_equal(c1, c2)

    c3 <- treestats::max_ladder(treestats::phylo_to_l(focal_tree))
    testthat::expect_equal(c1, c3)
  }
})

test_that("polytomies", {
  if (requireNamespace("treebalance")) {
    test_func <- function(phy) {
      max(phyloTop::ladderSizes(focal_tree)$ladderSizes)
    }

    test_polytomies(treestats::max_ladder, test_func)
  }
})



test_that("wrong_object", {
  testthat::expect_error(
    treestats::max_ladder(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::max_ladder(list()),
    "input object has to be phylo or ltable"
  )
})

#' Number of pitchforks
#' @description
#' Pitchforks are a clade with three tips, as introduced in the phyloTop
#' package.
#' @param input_obj phylo object or ltable
#' @param normalization "none" or "tips", in which case the found number of
#' pitchforks is divided by the expected number.
#' @return number of pitchforks
#' @export
pitchforks <- function(input_obj, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(input_obj,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(input_obj, "matrix")) {
    pitch_stat <- pitchforks_ltable_cpp(input_obj)
    if (normalization == "tips" || normalization == TRUE) {
      pitch_stat <- pitch_stat * 3 / length(input_obj[, 1])
    }
    return(pitch_stat)
  }
  if (inherits(input_obj, "phylo")) {
    pitch_stat <- pitchforks_cpp(as.vector(t(input_obj$edge)))
    if (normalization == "tips" || normalization == TRUE) {
      pitch_stat <- pitch_stat * 3 / length(input_obj$tip.label)
    }
    return(pitch_stat)
  }

  stop("input object has to be phylo or ltable")
}

#' Maximum depth metric
#' @description The maximum depth metric, measures the maximal path (in edges),
#' between the tips and the root.
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", in which case the resulting statistic
#' is divided by the number of tips in the tree.
#' @return Maximum depth (in number of edges)
#' @references  C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
#' transmission patterns. Evolution, Medicine, and Public Health,
#' 2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
#' @export
max_depth <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    max_d_stat <- calc_max_depth_ltable_cpp(phy)
    if (normalization == "tips" || normalization == TRUE) {
      max_d_stat <- max_d_stat / length(phy[, 1])
    }
    return(max_d_stat)
  }
  if (inherits(phy, "phylo")) {
    max_d_stat <- calc_max_depth_cpp(as.vector(t(phy$edge)))
    if (normalization == "tips" || normalization == TRUE) {
      max_d_stat <- max_d_stat / length(phy$tip.label)
    }
    return(max_d_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' Adjancency Matrix properties
#' @description Calculates the eigenvalues of the Adjancency Matrix, where the
#' Adjacency matrix is a square matrix indicate whether pairs of vertices
#' are adjacent or not on a graph - here, entries in the matrix indicate
#' connections between nodes (and betweens nodes and tips). Entries in the
#' adjacency matrix are weighted by branch length. Then, using the adjacency
#' matrix, we calculate the spectral properties of the matrix, e.g. the
#' minimum positive and maximum positive eigenvalues of the matrix.
#' When the R package \pkg{RSpectra} is available,
#' a faster calculation can be used, which does not calculate all eigenvalues,
#' but only the maximum and minimum. As such, when using this option,
#' the vector of all eigenvalues is not returned.
#' @param phy phylo object or ltable
#' @param use_rspectra boolean to indicate whether the helping package RSpectra
#' should be used, in which case only the minimum and maximum values are
#' returned
#' @param force_min boolean to indicate whether or not to also calculate
#' the minimum positive eigenvalue if the tree is larger than 23170 tips.
#' Default is FALSE, as this can be very slow, but in some cases might be
#' of interest.
#' @return List with the minimum and maximum eigenvalues
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." Plos one 16.12 (2021): e0259877.
#' @export
minmax_adj <- function(phy, use_rspectra = TRUE, force_min = FALSE) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {

    # when not using RSpectra, things are slow and restricted
    if (!use_rspectra || !requireNamespace("RSpectra")) {
      return(slow_calc(phy))
    } else {
      edges   <- phy$edge
      lengths <- phy$edge.length
      n_nodes <- max(edges)

      matvec_fun <- function(x, args) {
        Ax_tree(edges, lengths, x, n_nodes)
      }

      mat_size <- max(phy$edge)
      max_val <- RSpectra::eigs_sym(matvec_fun,
                                    n = mat_size,
                                    k = 1,
                                    which = "LA",
                                    opts = list(retvec = FALSE))$values

      if (mat_size > 46340) {  # floor(sqrt(2^31 - 1)))
        if (force_min == TRUE) {
          warning("calculating the minimum eigenvalue
                   is very costly for large trees")
          res <- RSpectra::eigs_sym(matvec_fun,
                                    n = n_nodes,
                                    # we don't need negative eigenvalues:
                                    k = n_nodes / 2,
                                    which = "SM",
                                    opts = list(retvec = FALSE))$values
          res <- res[round(res, 10) > 0]
          res_min <- min(res)
          return(list("max" = max_val,
                      "min" = res_min))
        } else {
          warning("skipped calculating the minimum eigenvalue of the adjacency
                  matrix, because the tree is too large")
          return(list("max" = max_val))
        }
      } else {
        # if the tree is not very large, then using the matrix is faster
        lap_mat <- matrix(0, mat_size, mat_size)
        lap_mat[phy$edge] <- phy$edge.length
        lap_mat <- lap_mat + t(lap_mat)

        # we can make use of RSpectra
        min_val <- RSpectra::eigs_sym(lap_mat, k = 1, which = "BE",
                                      sigma = 1e-9,
                                      opts = list(retvec = FALSE))$values

        return(list(min = min_val,
                    max = max_val))
      }
    }
  }

  stop("input object has to be phylo or ltable")
}


#' @keywords internal
slow_calc <- function(phy) {

  # prepping matrix in Rcpp yields no speed gain
  mat_size <- max(phy$edge)

  if (mat_size > 10000) {
    warning("calculating all eigenvalues without RSpectra can be slow
              for large trees")
  }

  if (mat_size > 46340) {  # floor(sqrt(2^31 - 1)))
    stop("tree too big, memory allocation fail")
  }
  lap_mat <- matrix(0, mat_size, mat_size)
  lap_mat[phy$edge] <- phy$edge.length
  lap_mat <- lap_mat + t(lap_mat)

  eigen_vals <- eigen(lap_mat,
                      symmetric = TRUE,
                      only.values = TRUE)$values

  eigen_vals <- round(eigen_vals, digits = 10)

  max_val <- max(eigen_vals)
  min_val <- min(eigen_vals[eigen_vals > 0])
  return(list("max" = max_val,
              "min" = min_val,
              "values" = eigen_vals))
}

context("minmax_adj")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 10, birth = 1, death = 0)

  if (requireNamespace("igraph")) {

    a1_1 <- treestats::minmax_adj(focal_tree)

    df <- as.data.frame(cbind(focal_tree$edge,
                              weight = focal_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)

    adj_mat <- igraph::as_adjacency_matrix(g, attr = "weight", sparse = FALSE)
    ref <- eigen(adj_mat)$values
    ref <- round(ref, digits = 10)

    testthat::expect_equal(a1_1$min, min(ref[ref > 0]))
    testthat::expect_equal(a1_1$max, max(ref))

    ltab <- treestats::phylo_to_l(focal_tree)
    a1_2 <- treestats::minmax_adj(ltab)
    testthat::expect_equal(a1_1$values, a1_2$values)
  }

  # test rspectra use
  ref <- treestats::minmax_adj(focal_tree,
                               use_rspectra = FALSE)

  a2 <- treestats::minmax_adj(focal_tree,
                              use_rspectra = TRUE)
  testthat::expect_equal(ref$min, a2$min)
  testthat::expect_equal(ref$max, a2$max)
})

test_that("polytomies", {
  if (requireNamespace("igraph")) {

    test_func <- function(phy) {
      df <- as.data.frame(cbind(phy$edge,
                                weight = phy$edge.length))
      g <- igraph::graph_from_data_frame(df, directed = FALSE)

      adj_mat <- igraph::as_adjacency_matrix(g, attr = "weight", sparse = FALSE)
      ref <- eigen(adj_mat)$values
      ref <- round(ref, digits = 10)
      return(ref)
    }

    min_ref <- function(phy) {
      ref <- test_func(phy)
      return(min(ref[ref > 0]))
    }
    max_ref <- function(phy) {
      return(max(test_func(phy)))
    }

    min_treestats <- function(phy) {
      res <- treestats::minmax_adj(phy)
      return(res$min)
    }
    max_treestats <- function(phy) {
      res <- treestats::minmax_adj(phy)
      return(res$max)
    }


    test_polytomies(min_treestats, min_ref, tol = 1e-5)
    test_polytomies(max_treestats, max_ref)
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::minmax_adj(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::minmax_adj(list()),
    "input object has to be phylo or ltable"
  )
})

#' Sackin index of (im)balance.
#' @description The Sackin index is calculated as the sum of ancestors for each
#' of the tips. Higher values indicate higher imbalance. Two normalizations
#' are available, where a correction is made for tree size, under either a
#' Yule expectation, or a pda expectation.
#' @param phy phylogeny or ltable
#' @param normalization normalization, either 'none' (default), "yule" or "pda".
#' @return Sackin index
#' @references M. J. Sackin (1972). "Good" and "Bad" Phenograms.
#' Systematic Biology. 21:225-226.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' sackin(balanced_tree)
#' sackin(unbalanced_tree) # should be much higher
sackin <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(calc_sackin_ltable_cpp(phy, normalization))
  }
  if (inherits(phy, "phylo")) {
    return(calc_sackin_cpp(as.vector(t(phy$edge)), normalization))
  }
  stop("input object has to be phylo or ltable")
}

context("sackin")

test_that("usage", {
  if (requireNamespace("treebalance")) {

    standard_test(treestats::sackin, treebalance::sackinI)

    test_polytomies(treestats::sackin, treebalance::sackinI)
  }

  if (requireNamespace("apTreeshape")) {

    test_func <- function(phy) {
      apTreeshape::sackin(apTreeshape::as.treeshape(phy))
    }

    standard_test(treestats::sackin, test_func)

    # no polytomies test, as apTreeshape does not support it.

    sackin <- treestats::sackin(extant_tree, normalization = "yule")
    sackin_check <- apTreeshape::sackin(apTreeshape::as.treeshape(extant_tree),
                                        norm = "yule")
    testthat::expect_equal(sackin, sackin_check, tol = 1e-5)

    sackin <- treestats::sackin(extant_tree, normalization = "pda")
    sackin_check <- apTreeshape::sackin(apTreeshape::as.treeshape(extant_tree),
                                        norm = "pda")
    testthat::expect_equal(sackin, sackin_check, tol = 1e-5)


    # test ltable functionality
    ltab <- treestats::phylo_to_l(extant_tree)
    testthat::expect_equal(treestats::sackin(extant_tree),
                           treestats::sackin(ltab))

    testthat::expect_equal(treestats::sackin(extant_tree,
                                             normalization = "yule"),
                           treestats::sackin(ltab,
                                             normalization = "yule"))

    testthat::expect_equal(treestats::sackin(extant_tree,
                                             normalization = "pda"),
                           treestats::sackin(ltab,
                                             normalization = "pda"))
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::sackin(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::sackin(list()),
    "input object has to be phylo or ltable"
  )
})

#' Maximum difference of widths of a phylogenetic tree
#' @description Calculates the maximum difference of widths of a phylogenetic
#' tree. First, the widths are calculated by collecting the depth of each node
#' and tip across the entire tree, where the depth represents the distance
#' (in nodes) to the root. Then, the width represents the number of occurrences
#' of each possible depth. Then, we take the difference between each consecutive
#' width, starting with the first width. The maximum difference is then
#' returned - whereas the original statistic designed by Colijn and Gardy used
#' the absolute maximum difference, we here use the modified version as
#' introduced in Fischer 2023: this returns the maximum value, without
#' absoluting negative widths. This ensures that this metric is a proper
#' (im)balance metric, following Fischer 2023.
#' @param phy phylogeny or ltable
#' @param normalization "none" or "tips", in which case the resulting statistic
#' is divided by the number of tips in the tree.
#' @return maximum difference of widths
#' @references C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
#' transmission patterns. Evolution, Medicine, and Public Health,
#' 2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
#'
#' Fischer, M., Herbst, L., Kersting, S., Kühn, A. L., & Wicke, K. (2023).
#' Tree Balance Indices: A Comprehensive Survey.
#' @export
max_del_width <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    max_dw_stat <- calc_max_del_width_ltable_cpp(phy)
    if (normalization == "tips" || normalization == TRUE) {
      max_dw_stat <- max_dw_stat / length(phy[, 1])
    }
    return(max_dw_stat)
  }
  if (inherits(phy, "phylo")) {
    # polytomies are checked at the Rcpp side
    max_dw_stat <- calc_max_del_width_cpp(as.vector(t(phy$edge)))

    if (normalization == "tips" || normalization == TRUE) {
      max_dw_stat <- max_dw_stat / length(phy$tip.label)
    }
    return(max_dw_stat)
  }

  stop("input object has to be phylo or ltable")
}

#' Blum index of (im)balance.
#' @description The Blum index of imbalance (also known as the s-shape
#' statistic (see \link{sshape})) calculates the sum of \eqn{log(N-1)} over all
#' internal nodes, where N represents the total number of extant tips connected
#' to that node. An alternative implementation can be found in the Castor R
#' package.
#' @param phy phylogeny or ltable
#' @param normalization because the Blum index sums over all nodes,
#' the resulting statistic tends to be correlated with the number of extant
#' tips. Normalization can be performed by dividing by the number of extant
#' tips.
#' @return Blum index of imbalance
#' @references M. G. B. Blum and O. Francois (2006). Which random processes
#' describe the Tree of Life? A large-scale study of phylogenetic tree
#' imbalance. Systematic Biology. 55:685-691.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#'   balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#'   unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#'   blum(balanced_tree)
#'   blum(unbalanced_tree) # should be higher
blum <- function(phy,
                 normalization = FALSE) {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(calc_blum_ltable_cpp(phy, normalization))
  }
  if (inherits(phy, "phylo")) {
    return(calc_blum_cpp(as.vector(t(phy$edge)), normalization))
  }
  stop("input object has to be phylo or ltable")
}


#' s shape statistic of (im)balance.
#' @description The s shape statistic of imbalance (also known as the Blum
#' statistic, see \link{blum}) calculates the sum of \eqn{log(N-1)} over all
#' internal nodes, where N represents the total number of extant tips connected
#' to that node. An alternative implementation can be found in the Castor R
#' package.
#' @param phy phylogeny or ltable
#' @param normalization because the sshape statistic sums over all nodes,
#' the resulting statistic tends to be correlated with the number of extant
#' tips. Normalization can be performed by dividing by the number of extant
#' tips.
#' @return s shape statistic of imbalance
#' @references M. G. B. Blum and O. Francois (2006). Which random processes
#' describe the Tree of Life? A large-scale study of phylogenetic tree
#' imbalance. Systematic Biology. 55:685-691.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#'   balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#'   unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#'   sshape(balanced_tree)
#'   sshape(unbalanced_tree) # should be higher
sshape <- function(phy,
                   normalization = FALSE) {
  return(blum(phy, normalization))
}

context("blum")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    standard_test(treestats::blum, treebalance::sShapeI)
  }
})

test_that("usage", {
  set.seed(5)
  focal_tree <- ape::rphylo(n = 4, birth = 1, death = 0)
  blum1 <- treestats::blum(focal_tree)
  blum_check <- log2(2 - 1) + log2(3 - 1) + log2(4 - 1)
  testthat::expect_equal(blum1, blum_check)
  ltab <- treestats::phylo_to_l(focal_tree)
  blum2 <- treestats::blum(ltab)
  testthat::expect_equal(blum1, blum2)

  sshape1 <- treestats::sshape(focal_tree)
  testthat::expect_equal(sshape1, blum1)
  sshape2 <- treestats::sshape(ltab)
  testthat::expect_equal(sshape2, blum1)
})

test_that("polytomies", {
  if (requireNamespace("treebalance")) {
    test_polytomies(treestats::blum, treebalance::sShapeI)
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::blum(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::blum(list()),
    "input object has to be phylo or ltable"
  )
})

test_that("normalisation", {
  c1 <- treestats::blum(extant_tree, normalization = FALSE)
  c2 <- treestats::blum(extant_tree, normalization = TRUE)
  testthat::expect_lt(c2, c1)
  c3 <- treestats::blum(treestats::phylo_to_l(extant_tree),
                        normalization = TRUE)
  testthat::expect_equal(c2, c3)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::blum(focal_tree))
    stats2 <- c(stats2, treestats::blum(focal_tree, normalization = TRUE))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_lt(a2, 0.5)
})

#' Branching times of a tree
#' @param phy phylo object or ltable
#' @return vector of branching times
#' @description C++ based alternative to
#' \link[ape:branching.times]{ape::branching.times}.
#' @export
branching_times <- function(phy) {

  if (inherits(phy, "matrix")) {
    return(branching_times_ltable_cpp(phy))
  }

  if (inherits(phy, "phylo")) {
    answ <- branching_times_cpp(phy)
    n <- length(phy$tip.label)
    if (is.null(phy$node.label)) {
       names(answ) <- (n + 1):(n + phy$Nnode)
    } else {
       names(answ) <- phy$node.label
    }
    return(answ)
  }
  stop("input object has to be phylo or ltable")
}

context("branching_times")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 500, birth = 1, death = 0)

  a1 <- ape::branching.times(focal_tree)
  a2 <- treestats::branching_times(focal_tree)

  testthat::expect_equal(a1, a2)

  # again, but with extinct lineages:

  a1 <- ape::branching.times(complete_tree)

  # oddly enough, ape::branching.times does not rescale correctly:
  if (min(a1) < 0) {
    a1 <- a1 - min(a1)
  }

  a2 <- treestats::branching_times(complete_tree)

  diff_brts <- abs(sort(a1) - sort(a2))
  mean_diff <- mean(diff_brts)
  testthat::expect_lte(mean_diff, min(a2))

  # using ltable
  l1 <- treestats::phylo_to_l(complete_tree)
  a3 <- treestats::branching_times(l1)
  testthat::expect_true(sum(a2 - a3) == 0)

  # with node labels:
  set.seed(42)
  focal_tree <- ape::rphylo(n = 5, birth = 1, death = 0)
  focal_tree$node.label <- 4:1
  b1 <- treestats::branching_times(focal_tree)
  b2 <- ape::branching.times(focal_tree)
  testthat::expect_true(all.equal(b1, b2))
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::branching_times(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::branching_times(list()),
    "input object has to be phylo or ltable"
  )
})

#' Total internal path length
#' @description The total internal path length describes the sums of the depths
#' of all inner vertices of the tree.
#' @param phy phylo object or ltable
#' @return Total internal path length
#' @references  Knuth, Donald E. The Art of Computer Programming:
#' Fundamental Algorithms, volume 1. Addison-Wesley Professional, 1997.
#' @export
tot_internal_path <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {
    t_i_p <- tot_internal_path_cpp(as.vector(t(phy$edge)))

    return(t_i_p)
  }
  stop("input object has to be phylo or ltable")
}

context("tot_internal_path")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    standard_test(treestats::tot_internal_path, treebalance::totIntPathLen)

    test_polytomies(treestats::tot_internal_path, treebalance::totIntPathLen)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::tot_internal_path(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::tot_internal_path(list()),
    "input object has to be phylo or ltable"
  )
})

#' Eigen vector centrality
#' @description Eigen vector centrality associates with each node \eqn{v}
#' the positive value \eqn{e(v)}, such that:
#' \eqn{ \sum_{e}^v w(uv) * e(u) = \lambda * e(v) }. Thus,
#' \eqn{e(v)} is the Perron-Frobenius eigenvector of the adjacency matrix of the
#' tree.
#' @param phy phylo object or ltable
#' @param weight if TRUE, uses branch lengths.
#' @param scale if TRUE, the Eigenvector is rescaled
#' @param use_rspectra boolean to indicate whether the helping package RSpectra
#' should be used, which is faster, but returns fewer eigen values.
#' @return List with the Eigen vector and the leading Eigen value
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." Plos one 16.12 (2021): e0259877.
#' @export
eigen_centrality <- function(phy,
                             weight = TRUE,
                             scale = FALSE,
                             use_rspectra = TRUE) {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {

    if (use_rspectra == FALSE || !requireNamespace("RSpectra")) {
      ev <- slow_eigen_centrality(phy, weight)
    } else {
      edges   <- phy$edge
      lengths <- phy$edge.length
      if (!weight) {
        lengths <- rep(1, length(phy$edge.length))
      }
      n_nodes <- max(edges)

      matvec_fun <- function(x, args) {
        Ax_tree(edges, lengths, x, n_nodes)
      }

      mat_size <- max(phy$edge)
      ev <- RSpectra::eigs_sym(matvec_fun,
                               n = mat_size,
                               k = 1,
                               which = "LM",
                               opts = list(retvec = TRUE))
    }

    evector <- abs(ev$vectors[, 1])

    evalue <- abs(ev$values[1])

    if (scale) {
      evector <- evector / max(evector)
    }

    return(list(eigenvector = evector,
                eigenvalue = evalue))
  }

  stop("input object has to be phylo or ltable")
}


slow_eigen_centrality <- function(phy, weight) {
  adj_matrix <- prep_adj_mat(as.vector(t(phy$edge)),
                             as.vector(phy$edge.length),
                             weight)
  ev <- eigen(adj_matrix, symmetric = TRUE)
  return(ev)
}

context("eigenvector")

test_that("usage", {
  a1_1 <- treestats::eigen_centrality(extant_tree,
                                      weight = TRUE, scale = FALSE)
  a1_2 <- treestats::eigen_centrality(extant_tree,
                                      weight = FALSE, scale = FALSE)

  # because treeCentrality is not available on CRAN, we precompute reference
  # values:
  a2_1 <- 3.585569  # treeCentrality::computeEigenvector(focal_tree
  #                                      weight = TRUE))
  a2_2 <- 2.609486  # treeCentrality::computeEigenvector(focal_tree
  #                                      weight = FALSE))

  testthat::expect_equal(a1_1$eigenvalue, a2_1, tolerance = 1e-4)
  testthat::expect_equal(a1_2$eigenvalue, a2_2, tolerance = 1e-4)
  a1_3 <- treestats::eigen_centrality(extant_tree, weight = TRUE, scale = TRUE)
  testthat::expect_equal(a1_1$eigenvalue,
                         a1_3$eigenvalue)
  testthat::expect_equal(max(a1_3$eigenvector), 1)

  ltab <- treestats::phylo_to_l(extant_tree)
  testthat::expect_equal(
    treestats::eigen_centrality(extant_tree, weight = TRUE)$eigenvalue,
    treestats::eigen_centrality(ltab, weight = TRUE)$eigenvalue)

  a1_1 <- treestats::eigen_centrality(complete_tree, weight = TRUE)
  a1_2 <- treestats::eigen_centrality(complete_tree, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:

  a2_1 <- 3.718206  # treeCentrality::computeEigenvector(focal_tree
  #                                      weight = TRUE))
  a2_2 <- 2.631605  # treeCentrality::computeEigenvector(focal_tree
  #                                      weight = FALSE))

  testthat::expect_equal(a1_1$eigenvalue, a2_1, tolerance = 1e-4)
  testthat::expect_equal(a1_2$eigenvalue, a2_2, tolerance = 1e-4)

  if (requireNamespace("RSpectra")) {
    a1_1 <- treestats::eigen_centrality(complete_tree, weight = TRUE,
                                        use_rspectra = TRUE)
    a1_2 <- treestats::eigen_centrality(complete_tree, weight = FALSE,
                                        use_rspectra = TRUE)
    testthat::expect_equal(a1_1$eigenvalue, a2_1, tolerance = 1e-4)
    testthat::expect_equal(a1_2$eigenvalue, a2_2, tolerance = 1e-4)
  }

  ltab <- treestats::phylo_to_l(complete_tree)
  testthat::expect_equal(
    treestats::eigen_centrality(complete_tree, weight = TRUE)$eigenvalue,
    treestats::eigen_centrality(ltab,          weight = TRUE)$eigenvalue)

  if (requireNamespace("igraph")) {
    df <- as.data.frame(cbind(extant_tree$edge,
                              weight = extant_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)

    ref <- igraph::eigen_centrality(g)

    a1 <- treestats::eigen_centrality(extant_tree)

    testthat::expect_equal(a1$eigenvalue, ref$value)

    ltab <- treestats::phylo_to_l(extant_tree)
    a1_2 <- treestats::eigen_centrality(ltab)
    testthat::expect_equal(a1_2$eigenvalue, ref$value, tolerance = 0.01)
  }

  # compare namespaces
  if (requireNamespace("Matrix")) {
    a1_1 <- treestats::eigen_centrality(extant_tree,
                                        weight = TRUE,
                                        scale = FALSE)
    a2_1 <- treestats::eigen_centrality(extant_tree,
                                        weight = FALSE,
                                        scale = FALSE)

    testthat::with_mocked_bindings(
      {
        # Now `myfun()` should behave as if `data.tree` is not installed
        a1_2 <- treestats::eigen_centrality(extant_tree,
                                            weight = TRUE,
                                            scale = FALSE)
        testthat::expect_equal(a1_1, a1_2)

        a1_3 <- treestats::eigen_centrality(extant_tree,
                                            weight = FALSE,
                                            scale = FALSE)
        testthat::expect_equal(a2_1, a1_3)
      },
      requireNamespace = function(pkg, quietly = TRUE) {
        if (pkg == "Matrix") {
          return(FALSE)
        }
        if (pkg == "RSpectra") {
          return(FALSE)
        }
        # Call the real `requireNamespace` for other packages
        base::requireNamespace(pkg, quietly = TRUE)
      },
      .package = "base"
    )
  }
})

test_that("polytomies", {
  if (requireNamespace("igraph")) {

    test_func <- function(phy) {
      df <- as.data.frame(cbind(phy$edge,
                                weight = phy$edge.length))
      g <- igraph::graph_from_data_frame(df, directed = FALSE)
      ref <- igraph::eigen_centrality(g)
      return(ref$value)
    }

    treestats_func <- function(phy) {
      res <- treestats::eigen_centrality(phy)
      return(res$eigenvalue)
    }

    test_polytomies(treestats_func, test_func)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::eigen_centrality(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::eigen_centrality(list()),
    "input object has to be phylo or ltable"
  )
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

calc_beta_cpp <- function(phy, upper_lim, algorithm, abs_tol, rel_tol) {
    .Call('_treestats_calc_beta_cpp', PACKAGE = 'treestats', phy, upper_lim, algorithm, abs_tol, rel_tol)
}

calc_beta_ltable_cpp <- function(ltable, upper_lim, algorithm, abs_tol, rel_tol) {
    .Call('_treestats_calc_beta_ltable_cpp', PACKAGE = 'treestats', ltable, upper_lim, algorithm, abs_tol, rel_tol)
}

branching_times_cpp <- function(phy) {
    .Call('_treestats_branching_times_cpp', PACKAGE = 'treestats', phy)
}

branching_times_ltable_cpp <- function(mat_in) {
    .Call('_treestats_branching_times_ltable_cpp', PACKAGE = 'treestats', mat_in)
}

calc_rho_complete_cpp <- function(phy) {
    .Call('_treestats_calc_rho_complete_cpp', PACKAGE = 'treestats', phy)
}

calc_rho_cpp <- function(phy) {
    .Call('_treestats_calc_rho_cpp', PACKAGE = 'treestats', phy)
}

calc_rho_ltable_cpp <- function(ltab) {
    .Call('_treestats_calc_rho_ltable_cpp', PACKAGE = 'treestats', ltab)
}

calc_phylodiv_cpp <- function(phy, t, extinct_acc) {
    .Call('_treestats_calc_phylodiv_cpp', PACKAGE = 'treestats', phy, t, extinct_acc)
}

calc_phylodiv_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_calc_phylodiv_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

calc_mean_branch_length_cpp <- function(edge_length) {
    .Call('_treestats_calc_mean_branch_length_cpp', PACKAGE = 'treestats', edge_length)
}

calc_mean_branch_length_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_calc_mean_branch_length_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

calc_gamma_cpp <- function(phy) {
    .Call('_treestats_calc_gamma_cpp', PACKAGE = 'treestats', phy)
}

calc_gamma_ltable_cpp <- function(ltab_in) {
    .Call('_treestats_calc_gamma_ltable_cpp', PACKAGE = 'treestats', ltab_in)
}

calc_nltt_cpp <- function(phy1, phy2) {
    .Call('_treestats_calc_nltt_cpp', PACKAGE = 'treestats', phy1, phy2)
}

calc_nltt_ltable_cpp <- function(ltab1, ltab2) {
    .Call('_treestats_calc_nltt_ltable_cpp', PACKAGE = 'treestats', ltab1, ltab2)
}

calc_crown_age_cpp <- function(phy) {
    .Call('_treestats_calc_crown_age_cpp', PACKAGE = 'treestats', phy)
}

calc_wiener_cpp <- function(edge_in, el, normalize, weight) {
    .Call('_treestats_calc_wiener_cpp', PACKAGE = 'treestats', edge_in, el, normalize, weight)
}

calc_max_betweenness_cpp <- function(edge_in, el) {
    .Call('_treestats_calc_max_betweenness_cpp', PACKAGE = 'treestats', edge_in, el)
}

calc_max_betweenness_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_max_betweenness_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_max_closeness_cpp <- function(edge_in, el, weight) {
    .Call('_treestats_calc_max_closeness_cpp', PACKAGE = 'treestats', edge_in, el, weight)
}

calc_diameter_cpp <- function(edge_in, el, weight) {
    .Call('_treestats_calc_diameter_cpp', PACKAGE = 'treestats', edge_in, el, weight)
}

calc_diameter_ltable_cpp <- function(l_from_R, weight) {
    .Call('_treestats_calc_diameter_ltable_cpp', PACKAGE = 'treestats', l_from_R, weight)
}

calc_colless_cpp <- function(parent_list, normalization) {
    .Call('_treestats_calc_colless_cpp', PACKAGE = 'treestats', parent_list, normalization)
}

calc_colless_ltable_cpp <- function(l_from_R, normalization) {
    .Call('_treestats_calc_colless_ltable_cpp', PACKAGE = 'treestats', l_from_R, normalization)
}

calc_eWcolless_cpp <- function(parent_list) {
    .Call('_treestats_calc_eWcolless_cpp', PACKAGE = 'treestats', parent_list)
}

calc_eWcolless_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_eWcolless_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

ILnumber_cpp <- function(tree_edge) {
    .Call('_treestats_ILnumber_cpp', PACKAGE = 'treestats', tree_edge)
}

ILnumber_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_ILnumber_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

calc_rquartet_cpp <- function(tree_edge) {
    .Call('_treestats_calc_rquartet_cpp', PACKAGE = 'treestats', tree_edge)
}

calc_rquartet_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_calc_rquartet_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

stairs_cpp <- function(tree_edge) {
    .Call('_treestats_stairs_cpp', PACKAGE = 'treestats', tree_edge)
}

stairs_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_stairs_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

stairs2_cpp <- function(tree_edge) {
    .Call('_treestats_stairs2_cpp', PACKAGE = 'treestats', tree_edge)
}

stairs2_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_stairs2_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

calc_rogers_cpp <- function(parent_list) {
    .Call('_treestats_calc_rogers_cpp', PACKAGE = 'treestats', parent_list)
}

calc_rogers_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_rogers_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_j_one_cpp <- function(parent_list) {
    .Call('_treestats_calc_j_one_cpp', PACKAGE = 'treestats', parent_list)
}

calc_j_one_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_j_one_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_Ibased_cpp <- function(parent_list) {
    .Call('_treestats_calc_Ibased_cpp', PACKAGE = 'treestats', parent_list)
}

calc_Ibased_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_Ibased_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

pitchforks_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_pitchforks_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

calc_colless_corr_cpp <- function(parent_list, normalization) {
    .Call('_treestats_calc_colless_corr_cpp', PACKAGE = 'treestats', parent_list, normalization)
}

calc_colless_corr_ltable_cpp <- function(l_from_R, normalization) {
    .Call('_treestats_calc_colless_corr_ltable_cpp', PACKAGE = 'treestats', l_from_R, normalization)
}

calc_colless_quad_cpp <- function(parent_list, normalization) {
    .Call('_treestats_calc_colless_quad_cpp', PACKAGE = 'treestats', parent_list, normalization)
}

calc_colless_quad_ltable_cpp <- function(l_from_R, normalization) {
    .Call('_treestats_calc_colless_quad_ltable_cpp', PACKAGE = 'treestats', l_from_R, normalization)
}

calc_root_imbalance_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_root_imbalance_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_double_cherries_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_double_cherries_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_four_prong_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_four_prong_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

check_is_binary_rcpp <- function(parent_list) {
    .Call('_treestats_check_is_binary_rcpp', PACKAGE = 'treestats', parent_list)
}

calc_max_del_width_cpp <- function(parent_list) {
    .Call('_treestats_calc_max_del_width_cpp', PACKAGE = 'treestats', parent_list)
}

calc_max_del_width_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_max_del_width_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_max_width_cpp <- function(parent_list) {
    .Call('_treestats_calc_max_width_cpp', PACKAGE = 'treestats', parent_list)
}

calc_max_width_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_max_width_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_max_depth_cpp <- function(parent_list) {
    .Call('_treestats_calc_max_depth_cpp', PACKAGE = 'treestats', parent_list)
}

calc_avg_vert_depth_cpp <- function(parent_list) {
    .Call('_treestats_calc_avg_vert_depth_cpp', PACKAGE = 'treestats', parent_list)
}

tot_internal_path_cpp <- function(parent_list) {
    .Call('_treestats_tot_internal_path_cpp', PACKAGE = 'treestats', parent_list)
}

calc_max_depth_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_max_depth_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_var_leaf_depth_cpp <- function(parent_list) {
    .Call('_treestats_calc_var_leaf_depth_cpp', PACKAGE = 'treestats', parent_list)
}

calc_var_leaf_depth_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_var_leaf_depth_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_sym_nodes_cpp <- function(parent_list) {
    .Call('_treestats_calc_sym_nodes_cpp', PACKAGE = 'treestats', parent_list)
}

calc_b1_cpp <- function(parent_list) {
    .Call('_treestats_calc_b1_cpp', PACKAGE = 'treestats', parent_list)
}

calc_b1_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_b1_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_b2_cpp <- function(parent_list) {
    .Call('_treestats_calc_b2_cpp', PACKAGE = 'treestats', parent_list)
}

calc_b2_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_b2_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_mw_over_md_cpp <- function(parent_list) {
    .Call('_treestats_calc_mw_over_md_cpp', PACKAGE = 'treestats', parent_list)
}

calc_mw_over_md_ltable_cpp <- function(l_from_R) {
    .Call('_treestats_calc_mw_over_md_ltable_cpp', PACKAGE = 'treestats', l_from_R)
}

calc_root_imbalance_cpp <- function(parent_list) {
    .Call('_treestats_calc_root_imbalance_cpp', PACKAGE = 'treestats', parent_list)
}

avgLadder_cpp <- function(tree_edge) {
    .Call('_treestats_avgLadder_cpp', PACKAGE = 'treestats', tree_edge)
}

max_ladder_cpp <- function(tree_edge) {
    .Call('_treestats_max_ladder_cpp', PACKAGE = 'treestats', tree_edge)
}

get_eigen_values_arma_cpp <- function(phy) {
    .Call('_treestats_get_eigen_values_arma_cpp', PACKAGE = 'treestats', phy)
}

prep_adj_mat <- function(parent_list, br_len, use_br_len) {
    .Call('_treestats_prep_adj_mat', PACKAGE = 'treestats', parent_list, br_len, use_br_len)
}

outer_cpp <- function(xx, x, sd) {
    .Call('_treestats_outer_cpp', PACKAGE = 'treestats', xx, x, sd)
}

calc_mpd_cpp <- function(edge, el) {
    .Call('_treestats_calc_mpd_cpp', PACKAGE = 'treestats', edge, el)
}

calc_var_mpd_cpp <- function(phy) {
    .Call('_treestats_calc_var_mpd_cpp', PACKAGE = 'treestats', phy)
}

calc_mntd_cpp <- function(phy) {
    .Call('_treestats_calc_mntd_cpp', PACKAGE = 'treestats', phy)
}

calc_mntd_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_calc_mntd_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

calc_inv_path_cpp <- function(edge, el, add_one) {
    .Call('_treestats_calc_inv_path_cpp', PACKAGE = 'treestats', edge, el, add_one)
}

#' Function to generate an ltable from a phy object.
#' @description This function is a C++
#' implementation of the function DDD::phylo2L. An L table summarises a
#' phylogeny in a table with four columns, being: 1) time at which a species
#' is born, 2) label of the parent of the species, where positive and negative
#' numbers indicate whether the species belongs to the left or right crown
#' lineage, 3) label of the daughter species itself (again positive or
#' negative depending on left or right crown lineage), and the last column 4)
#' indicates the time of extinction of a species, or -1 if the species is
#' extant.
#' @param phy phylo object
#' @return ltable (see description)
#' @export
#' @examples
#' simulated_tree <- ape::rphylo(n = 4, birth = 1, death = 0)
#' ltable <- phylo_to_l(simulated_tree)
#' reconstructed_tree <- DDD::L2phylo(ltable)
#' old_par <- par()
#' par(mfrow = c(1, 2))
#' # trees should be more or less similar, although labels may not match, and
#' # rotations might cause (initial) visual mismatches
#' plot(simulated_tree)
#' plot(reconstructed_tree)
#' par(old_par)
phylo_to_l <- function(phy) {
    .Call('_treestats_phylo_to_l', PACKAGE = 'treestats', phy)
}

l_to_newick <- function(ltable_R, drop_extinct) {
    .Call('_treestats_l_to_newick', PACKAGE = 'treestats', ltable_R, drop_extinct)
}

imbalance_steps_cpp <- function(ltable_R, normalization) {
    .Call('_treestats_imbalance_steps_cpp', PACKAGE = 'treestats', ltable_R, normalization)
}

Ax_tree <- function(edge, lengths, x, nNodes) {
    .Call('_treestats_Ax_tree', PACKAGE = 'treestats', edge, lengths, x, nNodes)
}

Lx_tree_weighted <- function(edge, w, x, nNodes) {
    .Call('_treestats_Lx_tree_weighted', PACKAGE = 'treestats', edge, w, x, nNodes)
}

calc_sackin_cpp <- function(tree_edge, normalization) {
    .Call('_treestats_calc_sackin_cpp', PACKAGE = 'treestats', tree_edge, normalization)
}

calc_sackin_ltable_cpp <- function(ltab, normalization) {
    .Call('_treestats_calc_sackin_ltable_cpp', PACKAGE = 'treestats', ltab, normalization)
}

calc_tot_coph_cpp <- function(tree_edge) {
    .Call('_treestats_calc_tot_coph_cpp', PACKAGE = 'treestats', tree_edge)
}

calc_tot_coph_ltable_cpp <- function(ltab) {
    .Call('_treestats_calc_tot_coph_ltable_cpp', PACKAGE = 'treestats', ltab)
}

calc_blum_cpp <- function(tree_edge, normalize) {
    .Call('_treestats_calc_blum_cpp', PACKAGE = 'treestats', tree_edge, normalize)
}

calc_blum_ltable_cpp <- function(ltab_in, normalize) {
    .Call('_treestats_calc_blum_ltable_cpp', PACKAGE = 'treestats', ltab_in, normalize)
}

cherries_cpp <- function(tree_edge) {
    .Call('_treestats_cherries_cpp', PACKAGE = 'treestats', tree_edge)
}

calc_double_cherries_cpp <- function(tree_edge) {
    .Call('_treestats_calc_double_cherries_cpp', PACKAGE = 'treestats', tree_edge)
}

calc_four_prong_cpp <- function(tree_edge) {
    .Call('_treestats_calc_four_prong_cpp', PACKAGE = 'treestats', tree_edge)
}

cherries_ltable_cpp <- function(ltable_R) {
    .Call('_treestats_cherries_ltable_cpp', PACKAGE = 'treestats', ltable_R)
}

pitchforks_cpp <- function(tree_edge) {
    .Call('_treestats_pitchforks_cpp', PACKAGE = 'treestats', tree_edge)
}

context("colless_quad")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    colless_c <- treestats::colless_quad(extant_tree)
    colless_check <- treebalance::collessI(extant_tree, method = "quadratic")
    testthat::expect_equal(colless_c, colless_check)

    # now, using ltable:
    focal_ltab <- treestats::phylo_to_l(extant_tree)

    colless_cl <- treestats::colless_quad(focal_ltab)
    testthat::expect_equal(colless_cl, colless_check)

    ## with extinct lineages:

    colless_c <- treestats::colless_quad(complete_tree)
    colless_check <- treebalance::collessI(complete_tree, method = "quadratic")
    testthat::expect_equal(colless_c, colless_check)

    focal_ltab <- treestats::phylo_to_l(complete_tree)

    colless_cl <- treestats::colless_quad(focal_ltab)
    testthat::expect_equal(colless_cl, colless_check)
  }
})

test_that("polytomies", {
  if (requireNamespace("treebalance")) {

    test_func <- function(phy) {
      treebalance::collessI(phy, method = "quadratic")
    }

    test_polytomies(treestats::colless_quad, test_func)
  }
})

test_that("yule", {
  testthat::skip_on_cran()
  found <- c()
  found2 <- c()
  for (r in 1:100) {
    focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
    found[r] <- treestats::colless_quad(focal_tree, "yule")
    found2[r] <- treestats::colless_quad(treestats::phylo_to_l(focal_tree),
                                         "yule")
  }
  testthat::expect_equal(mean(found), 1.0, tolerance = 0.1)
  testthat::expect_equal(mean(found), mean(found2))
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::colless_quad(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::colless_quad(list()),
    "input object has to be phylo or ltable"
  )
})

context("colless")

test_that("usage", {
  if (requireNamespace("apTreeshape")) {
     colless <- treestats::colless(extant_tree)
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(extant_tree))
    testthat::expect_equal(colless, colless_check)

    colless <- treestats::colless(extant_tree, normalization = "yule")
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(extant_tree),
                                          norm = "yule")
    testthat::expect_equal(colless, colless_check, tol = 1e-5)

    colless <- treestats::colless(extant_tree, normalization = "pda")
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(extant_tree), norm = "pda")
    testthat::expect_equal(colless, colless_check, tol = 1e-5)

    # now, using ltable:
    focal_ltab <- treestats::phylo_to_l(extant_tree)

    colless <- treestats::colless(focal_ltab)
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(extant_tree))
    testthat::expect_equal(colless, colless_check)

    colless <- treestats::colless(focal_ltab, normalization = "yule")
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(extant_tree),
                           norm = "yule")
    testthat::expect_equal(colless, colless_check, tol = 1e-5)

    colless <- treestats::colless(focal_ltab, normalization = "pda")
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(extant_tree), norm = "pda")
    testthat::expect_equal(colless, colless_check, tol = 1e-5)


    ## with extinct lineages:
    colless <- treestats::colless(complete_tree)
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(complete_tree))
    testthat::expect_equal(colless, colless_check)

    colless <- treestats::colless(complete_tree, normalization = "yule")
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(complete_tree),
                                          norm = "yule")
    testthat::expect_equal(colless, colless_check, tol = 1e-5)

    colless <- treestats::colless(complete_tree, normalization = "pda")
    colless_check <-
      apTreeshape::colless(apTreeshape::as.treeshape(complete_tree),
                           norm = "pda")
    testthat::expect_equal(colless, colless_check, tol = 1e-5)
  }
})

test_that("polytomies", {
  if (requireNamespace("apTreeshape")) {

    test_func <- function(phy) {
      apTreeshape::colless(apTreeshape::as.treeshape(phy))
    }

    test_polytomies(treestats::colless, test_func)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::colless(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::colless(list()),
    "input object has to be phylo or ltable"
  )
})

context("gamma")

test_that("usage", {
  if (requireNamespace("castor") &&
      requireNamespace("DDD")) {
    check_gamma <- function(focal_tree) {
      gammast <- treestats::gamma_statistic(focal_tree)
      gammast_check <- ape::gammaStat(focal_tree)
      gammast_check2 <- castor::gamma_statistic(focal_tree)

      ltab <- treestats::phylo_to_l(focal_tree)
      gammast_check4 <- treestats::gamma_statistic(ltab)

      testthat::expect_equal(gammast, gammast_check, tolerance = 1e-4)
      testthat::expect_equal(gammast, gammast_check2, tolerance = 1e-4)
      testthat::expect_equal(gammast, gammast_check4, tolerance = 1e-4)
    }

    set.seed(42)
    bd_tree <- ape::rphylo(n = 101, birth = 1, death = 0)

    check_gamma(bd_tree)

    # DDD tree has all speciation events at the beginning
    # expect strongly negative gamma.
    dd_tree <- DDD::dd_sim(pars = c(1, 0, 10), age = 20)$tas
    check_gamma(dd_tree)
    testthat::expect_lt(treestats::gamma_statistic(dd_tree), 0)
  }
})

test_that("polytomies", {
  for (focal_tree in poly_trees) {
    if (ape::is.ultrametric(focal_tree)) {
      res <- treestats::gamma_statistic(focal_tree)
      testthat::expect_true(!is.na(res))
    }
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::gamma_statistic(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::gamma_statistic(list()),
    "input object has to be phylo or ltable"
  )
})

context("inv_branch_dist")

test_that("usage", {
  # let's create an artifical tree for which we know the values
  set.seed(42)
  focal_tree <- ape::rphylo(n = 3, 1, 0)
  focal_tree <- stats::reorder(focal_tree)
  focal_tree$edge.length <- c(0.5, 1, 1, 1.5)
  expected_vals <- c(1 + 1 / 0.5, 1 + 1 / 0.5, 1 / 1.5)

  inv_dist <- treestats::inv_branch_dist(focal_tree, add_one = FALSE)
  testthat::expect_equal(as.vector(inv_dist), expected_vals)

  ltab <- treestats::phylo_to_l(focal_tree)
  inv_dist_ltab <- treestats::inv_branch_dist(ltab, add_one = FALSE)
  testthat::expect_equal(sum(as.vector(inv_dist_ltab)), sum(expected_vals))

  # now we add one
  expected_vals <- c(1 / (1 + 1) + 1 / 1.5, 1 / (1 + 1) + 1 / 1.5, 1 / 2.5)

  inv_dist <- treestats::inv_branch_dist(focal_tree, add_one = TRUE)
  testthat::expect_equal(as.vector(inv_dist), expected_vals)
})

test_that("polytomies", {
  local_stat <- max(treestats::inv_branch_dist(poly_trees[[3]]))
  testthat::expect_true(!is.na(local_stat))
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::inv_branch_dist(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::inv_branch_dist(list()),
    "input object has to be phylo or ltable"
  )
})

context("area per pair")

test_that("usage", {
  if (requireNamespace("treebalance")) {
    a1 <- treestats::area_per_pair(extant_tree)
    a2 <- treebalance::areaPerPairI(extant_tree)
    testthat::expect_equal(a1, a2)

    ltab <- treestats::phylo_to_l(extant_tree)
    testthat::expect_equal(treestats::area_per_pair(extant_tree),
                           treestats::area_per_pair(ltab))

    # with extinct species:
    a1 <- treestats::area_per_pair(complete_tree)
    a2 <- treebalance::areaPerPairI(complete_tree)
    testthat::expect_equal(a1, a2)

    ltab <- treestats::phylo_to_l(complete_tree)
    testthat::expect_equal(treestats::area_per_pair(complete_tree),
                           treestats::area_per_pair(ltab))

    # on polytomies
    test_polytomies(treestats::area_per_pair, treebalance::areaPerPairI)
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::area_per_pair(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::area_per_pair(list()),
    "input object has to be phylo or ltable"
  )
})

context("phylo2L")

test_that("usage", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 500, birth = 1, death = 0)

  ltable_1 <- DDD::phylo2L(focal_tree)
  ltable_2 <- treestats::phylo_to_l(focal_tree)

  diff <- ltable_1 - ltable_2

  for (i in 1:4) {
    testthat::expect_lt(mean(diff[, i]), 1e-6)
  }

  # again, but with extinct lineages:
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0.3, fossils = TRUE)

  ltable_1 <- DDD::phylo2L(focal_tree)
  ltable_2 <- treestats::phylo_to_l(focal_tree)

  diff <- ltable_1 - ltable_2
  for (i in 1:4) {
    testthat::expect_lt(mean(diff[, i]), 1e-6)
  }
})

context("ew_colless")

test_that("usage", {
  if (requireNamespace("treebalance")) {

    colless <- treestats::ew_colless(extant_tree)
    colless_check <- treebalance::ewCollessI(extant_tree)
    testthat::expect_equal(colless, colless_check)

    focal_ltab <- treestats::phylo_to_l(extant_tree)

    colless <- treestats::ew_colless(focal_ltab)
    testthat::expect_equal(colless, colless_check)

    ## with extinct lineages:
    colless <- treestats::ew_colless(complete_tree)
    colless_check <- treebalance::ewCollessI(complete_tree)
    testthat::expect_equal(colless, colless_check)

    focal_ltab <- treestats::phylo_to_l(complete_tree)

    colless <- treestats::ew_colless(focal_ltab)
    testthat::expect_equal(colless, colless_check)
  }
})

test_that("polytomies", {
  if (requireNamespace("treebalance")) {
    test_polytomies(treestats::ew_colless, treebalance::ewCollessI)
  }
})


test_that("wrong_object", {
  testthat::expect_error(
    treestats::ew_colless(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::ew_colless(list()),
    "input object has to be phylo or ltable"
  )
})

context("betweenness")

test_that("usage", {
  a1_1 <- treestats::max_betweenness(extant_tree)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:

  a2_1 <- 12795  # max(treeCentrality::computeBetweenness(focal_tree)) #nolint

  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)

  if (requireNamespace("igraph")) {
    df <- as.data.frame(cbind(extant_tree$edge,
                              weight = extant_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)
    ref_betweenness <- igraph::betweenness(g)
    testthat::expect_equal(a1_1, max(ref_betweenness), tolerance = 0.01)
  }

  ltab <- treestats::phylo_to_l(extant_tree)
  testthat::expect_equal(treestats::max_betweenness(extant_tree),
                         treestats::max_betweenness(ltab))

  a1_1 <- treestats::max_betweenness(complete_tree)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:
  a2_1 <- 20315 # max(treeCentrality::computeBetweenness(focal_tree)) #nolint
  if (requireNamespace("igraph")) {
    df <- as.data.frame(cbind(complete_tree$edge,
                              weight = complete_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)
    ref_betweenness <- igraph::betweenness(g)
    testthat::expect_equal(a1_1, max(ref_betweenness), tolerance = 0.01)
  }

  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)

  ltab <- treestats::phylo_to_l(complete_tree)
  testthat::expect_equal(treestats::max_betweenness(complete_tree),
                         treestats::max_betweenness(ltab))
})

test_that("polytomies", {
  if (requireNamespace("igraph")) {

    test_func <- function(phy) {
      df <- as.data.frame(cbind(phy$edge,
                                weight = phy$edge.length))
      g <- igraph::graph_from_data_frame(df, directed = FALSE)
      ref_betweenness <- igraph::betweenness(g)
      return(max(ref_betweenness))
    }

    # TODO: fix C++ error
    test_polytomies(treestats::max_betweenness, test_func)
  }
})


test_that("normalization", {
  if (requireNamespace("igraph")) {
    set.seed(42)
    focal_tree <- ape::rphylo(n = 30, birth = 1, death = 0)
    a1_1 <- treestats::max_betweenness(focal_tree, normalization = "tips")

    df <- as.data.frame(cbind(focal_tree$edge,
                             weight = focal_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)

    ref_betweenness <- igraph::betweenness(g, normalized = TRUE)
    testthat::expect_equal(a1_1, max(ref_betweenness), tolerance = 0.01)

    ltab <- treestats::phylo_to_l(focal_tree)
    a1_2 <- treestats::max_betweenness(ltab, normalization = "tips")
    testthat::expect_equal(a1_2, max(ref_betweenness), tolerance = 0.01)
  }
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::max_betweenness(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::max_betweenness(list()),
    "input object has to be phylo or ltable"
  )
})

poly_text <- c("((t6:1,t2:1)1:0.5,(((t15:1,t8:1,t11:1)1:1,t12:1,t13:1,t10:1)1:1,((t9:1,(t4:1,t14:1)1:1)1:1,t1:1)1:1,(t3:1,t7:1,t5:1)1:1)1:0.5);", #nolint
                "(((t9:1,t14:1)1:1,t4:1,t12:1)1:0.5,(((t6:1,t19:1,t10:1,t2:1)1:1,t8:1,(t5:1,t16:1)1:1)1:1,(((((t11:1,t17:1)1:1,(t7:1,t3:1)1:1)1:1,(t1:1,t18:1)1:1)1:1,t15:1)1:1,t13:1)1:1)1:0.5);", #nolint
                "((t1:2,t2:2,t3:2):1,((t4:1,t5:1):1,(t6:1,t7:1,t8:1,(t9:0.5,t10:0.5):0.5):1):1);") #nolint  ultrametric tree with polytomies

poly_trees <- list()
for (i in seq_along(poly_text)) {
  poly_trees[[i]] <- ape::read.tree(text = poly_text[i])
}

set.seed(42)
extant_tree   <- ape::rphylo(n = 100, birth = 1, death = 0)
complete_tree <- ape::rphylo(n = 100, birth = 1, death = 0.2, fossils = TRUE)

#' @keywords internal
test_stat <- function(phylo, func) {
  res <- tryCatch(expr = {func(phylo)},  #nolint
                  error = function(e) {return(NA_real_)})     #nolint
  return(res)
}


test_polytomies <- function(treestats_func, ref_func, tol = NA) {
  for (focal_tree in poly_trees) {
    a1 <- test_stat(focal_tree, treestats_func)
    a2 <- test_stat(focal_tree, ref_func)
    if (is.na(tol)) {
      testthat::expect_equal(a1, a2)
    } else {
      testthat::expect_equal(a1, a2, tolerance = tol)
    }
  }
}

standard_test <- function(treestats_func,
                          ref_func,
                          drop_complete = FALSE) {
  a1 <- treestats_func(extant_tree)
  a2 <- ref_func(extant_tree)
  testthat::expect_equal(a1, a2)

  ltab <- treestats::phylo_to_l(extant_tree)
  testthat::expect_equal(treestats_func(extant_tree),
                         treestats_func(ltab))

  if (drop_complete == FALSE) {
    # with extinct species:
    a1 <- treestats_func(complete_tree)
    a2 <- ref_func(complete_tree)
    testthat::expect_equal(a1, a2)

    ltab <- treestats::phylo_to_l(complete_tree)
    testthat::expect_equal(treestats_func(complete_tree),
                           treestats_func(ltab))
  }
}

context("max_closeness")

test_that("usage", {
  a1_1 <- treestats::max_closeness(extant_tree, weight = TRUE)
  a1_2 <- treestats::max_closeness(extant_tree, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:

  a2_1 <- 0.001400582  # max(treeCentrality::computeCloseness(focal_tree,
                       #                                      weight = TRUE))
  a2_2 <- 0.0007843137 # max(treeCentrality::computeCloseness(focal_tree,
                       #                                      weight = FALSE))

  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)
  testthat::expect_equal(a1_2, a2_2)

  ltab <- treestats::phylo_to_l(extant_tree)
  testthat::expect_equal(treestats::max_closeness(extant_tree, weight = TRUE),
                         treestats::max_closeness(ltab, weight = TRUE))

  a1_1 <- treestats::max_closeness(complete_tree, weight = TRUE)
  a1_2 <- treestats::max_closeness(complete_tree, weight = FALSE)
  # because treeCentrality is not available on CRAN, we precompute reference
  # values:

  a2_1 <- 0.001018909  # max(treeCentrality::computeCloseness(focal_tree,
  #                                      weight = TRUE))
  a2_2 <- 0.0006035003 # max(treeCentrality::computeCloseness(focal_tree,
  #                                      weight = FALSE))


  testthat::expect_equal(a1_1, a2_1, tolerance = 0.01)
  testthat::expect_equal(a1_2, a2_2)

  ltab <- treestats::phylo_to_l(complete_tree)
  testthat::expect_equal(treestats::max_closeness(complete_tree, weight = TRUE),
                         treestats::max_closeness(ltab,       weight = TRUE))

  if (requireNamespace("igraph")) {
   df <- as.data.frame(cbind(extant_tree$edge,
                             weight = extant_tree$edge.length))
    g <- igraph::graph_from_data_frame(df, directed = FALSE)

    ref <- igraph::closeness(g)

    a1 <- treestats::max_closeness(extant_tree)

    testthat::expect_equal(a1, max(ref), tolerance = 0.01)

    ltab <- treestats::phylo_to_l(extant_tree)
    a1_2 <- treestats::max_closeness(ltab)
    testthat::expect_equal(a1_2, max(ref), tolerance = 0.01)
  }
})


test_that("polytomies", {
  if (requireNamespace("igraph")) {

    test_func <- function(phy) {
      df <- as.data.frame(cbind(phy$edge,
                                weight = phy$edge.length))
      g <- igraph::graph_from_data_frame(df, directed = FALSE)
      ref_closeness <- igraph::closeness(g)
      return(max(ref_closeness))
    }


    test_polytomies(treestats::max_closeness, test_func)
  } else if (requireNamespace("ape")) {
    test_func <- function(phy) {
      vv <-  ape::dist.nodes(phy)
      farness <- rowSums(vv)
      closeness <- 1 / farness
      return(max(closeness))
    }
    test_polytomies(treestats::max_closeness, test_func)
  }
})


test_that("normalization", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 30, birth = 1, death = 0)

  c1 <- treestats::max_closeness(focal_tree, weight = TRUE)
  c2 <- treestats::max_closeness(focal_tree,  weight = TRUE,
                                 normalization = "tips")
  testthat::expect_lt(c1, c2)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::max_closeness(focal_tree))
    stats2 <- c(stats2, treestats::max_closeness(focal_tree,
                                                 normalization = "tips"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(abs(a2), abs(a1))
  testthat::expect_lt(abs(a2), 0.5)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::max_closeness(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::max_closeness(list()),
    "input object has to be phylo or ltable"
  )
})

context("rebase_ltable")

test_that("usage", {
  set.seed(42)
  # this is a random tree that happens to trigger a rebase:
  focal_tree <- ape::rphylo(n = 20, birth = 1, death = 0)

  ltab <- treestats::phylo_to_l(focal_tree)

  ltab2 <- treestats::rebase_ltable(ltab)
  testthat::expect_true(all.equal(ltab, ltab2) != TRUE)

  # test small tree:
  focal_tree <- ape::rphylo(n = 2, birth = 1, death = 0)

  ltab <- treestats::phylo_to_l(focal_tree)

  focal_tree2 <- treestats::rebase_ltable(ltab)
  testthat::expect_true(all.equal(ltab, focal_tree2))
})

#' Gamma statistic
#' @description The gamma statistic measures the relative position of
#' internal nodes within a reconstructed phylogeny. Under the Yule process,
#' the gamma values of a reconstructed tree follow a standard normal
#' distribution. If gamma > 0, the nodes are located more towards the tips of
#' the tree, and if gamma < 0, the nodes are located more towards the root of
#' the tree. Only available for ultrametric trees.
#' @param phy phylo object or ltable
#' @return gamma statistic
#' @export
#' @references Pybus, O. G. and Harvey, P. H. (2000) Testing macro-evolutionary
#' models using incomplete molecular phylogenies. Proceedings of the Royal
#' Society of London. Series B. Biological Sciences, 267, 2267–2272.
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' gamma_statistic(simulated_tree) # should be around 0.
#' if (requireNamespace("DDD")) {
#'   ddd_tree <- DDD::dd_sim(pars = c(1, 0, 10), age = 7)$tes
#'   gamma_statistic(ddd_tree) # because of diversity dependence, should be < 0
#' }
gamma_statistic <- function(phy) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = TRUE,
             require_rooted = TRUE)

  if (inherits(phy, "phylo")) {
    return(calc_gamma_cpp(phy))
  }
  if (inherits(phy, "matrix")) {
    return(calc_gamma_ltable_cpp(phy))
  }

  stop("input object has to be phylo or ltable")
}

#' Maximum closeness
#' @description Closeness is defined as 1 / Farness, where Farness is the sum
#' of distances from a node to all the other nodes in the tree. Here, we return
#' the node with maximum closeness.
#' @param phy phylo object or ltable
#' @param weight if TRUE, uses branch lengths.
#' @param normalization "none" or "tips", in which case an arbitrary post-hoc
#' correction is performed by dividing by the expectation of n log(n), where n
#' is the number of tips.
#' @return Maximum Closeness
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." Plos one 16.12 (2021): e0259877.
#' Wang W, Tang CY. Distributed computation of classic and exponential closeness
#' on tree graphs. Proceedings of the American Control Conference. IEEE; 2014.
#' p. 2090–2095.
#' @export
max_closeness <- function(phy, weight = TRUE, normalization = "none") {
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  normalization <- check_normalization_key(normalization)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {

    if (ape::is.rooted(phy)) {
      closeness_stat <- calc_max_closeness_cpp(as.vector(t(phy$edge)),
                                               phy$edge.length,
                                               weight)
    } else {
      mat_size <- max(phy$edge)
      if (mat_size > 46340) {  # floor(sqrt(2^31 - 1)))
        stop("tree too big, memory allocation fail")
      }

      if (!weight) phy$edge.length <- rep(1, length(phy$edge.length))
      node_dist <- ape::dist.nodes(phy)
      closeness_stat <- max(1 / rowSums(node_dist))
    }

    if (normalization == "tips" || normalization == TRUE) {
      n <- length(phy$tip.label)
      expectation <- 1.0 / (n * log(n))
      closeness_stat <- closeness_stat / expectation
    }
    return(closeness_stat)
  }
  stop("input object has to be phylo or ltable")
}

#' ILnumber
#' @description The ILnumber is the number of internal nodes with a
#' single tip child. Higher values typically indicate a tree that
#' is more unbalanced.
#' @param input_obj phylo object or ltable
#' @param normalization "none" or "tips", in which case the result is normalized
#' by dividing by N - 2, where N is the number of tips.
#' @return ILnumber
#' @export
ILnumber <- function(input_obj, normalization = "none") { # nolint
  normalization <- check_normalization_key(normalization)
  check_tree(input_obj,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(input_obj, "matrix")) {
    il_stat <- ILnumber_ltable_cpp(input_obj)
    if (normalization == "tips" || normalization == TRUE) {
      n <- length(input_obj[, 1])
      il_stat <- il_stat / (n - 2)
    }
    return(il_stat)
  }
  if (inherits(input_obj, "phylo")) {
    il_stat <- ILnumber_cpp(as.vector(t(input_obj$edge)))
    if (normalization == "tips" || normalization == TRUE) {
      n <- length(input_obj$tip.label)
      il_stat <- il_stat / (n - 2)
    }
    return(il_stat)
  }

  stop("input object has to be phylo or ltable")
}

#' Double Cherry index
#' @description Calculate the number of double cherries, where a single cherry
#' is a node connected to two tips, and a double cherry is a node connected
#' to two cherry nodes.
#' @param input_obj phylo object or ltable
#' @return number of cherries
#' @references Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." PloS one 16.12 (2021): e0259877.
#' @export
double_cherries <- function(input_obj) {
  check_tree(input_obj,
             require_binary = FALSE,
             require_ultrametric = FALSE)

  if (inherits(input_obj, "matrix")) {
    return(calc_double_cherries_ltable_cpp(input_obj))
  }
  if (inherits(input_obj, "phylo")) {
    return(calc_double_cherries_cpp(as.vector(t(input_obj$edge))))
  }
  stop("input object has to be phylo or ltable")
}

#' Maximum width of branch depths.
#' @description Calculates the maximum width, this is calculated by first
#' collecting the depth of each node and tip across the entire tree, where the
#' depth represents the distance (in nodes) to the root. Then, the width
#' represents the number of occurrences of each possible depth. The maximal
#' width then returns the maximum number of such occurences.
#' @param phy phylogeny or ltable
#' @param normalization "none" or "tips", in which case the resulting statistic
#' is divided by the number of tips in the tree.
#' @return maximum width
#' @references C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
#' transmission patterns. Evolution, Medicine, and Public Health,
#' 2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
#' @export
max_width <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    max_w_stat <- calc_max_width_ltable_cpp(phy)
    if (normalization == "tips" || normalization == TRUE) {
      max_w_stat <- max_w_stat / length(phy[, 1])
    }
    return(max_w_stat)
  }
  if (inherits(phy, "phylo")) {
    # polytomies are checked at the Rcpp side
    max_w_stat <-  calc_max_width_cpp(as.vector(t(phy$edge)))

    if (normalization == "tips" || normalization == TRUE) {
      max_w_stat <- max_w_stat / length(phy$tip.label)
    }
    return(max_w_stat)
  }

  stop("input object has to be phylo or ltable")
}

#' Colless index of (im)balance.
#' @description The Colless index is calculated as the sum of
#' \eqn{abs(L - R)} over all nodes, where L (or R) is the number of extant tips
#' associated with the L (or R) daughter branch at that node.  Higher values
#' indicate higher imbalance. Two normalizations are available,
#' where a correction is made for tree size, under either a yule expectation,
#' or a pda expectation.
#' @param phy phylo object or ltable
#' @param normalization A character string equals to "none" (default) for no
#' normalization or one of "pda" or "yule".
#' @return colless index
#' @references  Colless D H. 1982. Review of: Phylogenetics: The Theory and
#' Practice of Phylogenetic Systematics. Systematic Zoology 31:100-104.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' colless(balanced_tree)
#' colless(unbalanced_tree) # should be higher
colless <- function(phy,
                    normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(calc_colless_ltable_cpp(phy, normalization))
  }
  if (inherits(phy, "phylo")) {
    return(calc_colless_cpp(as.vector(t(phy$edge)), normalization))
  }
  stop("input object has to be phylo or ltable")
}

#' Equal weights Colless index of (im)balance.
#' @description The equal weights Colless index is calculated as the sum of
#' \eqn{abs(L - R) / (L + R - 2)} over all nodes where \eqn{L + R > 2},
#' where L (or R) is the number of extant tips associated with the L (or R)
#' daughter branch at that node.  Maximal imbalance is associated with a value
#' of 1.0. The ew_colless index is not sensitive to tree size.
#' @param phy phylo object or ltable
#' @return colless index
#' @references  A. O. Mooers and S. B. Heard. Inferring Evolutionary Process
#' from Phylogenetic Tree Shape. The Quarterly Review of Biology, 72(1), 1997.
#' doi: 10.1086/419657.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' ew_colless(balanced_tree)
#' ew_colless(unbalanced_tree) # should be higher
ew_colless <- function(phy) {
  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)
  if (inherits(phy, "matrix")) {
    return(calc_eWcolless_ltable_cpp(phy))
  }
  if (inherits(phy, "phylo")) {
    return(calc_eWcolless_cpp(as.vector(t(phy$edge))))
  }
  stop("input object has to be phylo or ltable")
}



#' Corrected Colless index of (im)balance.
#' @description The Corrected Colless index is calculated as the sum of
#' \eqn{abs(L - R)} over all nodes, corrected for tree size by dividing over
#' \eqn{(n-1) * (n-2)}, where n is the number of nodes.
#' @param phy phylo object or ltable
#' @param normalization A character string equals to "none" (default) for no
#' normalization or "yule", in which case the obtained index is divided by
#' the Yule expectation.
#' @return corrected colless index
#' @references  Heard, Stephen B. "Patterns in tree balance among cladistic,
#' phenetic, and randomly generated phylogenetic trees." Evolution 46.6 (1992):
#' 1818-1826.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' colless_corr(balanced_tree)
#' colless_corr(unbalanced_tree) # should be higher
colless_corr <- function(phy,
                         normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(calc_colless_corr_ltable_cpp(phy, normalization))
  }
  if (inherits(phy, "phylo")) {
    return(calc_colless_corr_cpp(as.vector(t(phy$edge)), normalization))
  }
  stop("input object has to be phylo or ltable")
}

#' Quadratic Colless index of (im)balance.
#' @description The Quadratic Colless index is calculated as the sum of
#' \eqn{(L - R)^2} over all nodes.
#' @param phy phylo object or ltable
#' @param normalization A character string equals to "none" (default) for no
#' normalization or "yule"
#' @return quadratic colless index
#' @references  Bartoszek, Krzysztof, et al. "Squaring within the Colless index
#' yields a better balance index." Mathematical Biosciences 331 (2021): 108503.
#' @export
#' @examples simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
#' balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
#' unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
#' colless_quad(balanced_tree)
#' colless_quad(unbalanced_tree) # should be higher
colless_quad <- function(phy,
                         normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = TRUE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    return(calc_colless_quad_ltable_cpp(phy, normalization))
  }
  if (inherits(phy, "phylo")) {
    return(calc_colless_quad_cpp(as.vector(t(phy$edge)), normalization))
  }
  stop("input object has to be phylo or ltable")
}

#' Sum of Inverse branch length
#' @description This statistic calculates per tip, the sum of the inverse of
#' all branch lengths of the shortest path between the root and the tip.
#' @param phy phylo object or ltable
#' @param add_one should we take 1 / bl or 1 / (1 + bl) ? if TRUE, then
#' 1 / (1 + bl) is used. Default: FALSE.
#' @references Williams, P. H., Alonso-Alonso, P., Arbetman, M., Françoso,
#' E., Ghisbain, G., Huang, J., Orr, M. C., Ren, Z.-X., Streinzer, M., T
#' hanoosing, C., Vandame, R., Waite, M., & Brace, S. (2026). Evolutionary
#' Tree for All Bumblebee Species World-Wide Estimated by Combining
#' Information from Fast-Evolving Genes, Slow-Evolving Genes, and Genomic
#' Data (Apidae, Bombus). Insects, 17(6), 540.
#' https://doi.org/10.3390/insects17060540
#' @export
inv_branch_dist <- function(phy, add_one = FALSE) {

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = TRUE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy, drop_extinct = FALSE)
  }
  if (inherits(phy, "phylo")) {
    res <- calc_inv_path_cpp(as.vector(t(phy$edge)),
                             phy$edge.length,
                             add_one)
    distances <- res$distances
    names(distances) <- phy$tip.label[res$tip_ids]
    distances <- distances[sort(names(distances))]
    return(distances)
  }
  stop("input object has to be phylo or ltable")
}

#' Area per pair index
#' @description The area per pair index calculates the sum of the number of
#' edges on the path between all two leaves. Instead, the area per pair index
#' (APP) can also be derived from the Sackin (S) and total cophenetic index
#' (TC):
#' \eqn{ APP = \frac{2}{n}\cdot S - \frac{4}{n(n-1)}\cdot TC}
#' \eqn{APP = 2/n * S - 4/(n(n-1)) * TC}
#' @param phy phylo object or ltable
#' @param normalization "none" or "yule", in which case the acquired result
#' is divided by the expectation for the Yule model.
#' @return Area per pair index
#' @references  T. Araújo Lima, F. M. D. Marquitti, and M. A. M. de Aguiar.
#' Measuring Tree Balance with Normalized Tree Area. arXiv e-prints, art.
#' arXiv:2008.12867, 2020.
#' @export
area_per_pair <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)
  n <- 0

  if (inherits(phy, "matrix")) {
    n <- length(phy[, 1])
  } else if (inherits(phy, "phylo")) {
    n <- length(phy$tip.label)
  } else {
    stop("input object has to be phylo or ltable")
  }

  result <- 2 / n * treestats::sackin(phy) -
            4 / (n * (n - 1)) * treestats::tot_coph(phy)

  if (normalization == "yule") {
    h_n <- sum(1 / (1:n))
    a <- (n + 1) / (n - 1)

    expected_value <- 4 * ((h_n - 1) * a - 1)
    result <- result / expected_value
  }

  return(result)
}

#' Provides a list of all available statistics in the package
#' @param only_balance_stats only return those statistics associated with
#' measuring balance of a tree
#' @return vector with names of summary statistics
#' @export
list_statistics <- function(only_balance_stats = FALSE) {
  all_statistics <- c("gamma", "sackin",
                      "colless", "colless_corr", "colless_quad",
                      "beta", "crown_age",
                      "tree_height", "pigot_rho", "number_of_lineages",
                      "nltt_base", "phylogenetic_div", "avg_ladder",
                      "max_ladder", "cherries", "double_cherries",
                      "il_number", "pitchforks", "four_prong",
                      "stairs", "laplace_spectrum_a", "laplace_spectrum_p",
                      "laplace_spectrum_e", "laplace_spectrum_g",
                      "min_laplace", "max_laplace",
                      "min_adj", "max_adj",
                      "imbalance_steps", "j_one", "b1", "b2", "area_per_pair",
                      "average_leaf_depth", "avg_vert_depth", "i_stat",
                      "ew_colless",
                      "max_del_width", "max_depth", "max_width", "mw_over_md",
                      "rogers", "stairs2", "tot_coph", "var_depth",
                      "symmetry_nodes", "sshape",
                      "mpd", "psv", "vpd", "mntd", "j_stat", "rquartet",
                      "wiener", "max_betweenness",
                      "max_closeness", "max_closenessW",
                      "diameter",
                      "eigen_centrality", "eigen_centralityW",
                      "tot_path", "tot_internal_path",
                      "mean_branch_length", "var_branch_length",
                      "mean_branch_length_int", "mean_branch_length_ext",
                      "var_branch_length_int", "var_branch_length_ext",
                      "treeness", "root_imbalance", "mean_inv_branch_dist")

  all_statistics <- sort(all_statistics)

  bal_stats <- c("sackin", "colless", "beta", "sshape", "avg_ladder",
                 "max_ladder", "cherries", "il_number", "pitchforks", "stairs",
                 "b1", "b2", "area_per_pair", "average_leaf_depth", "i_stat",
                 "ew_colless", "max_del_width", "max_depth", "max_width",
                 "rogers", "stairs2", "tot_coph", "var_depth", "symmetry_nodes",
                 "rquartet", "imbalance_steps", "j_one", "diameter")

  bal_stats <- sort(bal_stats)

  if (only_balance_stats) return(bal_stats)

  return(all_statistics)
}

#' Phylogenetic Species Variability.
#' @description The phylogenetic species variability is bounded in [0, 1]. The
#' psv quantifies how phylogenetic relatedness decrease the variance of a
#' (neutral) trait shared by all species in the tree. As species become more
#' related, the psv tends to 0.
#' Please note that the psv is a special case of the Mean Pair Distance (see
#' appendix of Tucker et al. 2017 for a full derivation), and thus correlates
#' directly.
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", in which case the obtained mean
#' pairwise distance is normalized by the factor 2log(n), where n is the number
#' of tips.
#' @return Phylogenetic Species Variability
#' @references  Helmus M.R., Bland T.J., Williams C.K. & Ives A.R. (2007)
#' Phylogenetic measures of biodiversity. American Naturalist, 169, E68-E83
#'
#' Tucker, Caroline M., et al. "A guide to phylogenetic metrics for
#' conservation, community ecology and macroecology."
#' Biological Reviews 92.2 (2017): 698-715.
#' @export
psv <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)

  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = FALSE)

  if (inherits(phy, "matrix")) {
    phy <- treestats::l_to_phylo(phy)
  }
  if (inherits(phy, "phylo")) {
    psv_stat <- treestats::mean_pair_dist(phy) * 0.5

    if (normalization == "tips" || normalization == TRUE) {
      n <- length(phy$tip.label)
      psv_stat <- psv_stat / (2 * log(n))
    }
    return(psv_stat)
  }
  stop("input object has to be phylo or ltable")
}

context("psv")

test_that("usage", {
  if (requireNamespace("picante")) {

    test_func <- function(phy) {
      n <- length(phy$tip.label)
      sample_mat <- matrix(data = 1, nrow = n, ncol = n)
      colnames(sample_mat) <- phy$tip.label

      return(picante::psv(sample_mat,
                          phy,
                          compute.var = FALSE,
                          scale.vcv = FALSE)[1, 1])
    }

    standard_test(treestats::psv, test_func, drop_complete = TRUE)

    test_polytomies(treestats::psv, test_func)
  }
})

test_that("unrooted", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
  a1 <- treestats::psv(focal_tree)
  a2 <- treestats::psv(ape::unroot(focal_tree))
  testthat::expect_equal(a1, a2)
})

test_that("normalisation", {
  set.seed(42)
  focal_tree <- ape::rphylo(n = 30, birth = 1, death = 0)

  c1 <- treestats::psv(focal_tree)
  c2 <- treestats::psv(focal_tree, normalization = "tips")
  testthat::expect_lt(c2, c1)

  stats1 <- c()
  stats2 <- c()
  for (n in seq(100, 200, by = 10)) {
    focal_tree <- ape::rphylo(n = n, birth = 1, death = 0)
    stats1 <- c(stats1, treestats::psv(focal_tree))
    stats2 <- c(stats2, treestats::psv(focal_tree, normalization = "tips"))
  }

  a1 <- cor(stats1, seq(100, 200, by = 10))
  a2 <- cor(stats2, seq(100, 200, by = 10))

  testthat::expect_lt(a2, a1)
  testthat::expect_lt(a2, 0.4)
})

test_that("wrong_object", {
  testthat::expect_error(
    treestats::psv(10),
    "input object has to be phylo or ltable"
  )

  testthat::expect_error(
    treestats::psv(list()),
    "input object has to be phylo or ltable"
  )
})

#' rebase ltable
#' @description a function to modify an ltable,
#' such that the longest path in the phylogeny is a crown lineage.
#' @param ltable ltable
#' @return modified ltable
#' @export
rebase_ltable <- function(ltable) {
  if (length(ltable[, 1]) == 2) return(ltable)
  prev_main_attractor <- c()
  while (TRUE) {
    res <- swap_deepest(ltable)
    ltable <- res$ltab
    prev_main_attractor <- c(prev_main_attractor,
                             res$main_attractor)
    if (length(prev_main_attractor) > 5) {
      end <- length(prev_main_attractor)

      prev_main_attractor <- prev_main_attractor[(end - 5):end]
    }

    if (stats::sd(prev_main_attractor) == 0 &&
        length(prev_main_attractor) > 3) {
      stop("Stuck in endless loop, possibly due to polytomies")
    }

    if (res$stop) break
  }

  new_ltable <- renumber_ltable(ltable)

  return(new_ltable)
}

#' @keywords internal
renumber_ltable <- function(ltab) {
  temp_new_ltab <- ltab

  for (i in  seq_along(temp_new_ltab[, 1])) {
    current_label <- ltab[i, 3]
    if (abs(current_label) != i) {
      new_label <- i * sign(ltab[i, 3])

      temp_new_ltab[i, 3] <- new_label
      daughters <- which(ltab[, 2] == current_label &
                           ltab[, 1] <= ltab[i, 1])
      if (length(daughters) > 0) {
        temp_new_ltab[daughters, 2] <- i * sign(ltab[i, 3])
      }

      other_instances <- which(ltab[, 3] == i &
                                 ltab[, 1] < ltab[i, 1])
      if (length(other_instances) > 0) {
        temp_new_ltab[other_instances, 3] <- current_label
      }
    }
  }
  return(temp_new_ltab)
}

#' @keywords internal
swap_deepest <- function(ltab) {
  depths <- rep(0, length(ltab[, 1]))
  depths[c(1, 2)] <- 1
  for (i in 3:length(ltab[, 1])) {
    parent_index <- abs(ltab[i, 2])
    depths[parent_index] <- depths[parent_index] + 1
    depths[i] <- depths[parent_index]
  }
  main_attractor <- which.max(depths)
  focal_index <- which(abs(ltab[, 3]) == main_attractor)
  main_attractor <- ltab[focal_index, 3]

  new_ltab <- ltab
  finalized <- FALSE

  if (!focal_index %in% c(1, 2)) {
    parent <- ltab[focal_index, 2]
    index_parent <- which(ltab[, 3] == parent)
    # switch them around, label wise:
    new_ltab[focal_index, 3] <- parent
    new_ltab[focal_index, 2] <- main_attractor
    new_ltab[index_parent, 3] <- main_attractor
    daughters <- which(new_ltab[, 2] == parent &
                         new_ltab[, 1] > new_ltab[focal_index, 1])
    if (length(daughters) > 0) {
      new_ltab[daughters, 2] <- main_attractor
    }

    new_ltab <- renumber_ltable(new_ltab)

  } else {
    finalized <- TRUE
  }

  return(list("ltab" = new_ltab,
              "stop" = finalized,
              "main_attractor" = main_attractor))
}

#' Maximum betweenness centrality.
#' @description Betweenness centrality associates with each node v, the two
#' nodes u, w, for which the shortest path between u and w runs through v, if
#' the tree were re-rooted at node v. Then, we report the node with maximum
#' betweenness centrality.
#' @param phy phylo object or ltable
#' @param normalization "none" or "tips", if tips is chosen, the obtained
#' maximum betweenness is normalized by the total amount of node pair
#' combinations considered, e.g. (n-2)*(n-1), where n is the number of tips.
#' @return Maximum Betweenness
#' @references  Chindelevitch, Leonid, et al. "Network science inspires novel
#' tree shape statistics." Plos one 16.12 (2021): e0259877.
#' @export
max_betweenness <- function(phy, normalization = "none") {
  normalization <- check_normalization_key(normalization)
  check_tree(phy,
             require_binary = FALSE,
             require_ultrametric = FALSE,
             require_rooted = TRUE)

  if (inherits(phy, "matrix")) {
    betweenness_stat <- calc_max_betweenness_ltable_cpp(phy)
    if (normalization == "tips" || normalization == TRUE) {
      n <- 2 * length(phy[, 1]) - 1
      betweenness_stat <- 2 * betweenness_stat / ((n - 2) * (n - 1))
    }
    return(betweenness_stat)
  }
  if (inherits(phy, "phylo")) {
    betweenness_stat <- calc_max_betweenness_cpp(as.vector(t(phy$edge)),
                                                 phy$edge.length)
    if (normalization == "tips" || normalization == TRUE) {
      n <- 1 + length(phy$edge[, 1])
      betweenness_stat <- 2 * betweenness_stat / ((n - 2) * (n - 1))
    }
    return(betweenness_stat)
  }
  stop("input object has to be phylo or ltable")
}

test_that("Package style", {
  testthat::skip_if_not_installed("lintr")
  testthat::skip_on_cran()
  testthat::skip_on_ci()
  lintr::expect_lint_free(
    linters = lintr::linters_with_defaults(object_name_linter = NULL,
                                           indentation_linter = NULL,
                                           return_linter = NULL,
                                           commented_code_linter = NULL))
})

---
title: 'Speed improvement'
author: "Thijs Janzen"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Speed improvement}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE)
knitr::opts_chunk$set(fig.width = 6)
knitr::opts_chunk$set(fig.height = 6)
speed_data <- read.table("https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/Figure_S3_S4/timings.txt", header = TRUE)  # nolint

```

## Speed

An important goal during the development of treestats, has been to
provide not only a vast collection of phylogenetic tree statistics, but
also ensure that calculation of these statistics is fast and reliable.
During development, results of the developed code have continuously been
tested against reference code from other packages. After ensuring
correctness, profiling methods have been used to improve calculation
speed.

### Timing the relationship with tree size

Tree size is an important factor in determining calculation performance.
We have varied tree size from 10 to 1000 in logarithmicly appropriate
steps, calculating statistics for 10 randomly generated Yule trees per
tree size.

```{r out.width="100%", echo=FALSE}
knitr::include_graphics("https://github.com/thijsjanzen/treestats/blob/main/layout/Figure_S3.png?raw=true") # nolint
```

Results show that for many statistics, treestats code provides a serious speed
improvement over the old R code. Furthermore, scaling with tree size is much 
better (consider for instance mpd and mntd, which scale exponentially with tree
size using R, but linear with using treestats). Two functions are faster in 
other packages: colless and gamma. These two functions are slower in treestats
due to additional checks before performing the calculations, these checks verify
whether the tree is binary (for colless) or ultrametric (for the gamma statistic).

### Worst culprits

There are three functions that require a considerably larger amount of time to
be calculated on a tree of 1000 tips, compared to all other functions (see 
Figure below). These three functions rely on calculation of the Eigen values of
a (Laplacian) matrix. Unfortunately, calculating Eigen values is a costly 
endeavour, and can not be further optimized in-house - we use the base Eigen
functions for that already. 
```{r summarise_data, echo=FALSE}
res <- c()
for (x in unique(speed_data$treestatsfunction)) {
  a <- subset(speed_data, speed_data$treestatsfunction == x &
                          speed_data$ntips == 1000 &
                          speed_data$method == "treestats")
  res <- rbind(res, c(x, mean(a$time)))
}
res <- data.frame("treestats_function" = res[, 1],
                     "time" = res[, 2])
res$time <- as.numeric(res$time)

res2 <- res[order(res$time), ]

```

```{r out.width="100%", echo = FALSE}
opar <- par(no.readonly = TRUE)
par(mar = c(8, 4, 4, 4))
barplot(res2$time, names.arg = res2$treestats_function, las = 2,
        cex.names = 0.4, ylab = "Time", log = "y")
par(opar)
```

### Speed improvement
So how much does the speed improve for large trees? We see that for many statistics,
the speed improvement is at least a factor 10, with for some statistics (mntd, mpd,
psv) a speed improvement around four orders of magnitude. Currently, for many
functions speed is limited by preliminary checks if the used tree is ultrametric
and/or binary (if required), and code converting a phy object to C++ objects.
Nevertheless, improvements are considerable and we hope that these improvements
in calculation speed allow for a broader application of many summary statistics.

```{r out.width="100%", echo=FALSE}
knitr::include_graphics("https://github.com/thijsjanzen/treestats/blob/main/layout/Figure_S4.png?raw=true") # nolint
```

---
title: "tree_size"
author: "Thijs Janzen"
date: "`r Sys.Date()`"
output: html_document
vignette: >
  %\VignetteIndexEntry{Tree Size}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
require(TreeSim)
require(ggplot2)
require(tibble)
require(tidyr)
require(dplyr)
```

## Tree Size

Most summary statistics for phylogenetic trees are sensitive to tree size. 
Fortunately, some statistics provde corrections for tree size, based on the 
expectation for the statistic under the Yule model. Let's explore to what 
extent these corrections are effective, and to what extent the non-corrected
statistics are influenced by tree size.

### Simulating trees

To simulate trees, we simulate trees under the Yule model, where we vary the
size of the tree, and we also vary the crown age, in order to sample trees from
a similar distribution. We keep the speciation rate constant. Alternatively, we
could also keep crown age constant and vary the speciation rate.

```{r simulate_trees}
found <- c()
for (tree_size in ceiling(10^seq(1, 3, length.out = 10))) {
  cat(tree_size, "\n")
  stats <- c()

  # more replicates for smaller trees:
  num_repl <- min(100, ceiling(1 + 1000 / tree_size))
  speciation <- 0.5

  # calculate the expected crown age for the relevant tree size:
  max_t <- (1 / speciation) * log(tree_size / 2)

  for (r in 1:num_repl) {
    tree <- TreeSim::sim.bd.taxa.age(n = tree_size,
                                     numbsim = 1,
                                     lambda = speciation,
                                     mu = 0,
                                     age = max_t,
                                     mrca = TRUE)[[1]]
    stats <- rbind(stats, treestats::calc_all_stats(tree, normalize = TRUE))
  }
  stats2 <- apply(stats, 2, median)
  found <- rbind(found, stats2)
}

found <- as_tibble(found)
```

### Yule trees
Before plotting relationships, we collect information on the type of
normalization applied to each statistic. 

```{r collect_information}
norm_info <- read.table("https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/datasets/normalize.txt", header = TRUE) # nolint

found2 <- found |>
  tidyr::gather(key = "statistic", value = "value", -number_of_lineages)

found3 <- left_join(found2, norm_info)
```

Now, we can plot the relationship between tree size and the statistic:
```{r yule_plots}
found3 |>
  filter(normalize == TRUE & type == "Yule") |>
  ggplot(aes(x = number_of_lineages, y = value)) +
  geom_point(col = "blue", alpha = 0.3) +
  geom_line(col = "gold") +
  scale_x_log10() +
  facet_wrap(~statistic, scales = "free_y", ncol = 3) +
  theme_minimal() +
  xlab("Extant tree size") +
  ggtitle("Statistics normalized by the Yule expectation")
```

We see that variation in statistic values is reduced, but not zero.

### Normalization by tree size
For some statistics, no expectations based on the Yule model are available, and
instead, we can normalize by the number of tips. We then find:

```{r size_plots}
found3 |>
  filter(normalize == TRUE & type == "Tips") |>
  ggplot(aes(x = number_of_lineages, y = value)) +
  geom_point(col = "blue", alpha = 0.3) +
  geom_line(col = "gold") +
  scale_x_log10() +
  facet_wrap(~statistic, scales = "free_y", ncol = 4) +
  theme_minimal() +
  xlab("Extant tree size") +
  ggtitle("Statistics normalized by tree size")
```

Clearly, this is not perfect either.

### No normalization

Lastly, we have those statistics for which no normalization is available

```{r no_norm}
found3 |>
  filter(normalize == FALSE) |>
  ggplot(aes(x = number_of_lineages, y = value)) +
  geom_point(col = "blue", alpha = 0.3) +
  geom_line(col = "gold") +
  scale_x_log10() +
  facet_wrap(~statistic, scales = "free_y", ncol = 5) +
  theme_minimal() +
  xlab("Extant tree size") +
  ggtitle("Statistics not normalized")
```

Here we now see clear patterns with tree size. This is worrying, as for some
statistics, patterns might arise that are simply a side-effect of creating 
larger trees, rather than directly influencly the statistic of choice. 
Overall, it appears there are few statistics insensitive to tree size. When
the Yule model correction is used, this tends to work when we are looking at
trees of varying size that generated under the Yule model, but when the 
generating model is different, the correction no longer works. See the original
treestats paper (Janzen & Etienne 2024) for a more thorough study on this. 

---
title: "Getting started"
author: "Thijs Janzen"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
if (!requireNamespace("Matrix", quietly = TRUE)) {
    stop("Matrix needed for this vignette to compile. Please install it.")
}
knitr::opts_chunk$set(echo = TRUE)
knitr::opts_chunk$set(fig.width = 6)
knitr::opts_chunk$set(fig.height = 6)
library(treestats)
```

# Using treestats

The treestats package provides an easy to use interface to calculate summary
statistics on phylogenetic trees. To obtain a list of all supported summary 
statistics use:

```{r list}
list_statistics()
```

If your favourite summary statistic is missing, please let the maintainer know,
treestats is a dynamic package always under development, and the maintainers are
always looking for new statistics!

Given a phylogenetic tree, you can now use of the available functions to
calculate your summary statistic of choice. Let's take for instance the Colless
statistic (and we generate a dummy tree):

```{r colless}
phy <- ape::rphylo(n = 100, birth = 1, death = 0.1)

treestats::colless(phy)
```

Looking at the documentation of the colless statistic (`?colless`), we find that
the function also includes options to normalize for size: either 'pda' or
'yule':
```{r colless2}
treestats::colless(phy, normalization = "yule")
```

## Multiple statistics
The treestats package supports calculating many statistics in one go. For this,
several functions have been set up aptly. Firstly, the function 
`calc_all_stats` will calculate all statistics:
```{r all_stats}
all_stats <- calc_all_stats(phy)
```

Similarly, we can also blanket apply all topology associated summary statistics:
```{r all_stats3}
balance_stats <- calc_topology_stats(phy)
unlist(balance_stats)
```

---
title: 'Example: data'
author: "Thijs Janzen"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Example: data}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r setup, include=FALSE}
required <- c("ape", "pheatmap", "phytools")
if (!all(unlist(lapply(required,
                       function(pkg) requireNamespace(pkg, quietly = TRUE)))))
  knitr::opts_chunk$set(eval = FALSE)

knitr::opts_chunk$set(echo = TRUE)
knitr::opts_chunk$set(fig.width = 6)
knitr::opts_chunk$set(fig.height = 6)


```

## Data

The treestats package can rapidly calculate summary statistics on
phylogenetic trees, and in this vignette, we demonstrate this on
empirical trees. We will make use of family-level pruned trees stemming
from the clootl supertree of birds. These were created for the original
publication accompanying the treestats paper.

```{r read_trees}
focal_trees <- ape::read.tree(file = "https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/datasets/phylogenies/fracced/birds.trees")  # nolint
```

We can now calculate all summary statistics for all trees:

```{r calc_sum_stats}
all_stats <- c()
for (i in seq_along(focal_trees)) {
  if (!ape::is.ultrametric(focal_trees[[i]])) {
    testthat::expect_output(
    focal_trees[[i]] <- phytools::force.ultrametric(focal_trees[[i]])
    )
  }
  focal_stats <- treestats::calc_all_stats(focal_trees[[i]])
  all_stats <- rbind(all_stats, focal_stats)
}
all_stats <- as.data.frame(all_stats)
```

We can now, for instance, plot the distribution of family sizes in
birds:

```{r plot_size}
hist(all_stats$number_of_lineages)
```

Furthermore, we can make a heatmap of all correlations:

```{r corr}
cor.dist <- cor(all_stats)
diag(cor.dist) <- NA
heatmap(cor.dist)
```

This will generate a distorted image: correlations are not corrected for
tree size. We can study this a bit more in detail:

```{r plot with tree size}
opar <- par()
par(mfrow = c(3, 3))
for (stat in c("area_per_pair", "colless", "eigen_centrality",
               "four_prong", "max_betweenness", "max_width",
               "mntd", "sackin", "wiener")) {
  if (stat != "number_of_lineages") {
    x <- all_stats[, colnames(all_stats) == "number_of_lineages"]
    y <- all_stats[, colnames(all_stats) == stat]
    plot(y ~ x, xlab = "Tree size", ylab = stat, pch = 16)
  }
}
par(opar)
```

To correct for this, we will have to go over the entire correlation
matrix.

```{r correct_corr}
tree_size <- all_stats[, colnames(all_stats) == "number_of_lineages"]

for (i in seq_len(nrow(cor.dist))) {
  for (j in seq_len(ncol(cor.dist))) {
    stat1 <- rownames(cor.dist)[i]
    stat2 <- colnames(cor.dist)[j]
    x <- all_stats[, colnames(all_stats) == stat1]
    y <- all_stats[, colnames(all_stats) == stat2]

    a1 <- lm(x ~ tree_size)
    a2 <- lm(y ~ tree_size)
    new_cor <- cor(a1$residuals, a2$residuals)
    cor.dist[i, j] <- new_cor
  }
}
diag(cor.dist) <- NA
heatmap(cor.dist)
```

A nicer way to visualize this is given by the package ppheatmap:

```{r nicer plot, out.width="100%" }
if (requireNamespace("pheatmap")) pheatmap::pheatmap(cor.dist)
```

---
title: "Correlations"
author: "Thijs Janzen"
date: "`r Sys.Date()`"
output: html_document
vignette: >
  %\VignetteIndexEntry{Correlations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(ggplot2)
library(tibble)
library(dendextend)
library(ggdendro)
library(pheatmap)
```

## Correlations

With the ongoing development and addition of new statistics, here we document 
the correlations between all statistics, as measured on the collection of 
empirical trees, as used in Janzen (2024).

### Loading data

We load the data from the supplement, hosted on GitHub.

```{r load_data}
begin <- "https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/datasets/phylogenies/fracced/" # nolint

tree_collection_files <- c("amphibia_fracced.rds",
                           "birds_fracced.rds",
                           "ferns_fracced.rds",
                           "mammals_fracced.rds",
                           "ray_finned_fish_fracced.rds",
                           "sharks_fracced.rds",
                           "vascular_plants_fracced.rds")

taxa_names <- c("Amphibians", "Birds", "Ferns", "Mammals",
                "Ray finned Fish", "Cartaliginous Fish", "Vascular Plants")

found_stats <- c()
for (i in seq_along(tree_collection_files)) {
  raw_url <- paste0(begin, tree_collection_files[i])
  con <- url(raw_url, open = "rb")
  on.exit(close(con))

  tree_collection <- readRDS(con)

  families <- names(tree_collection)
  cat(i, "\n")
  for (j in seq_along(tree_collection)) {
    focal_tree <- tree_collection[[j]]
    if (length(focal_tree$tip.label) >= 10) {
      all_stats <- treestats::calc_all_stats(focal_tree, FALSE)
      to_add <- unlist(all_stats)
      to_add <- c(families[j], to_add)
      found_stats <- rbind(found_stats, to_add)
    }
  }
}
```

We add one statistic with random values to the dataset, as a benchmark for a 
completely uninformative statistic.

```{r add_random}
num_stats <- length(names(all_stats))
colnames(found_stats) <- c("Family", names(all_stats))
found_stats <- tibble::as_tibble(found_stats)
found_stats$random <- runif(n = length(found_stats$Family))

# remove failed calculations:
to_remove <- which(is.na(found_stats$sackin))
found_stats <- found_stats[-to_remove, ]

found_stats2 <- found_stats |>
  dplyr::mutate_at(2:(num_stats + 2), as.numeric)

df <- as.data.frame(found_stats2[, 2:(num_stats + 2)])
```

Statistics are known to correlate with tree size (see also the corresponding 
vignette), so we correct for this effect by calculating the residuals of a 
linear regression of each statistic against tree size, and then calculating the 
correlations between these residuals.

```{r correct_for_tree_size}
get_cor <- function(local_stats) {

  local_stats <- local_stats[!is.na(local_stats$gamma), ]
  local_stats <- local_stats[!is.na(local_stats$sackin), ]

  local_stats2 <- local_stats |>
    dplyr::mutate_at(2:(num_stats + 2), as.numeric)

  local_stats2 <- as.data.frame(local_stats2[, 2:(num_stats + 2)])

  res_cor <- stats::cor(as.data.frame(local_stats2), method = "pearson")

  res_cor2 <- res_cor

  for (i in seq_along(res_cor[1, ])) {
    for (j in seq_along(res_cor[1, ])) {
      stat1 <- colnames(res_cor)[i]
      stat2 <- colnames(res_cor)[j]

      if (stat1 != stat2) {
        if (stat1 != "number_of_lineages" && stat2 != "number_of_lineages") {
          x <- unlist(as.vector(local_stats2[stat1]))  # nolint
          y <- unlist(as.vector(local_stats2[stat2]))  # nolint
          z <- unlist(as.vector(local_stats2["number_of_lineages"])) # nolint

          a1 <- nlme::gls(y ~ z)
          a2 <- nlme::gls(x ~ z)

          found_cor <- cor(a1$residuals, a2$residuals)

          res_cor2[i, j] <- found_cor
        }
      }
    }
  }
  return(res_cor2)
}

master_cor <- get_cor(found_stats2)
```

Then, we plot it as a dendrogram, with added shading for groups, where 
closeness in the tree is calculated as a function of 1 - abs(correlation), 
so that both positive and negative correlations are considered as close, thus
statistics that group together in the dendrogram are those strongly related / 
covering the same information. Shaded areas are based on grouping those groups
together with at least a correlation of 0.2, which is somewhat arbitrary, but
seems to look reasonable, and avoids creating many small groups.

```{r plot_dendrogram, fig.width=25, fig.height=25}

cor1 <- master_cor

cor1 <- as.matrix(cor1)
to_remove <- which(colnames(cor1) == "number_of_lineages")

cor1 <- cor1[-to_remove, -to_remove]
cor1 <- as.data.frame(cor1)
cor1 <- tibble::as_tibble(cor1)
cor1 <- cor1 |>
  dplyr::mutate_at(seq_len(ncol(cor1)), as.numeric)

res_dist <- stats::as.dist(1 - abs(as.matrix(cor1)))

hc <- hclust(res_dist, method = "average")
dend0 <- stats::as.dendrogram(hc)
ddata <- ggdendro::dendro_data(hc, type = "rectangle")

xmax <- 0.8

clust_ref <- dendextend::cutree(dend0, h = xmax)
xmin <- 0
all_rect <- c() # xmin, xmax, ymin, ymax

for (a in unique(clust_ref)) {
  b <- clust_ref[clust_ref == a]
  in_plot <- subset(ddata$labels, ddata$labels$label %in% names(b))
  ymin <- min(in_plot$x) - 0.5
  ymax <- max(in_plot$x) + 0.5
  to_add <- c(xmin, xmax, ymin, ymax)
  all_rect <- rbind(all_rect, to_add)
}

rect_plot <- data.frame(xmin = all_rect[, 1],
                        xmax = all_rect[, 2],
                        ymin = all_rect[, 3],
                        ymax = all_rect[, 4],
                        categ = seq_along(all_rect[, 1]))

lvls <- sort(unique(as.numeric(rect_plot$categ)))
rect_plot$categ <- factor(rect_plot$categ, levels = lvls)

ggplot() +
  geom_rect(data = rect_plot,
            aes(xmin = ymin, xmax = ymax, ymin = xmin, ymax = xmax,
                fill = categ), alpha = 1) +
  scale_fill_brewer(type = "qual", palette = 3) +
  geom_segment(data = ggdendro::segment(ddata),
               aes(x = x, y = y, xend = xend, yend = yend)
  ) +
  geom_text(data = ggdendro::label(ddata),
            aes(x = x, y = y, label = label, hjust = 0),
            size = 10
  ) +
  coord_flip() +
  ylim(1, -0.3) +
  theme_classic() +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        legend.position = "none") +
  ylab("1 - abs(correlation)") +
  xlab("")
```

Or as a heatmap, where red colors indicate positive correlations and blue colors
indicate negative correlations.

```{r heatmap, , fig.width=25, fig.height=25}
breakz <- seq(-1, 1, length.out = 99)
cor2 <- as.matrix(cor1)
hm1 <- pheatmap::pheatmap(mat = cor2,
                          breaks = breakz,
                          clustering_method = "average",
                          fontsize_col = 18,
                          fontsize_row = 18)
```

---
title: "Correlations"
author: "Thijs Janzen"
date: "`r Sys.Date()`"
output: html_document
vignette: >
  %\VignetteIndexEntry{Correlations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
library(ggplot2)
library(tibble)
library(dplyr)
library(dendextend)
library(ggdendro)
library(pheatmap)
```

## Correlations

With the ongoing development and addition of new statistics, here we document 
the correlations between all statistics, as measured on the collection of 
empirical trees, as used in Janzen (2024).

### Loading data

We load the data from the supplement, hosted on GitHub.

```{r load_data}
begin <- "https://raw.githubusercontent.com/thijsjanzen/treestats-scripts/main/datasets/phylogenies/fracced/" # nolint

tree_collection_files <- c("amphibia_fracced.rds",
                           "birds_fracced.rds",
                           "ferns_fracced.rds",
                           "mammals_fracced.rds",
                           "ray_finned_fish_fracced.rds",
                           "sharks_fracced.rds",
                           "vascular_plants_fracced.rds")

taxa_names <- c("Amphibians", "Birds", "Ferns", "Mammals",
                "Ray finned Fish", "Cartaliginous Fish", "Vascular Plants")

found_stats <- c()
for (i in seq_along(tree_collection_files)) {
  raw_url <- paste0(begin, tree_collection_files[i])
  con <- url(raw_url, open = "rb")
  on.exit(close(con))

  tree_collection <- readRDS(con)

  families <- names(tree_collection)
  cat(taxa_names[i], "\n")
  for (j in seq_along(tree_collection)) {
    focal_tree <- tree_collection[[j]]
    if (length(focal_tree$tip.label) >= 10) {
      if (!ape::is.ultrametric(focal_tree)) {
        testthat::expect_output(
          focal_tree <- phytools::force.ultrametric(focal_tree)
        )
      }
      all_stats <- treestats::calc_all_stats(focal_tree, FALSE)
      to_add <- unlist(all_stats)
      # we also need to add if they have polytomies
      is_binary <- treestats::check_binary(focal_tree)

      to_add <- c(families[j], is_binary, to_add)

      found_stats <- rbind(found_stats, to_add)
    }
  }
}
```

We add one statistic with random values to the dataset, as a benchmark for a 
completely uninformative statistic.

```{r add_random}
num_stats <- length(names(all_stats))
colnames(found_stats) <- c("Family", "is_binary", names(all_stats))
found_stats <- tibble::as_tibble(found_stats)
found_stats$random <- runif(n = length(found_stats$Family))

found_stats2 <- found_stats |>
  dplyr::mutate_at(3:(num_stats + 3), as.numeric)
```

Statistics are known to correlate with tree size (see also the corresponding 
vignette), so we correct for this effect by calculating the residuals of a 
linear regression of each statistic against tree size, and then calculating the 
correlations between these residuals.

```{r correct_for_tree_size}
get_cor <- function(local_stats) {

  local_stats2 <- local_stats |>
    dplyr::mutate_at(3:(num_stats + 3), as.numeric)

  local_stats2 <- as.data.frame(local_stats2[, 3:(num_stats + 3)])

  # remove NA statistics
  to_remove <- c()
  for (i in seq_len(ncol(local_stats2))) {
    if (any(is.na(local_stats2[, i]))) {
      to_remove <- c(to_remove, i)
    }
  }
  if (length(to_remove)) local_stats2 <- local_stats2[, -to_remove]

  res_cor <- stats::cor(as.data.frame(local_stats2), method = "pearson")

  res_cor2 <- res_cor

  for (i in seq_along(res_cor[1, ])) {
    for (j in seq_along(res_cor[1, ])) {
      stat1 <- colnames(res_cor)[i]
      stat2 <- colnames(res_cor)[j]

      if (stat1 != stat2) {
        if (stat1 != "number_of_lineages" && stat2 != "number_of_lineages") {
          x <- unlist(as.vector(local_stats2[stat1]))  # nolint
          y <- unlist(as.vector(local_stats2[stat2]))  # nolint
          z <- unlist(as.vector(local_stats2["number_of_lineages"])) # nolint

          a1 <- nlme::gls(y ~ z)
          a2 <- nlme::gls(x ~ z)

          found_cor <- cor(a1$residuals, a2$residuals)

          res_cor2[i, j] <- found_cor
        }
      }
    }
  }
  return(res_cor2)
}
```
```{r cor_binary} 
found_stats_binary <- found_stats2 |>
  filter(is_binary == TRUE)
master_cor_binary <- get_cor(found_stats_binary)
```

```{r cor_poly}
found_stats_poly <- found_stats2 |>
  filter(is_binary == FALSE)
master_cor_polytomy <- get_cor(found_stats_poly)
```

## Binary trees

Then, we plot it as a dendrogram, with added shading for groups, where 
closeness in the tree is calculated as a function of 1 - abs(correlation), 
so that both positive and negative correlations are considered as close, thus
statistics that group together in the dendrogram are those strongly related / 
covering the same information. Shaded areas are based on grouping those groups
together with at least a correlation of 0.2, which is somewhat arbitrary, but
seems to look reasonable, and avoids creating many small groups.

```{r plot_dendrogram, fig.width=25, fig.height=25}

cor1 <- master_cor_binary

cor1 <- as.matrix(cor1)
to_remove <- which(colnames(cor1) == "number_of_lineages")

cor1 <- cor1[-to_remove, -to_remove]
cor1 <- as.data.frame(cor1)
cor1 <- tibble::as_tibble(cor1)
cor1 <- cor1 |>
  dplyr::mutate_at(seq_len(ncol(cor1)), as.numeric)

res_dist <- stats::as.dist(1 - abs(as.matrix(cor1)))

hc <- hclust(res_dist, method = "average")
dend0 <- stats::as.dendrogram(hc)
ddata <- ggdendro::dendro_data(hc, type = "rectangle")

xmax <- 0.8

clust_ref <- dendextend::cutree(dend0, h = xmax)
xmin <- 0
all_rect <- c() # xmin, xmax, ymin, ymax

for (a in unique(clust_ref)) {
  b <- clust_ref[clust_ref == a]
  in_plot <- subset(ddata$labels, ddata$labels$label %in% names(b))
  ymin <- min(in_plot$x) - 0.5
  ymax <- max(in_plot$x) + 0.5
  to_add <- c(xmin, xmax, ymin, ymax)
  all_rect <- rbind(all_rect, to_add)
}

rect_plot <- data.frame(xmin = all_rect[, 1],
                        xmax = all_rect[, 2],
                        ymin = all_rect[, 3],
                        ymax = all_rect[, 4],
                        categ = seq_along(all_rect[, 1]))

lvls <- sort(unique(as.numeric(rect_plot$categ)))
rect_plot$categ <- factor(rect_plot$categ, levels = lvls)

ggplot() +
  geom_rect(data = rect_plot,
            aes(xmin = ymin, xmax = ymax, ymin = xmin, ymax = xmax,
                fill = categ), alpha = 1) +
  scale_fill_brewer(type = "qual", palette = 3) +
  geom_segment(data = ggdendro::segment(ddata),
               aes(x = x, y = y, xend = xend, yend = yend)
  ) +
  geom_text(data = ggdendro::label(ddata),
            aes(x = x, y = y, label = label, hjust = 0),
            size = 10
  ) +
  coord_flip() +
  ylim(1, -0.3) +
  theme_classic() +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        legend.position = "none") +
  ylab("1 - abs(correlation)") +
  xlab("") +
  ggtitle("On Binary Trees")
```

Or as a heatmap, where red colors indicate positive correlations and blue colors
indicate negative correlations.

```{r heatmap, , fig.width=25, fig.height=25}
breakz <- seq(-1, 1, length.out = 99)
cor2 <- as.matrix(cor1)
rownames(cor2) <- colnames(cor2)
hm1 <- pheatmap::pheatmap(mat = cor2,
                          breaks = breakz,
                          clustering_method = "average",
                          fontsize_col = 18,
                          fontsize_row = 18)
```

## Polytomies

Then, we plot it as a dendrogram, with added shading for groups, where 
closeness in the tree is calculated as a function of 1 - abs(correlation), 
so that both positive and negative correlations are considered as close, thus
statistics that group together in the dendrogram are those strongly related / 
covering the same information. Shaded areas are based on grouping those groups
together with at least a correlation of 0.2, which is somewhat arbitrary, but
seems to look reasonable, and avoids creating many small groups.

```{r plot_dendrogram_poly, fig.width=25, fig.height=25}

cor1 <- master_cor_polytomy

cor1 <- as.matrix(cor1)
to_remove <- which(colnames(cor1) == "number_of_lineages")

cor1 <- cor1[-to_remove, -to_remove]
cor1 <- as.data.frame(cor1)
cor1 <- tibble::as_tibble(cor1)
cor1 <- cor1 |>
  dplyr::mutate_at(seq_len(ncol(cor1)), as.numeric)

res_dist <- stats::as.dist(1 - abs(as.matrix(cor1)))

hc <- hclust(res_dist, method = "average")
dend0 <- stats::as.dendrogram(hc)
ddata <- ggdendro::dendro_data(hc, type = "rectangle")

xmax <- 0.8

clust_ref <- dendextend::cutree(dend0, h = xmax)
xmin <- 0
all_rect <- c() # xmin, xmax, ymin, ymax

for (a in unique(clust_ref)) {
  b <- clust_ref[clust_ref == a]
  in_plot <- subset(ddata$labels, ddata$labels$label %in% names(b))
  ymin <- min(in_plot$x) - 0.5
  ymax <- max(in_plot$x) + 0.5
  to_add <- c(xmin, xmax, ymin, ymax)
  all_rect <- rbind(all_rect, to_add)
}

rect_plot <- data.frame(xmin = all_rect[, 1],
                        xmax = all_rect[, 2],
                        ymin = all_rect[, 3],
                        ymax = all_rect[, 4],
                        categ = seq_along(all_rect[, 1]))

lvls <- sort(unique(as.numeric(rect_plot$categ)))
rect_plot$categ <- factor(rect_plot$categ, levels = lvls)

ggplot() +
  geom_rect(data = rect_plot,
            aes(xmin = ymin, xmax = ymax, ymin = xmin, ymax = xmax,
                fill = categ), alpha = 1) +
  scale_fill_brewer(type = "qual", palette = 3) +
  geom_segment(data = ggdendro::segment(ddata),
               aes(x = x, y = y, xend = xend, yend = yend)
  ) +
  geom_text(data = ggdendro::label(ddata),
            aes(x = x, y = y, label = label, hjust = 0),
            size = 10
  ) +
  coord_flip() +
  ylim(1, -0.3) +
  theme_classic() +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        legend.position = "none") +
  ylab("1 - abs(correlation)") +
  xlab("") +
  ggtitle("On Trees with Polytomies")
```

Or as a heatmap, where red colors indicate positive correlations and blue colors
indicate negative correlations.

```{r heatmap poly, , fig.width=25, fig.height=25}
breakz <- seq(-1, 1, length.out = 99)
cor2 <- as.matrix(cor1)
rownames(cor2) <- colnames(cor2)
hm1 <- pheatmap::pheatmap(mat = cor2,
                          breaks = breakz,
                          clustering_method = "average",
                          fontsize_col = 18,
                          fontsize_row = 18)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean_branch_length.R
\name{var_branch_length_ext}
\alias{var_branch_length_ext}
\title{Variance of external branch lengths of a tree}
\usage{
var_branch_length_ext(phy)
}
\arguments{
\item{phy}{phylo object or Ltable}
}
\value{
variance of external branch lengths
}
\description{
Variance of external branch lengths of a tree, e.g. of
branches leading to a tip.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_all_stats.R
\name{calc_all_stats}
\alias{calc_all_stats}
\title{Apply all available tree statistics to a single tree}
\usage{
calc_all_stats(phylo, normalize = FALSE)
}
\arguments{
\item{phylo}{phylo object}

\item{normalize}{if set to TRUE, results are normalized (if possible) under
either the  Yule expectation (if available), or the number of tips}
}
\value{
List with statistics
}
\description{
this function applies all tree statistics available in
this package to a
single tree, being:
\itemize{
  \item gamma
  \item Sackin
  \item Colless
  \item corrected Colless
  \item quadratic Colless
  \item Aldous' beta statistic
  \item Blum
  \item crown age
  \item tree height
  \item Pigot's rho
  \item number of lineages
  \item nLTT with empty tree
  \item phylogenetic diversity
  \item avgLadder index
  \item cherries
  \item double cherries
  \item ILnumber
  \item pitchforks
  \item stairs
  \item stairs2
  \item laplacian spectrum
  \item B1
  \item B2
  \item area per pair (aPP)
  \item average leaf depth (aLD)
  \item I statistic
  \item ewColless
  \item max Delta Width (maxDelW)
  \item maximum of Depth
  \item variance of Depth
  \item maximum Width
  \item Rogers
  \item total Cophenetic distance
  \item symmetry Nodes
  \item mean of pairwise distance (mpd)
  \item variance of pairwise distance (vpd)
  \item Phylogenetic Species Variability (psv)
  \item mean nearest taxon distance (mntd)
  \item J statistic of entropy
  \item rquartet index
  \item Wiener index
  \item max betweenness
  \item max closeness
  \item diameter, without branch lenghts
  \item maximum eigen vector value
  \item mean branch length
  \item variance of branch length
  \item mean external branch length
  \item variance of external branch length
  \item mean internal branch length
  \item variance of internal branch length
  \item number of imbalancing steps
  \item j_one statistic
  \item treeness statistic
  \item branch weighted colless statistic
  \item minimum eigenvalue of the Laplacian matrix
  \item maximum eigenvalue of the Laplacian matrix
  \item minimum eigenvalue of the adjacency matrix
  \item maximum eigenvalue of the adjacency matrix
  \item mean of the sum of inverse branch lengths
}

For the Laplacian spectrum properties, four properties of the eigenvalue
distribution are returned: 1) asymmetry, 2) peakedness, 3) log(principal
eigenvalue) and 4) eigengap.
Please notice that for some very small or very large trees, some of the
statistics can not be calculated. The function will report an NA for this
statistic, but will not break, to facilitate batch analysis of large numbers
of trees.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rquartet.R
\name{rquartet}
\alias{rquartet}
\title{Rquartet index.}
\usage{
rquartet(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{The index can be normalized by the expectation under
the Yule ("yule") or PDA model ("pda").}
}
\value{
rquartet index
}
\description{
The rquartet index counts the number of potential fully
balanced rooted subtrees of 4 tips in the tree. The function in treestats
assumes a bifurcating tree. For trees with polytomies, we use
treebalance::rquartedI, which can also take polytomies into account.
}
\references{
T. M. Coronado, A. Mir, F. Rosselló, and G. Valiente. A balance
index for phylogenetic trees based on rooted quartets. Journal of
Mathematical Biology, 79(3):1105-1148, 2019. doi: 10.1007/s00285-019-01377-w.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_closeness.R
\name{max_closeness}
\alias{max_closeness}
\title{Maximum closeness}
\usage{
max_closeness(phy, weight = TRUE, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{weight}{if TRUE, uses branch lengths.}

\item{normalization}{"none" or "tips", in which case an arbitrary post-hoc
correction is performed by dividing by the expectation of n log(n), where n
is the number of tips.}
}
\value{
Maximum Closeness
}
\description{
Closeness is defined as 1 / Farness, where Farness is the sum
of distances from a node to all the other nodes in the tree. Here, we return
the node with maximum closeness.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." Plos one 16.12 (2021): e0259877.
Wang W, Tang CY. Distributed computation of classic and exponential closeness
on tree graphs. Proceedings of the American Control Conference. IEEE; 2014.
p. 2090–2095.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/area_per_pair.R
\name{area_per_pair}
\alias{area_per_pair}
\title{Area per pair index}
\usage{
area_per_pair(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "yule", in which case the acquired result
is divided by the expectation for the Yule model.}
}
\value{
Area per pair index
}
\description{
The area per pair index calculates the sum of the number of
edges on the path between all two leaves. Instead, the area per pair index
(APP) can also be derived from the Sackin (S) and total cophenetic index
(TC):
\eqn{ APP = \frac{2}{n}\cdot S - \frac{4}{n(n-1)}\cdot TC}
\eqn{APP = 2/n * S - 4/(n(n-1)) * TC}
}
\references{
T. Araújo Lima, F. M. D. Marquitti, and M. A. M. de Aguiar.
Measuring Tree Balance with Normalized Tree Area. arXiv e-prints, art.
arXiv:2008.12867, 2020.
}

\name{treestats-package}
\alias{treestats-package}
\alias{treestats}
\docType{package}
\title{
  Collection of phylogenetic tree statistics
}
\description{
  The 'treestats' package contains a collection of phylogenetic tree statistics,
  implemented in C++ to ensure high speed.
}
\details{
  Given a phylogenetic tree as a phylo object, the 'treestats' package provides
  a wide range of individual functions returning the relevant statistic. In
  addition, there are three functions available that calculate a collection
  of statistics at once: calc_all_statistics (which calculates all currently
  implemented statistics of treestats), calc_balance_stats, which calculates
  all (im)balance related statistics and calc_brts_stats, which calculates all
  branching times and branch length related statistics.
  Furthermore, there are a number of additional tools available that allow for
  phylogenetic tree manipulation: make_unbalanced_tree, which creates an
  imbalanced tree in a stepwise fashion. Then there are two functions related to
  conversion from and to an ltable, an alternative notation method used in some
  simulations. These are l_to_phylo which is a C++ based
  version of DDD::L2phylo, which converts an ltable to a phylo object, and
  phylo_to_l, which is a C+ based version of DDD::phylo2L, which converts a
  phylo object to an ltable. Lastly, the treestats package also includes a
  faster, C++ based, implementation of ape::branching.times (the
  function branching_times), which yields the same sequence of branching times,
  but omits the branching names in favour of speed.
}
\author{
Maintainer: Thijs Janzen <thijsjanzen@gmail.com>
}
\references{
Phylogenetic tree statistics: a systematic overview using the new R package
'treestats' Thijs Janzen, Rampal S. Etienne bioRxiv 2024.01.24.576848;
doi: https://doi.org/10.1101/2024.01.24.576848
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imbal_steps.R
\name{imbalance_steps}
\alias{imbalance_steps}
\title{Imbalance steps index}
\usage{
imbalance_steps(input_obj, normalization = FALSE)
}
\arguments{
\item{input_obj}{phylo object or ltable}

\item{normalization}{if true, the number of steps taken is normalized by tree
size, by dividing by the maximum number of moves required to move from a
fully balanced to a fully imbalanced tree, which is \eqn{N - log2(N) - 1},
where N is the number of extant tips.}
}
\value{
required number of moves
}
\description{
Calculates the number of moves required to transform the focal
tree into a fully imbalanced (caterpillar) tree. Higher value indicates a
more balanced tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylogenetic_diversity.R
\name{phylogenetic_diversity}
\alias{phylogenetic_diversity}
\title{Phylogenetic diversity at time point t}
\usage{
phylogenetic_diversity(input_obj, t = 0, extinct_tol = NULL)
}
\arguments{
\item{input_obj}{phylo object or Ltable}

\item{t}{time point at which to measure phylogenetic diversity, alternatively
a vector of time points can also be provided. Time is measured with 0 being
the present.}

\item{extinct_tol}{tolerance to determine if a lineage is extinct at time t.
Default is 1/100 * smallest branch length of the tree.}
}
\value{
phylogenetic diversity, or vector of phylogenetic diversity measures
if a vector of time points is used as input.
}
\description{
The phylogenetic diversity at time t is given by the total
branch length of the tree reconstructed up until time point t. Time is
measured increasingly, with the crown age equal to 0. Thus, the time at
the present is equal to the crown age.
}
\references{
Faith, Daniel P. "Conservation evaluation and phylogenetic
diversity." Biological conservation 61.1 (1992): 1-10.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stairs.R
\name{stairs2}
\alias{stairs2}
\title{Stairs2 index}
\usage{
stairs2(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
number of stairs
}
\description{
Calculates the stairs2 measure, from the phyloTop package. The
stairs2 reflects the imbalance at each node, where it represents the average
across measure at each node, the measure being \eqn{min(L, R) / max(L, R)},
where L and R reflect the number of tips connected at the left (L) and right
(R) daughter.
}
\references{
Norström, Melissa M., et al. "Phylotempo: a set of r scripts for
assessing and visualizing temporal clustering in genealogies inferred from
serially sampled viral sequences." Evolutionary Bioinformatics 8 (2012):
EBO-S9738.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_betweenness.R
\name{max_betweenness}
\alias{max_betweenness}
\title{Maximum betweenness centrality.}
\usage{
max_betweenness(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", if tips is chosen, the obtained
maximum betweenness is normalized by the total amount of node pair
combinations considered, e.g. (n-2)*(n-1), where n is the number of tips.}
}
\value{
Maximum Betweenness
}
\description{
Betweenness centrality associates with each node v, the two
nodes u, w, for which the shortest path between u and w runs through v, if
the tree were re-rooted at node v. Then, we report the node with maximum
betweenness centrality.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." Plos one 16.12 (2021): e0259877.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minmax_laplace.R
\name{minmax_laplace}
\alias{minmax_laplace}
\title{Laplacian Matrix properties}
\usage{
minmax_laplace(phy, use_rspectra = TRUE, force_min = FALSE)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{use_rspectra}{boolean to indicate whether the helping package RSpectra
should be used, in which case only the minimum and maximum values are
returned, and not the full list of eigenvalues, which makes the computation
much faster.}

\item{force_min}{boolean to indicate whether or not to also calculate
the minimum positive eigenvalue if the tree is larger than 23170 tips.
Default is FALSE, as this can be very slow, but in some cases might be
of interest.}
}
\value{
List with the minimum and maximum eigenvalues
}
\description{
Calculates the eigenvalues of the Laplacian Matrix, where the
Laplacian matrix is the matrix representation of a graph, in this case a
phylogeny.
When the R package \pkg{RSpectra} is available, a faster
calculation can be used, which does not calculate all eigenvalues, but only
the maximum and minimum. As such, when using this option, the vector of all
eigenvalues is not returned.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." Plos one 16.12 (2021): e0259877.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/number_of_lineages.R
\name{number_of_lineages}
\alias{number_of_lineages}
\title{Number of tips of a tree, including extinct tips.}
\usage{
number_of_lineages(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
number of lineages
}
\description{
Number of tips of a tree, including extinct tips.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/beta.R
\name{beta_statistic}
\alias{beta_statistic}
\title{Aldous' beta statistic.}
\usage{
beta_statistic(
  phy,
  upper_lim = 10,
  algorithm = "COBYLA",
  abs_tol = 1e-04,
  rel_tol = 1e-06
)
}
\arguments{
\item{phy}{phylogeny or ltable}

\item{upper_lim}{Upper limit for beta parameter, default = 10.}

\item{algorithm}{optimization algorithm used, default is "COBYLA"
(Constrained Optimization BY Linear Approximations), also available are
"subplex" and "simplex". Subplex and Simplex seem to have difficulties with
unbalanced trees, e.g. if beta < 0.}

\item{abs_tol}{absolute stopping criterion of optimization. Default is 1e-4.}

\item{rel_tol}{relative stopping criterion of optimization. Default is 1e-6.}
}
\value{
Beta value
}
\description{
The Beta statistic fits a beta splitting model to each node,
assuming that the number of extant descendants of each daughter branch is
split following a beta distribution, such that the number of extant
descendants x and y at a node follows \eqn{q(x, y) = s_n(\beta)^{-1}
\frac{(\gamma(x + 1 + \beta)\gamma(y + 1 + \beta))}{\gamma(x+1)\gamma(y+1)}},
where \eqn{s_n(\beta)^{-1}} is a normalizing constant. When this model is fit
to a tree,
different values of beta correspond to the expectation following from
different diversification models, such that a beta of 0 corresponds to a
Yule tree, a beta of \eqn{-3/2} to a tree following from a PDA model.
In general,
negative beta values correspond to trees more unbalanced than Yule trees, and
beta values larger than zero indicate trees more balanced than Yule trees.
The lower bound of the beta splitting parameter is -2.
}
\examples{
simulated_tree <- ape::rphylo(n = 100, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
beta_statistic(balanced_tree) # should be approximately 10
beta_statistic(simulated_tree) # should be near 0
beta_statistic(unbalanced_tree) # should be approximately -2
}
\references{
Aldous, David. "Probability distributions on cladograms." Random
discrete structures. Springer, New York, NY, 1996. 1-18.
Jones, Graham R. "Tree models for macroevolution and phylogenetic analysis."
Systematic biology 60.6 (2011): 735-746.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pigot_rho.R
\name{pigot_rho}
\alias{pigot_rho}
\title{Pigot's rho}
\usage{
pigot_rho(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
rho
}
\description{
Calculates the change in rate between the first half and the
second half of the extant phylogeny. \eqn{\rho = (r2 - r1) / (r1 + r2)},
where r reflects the rate in either the first or second half.
The rate within a half is given by \eqn{(log(n2) - log(n1) / t}, where n2 is
the number of lineages at the end of the half, and n1 the number of lineages
at the start of the half. Rho varies between -1 and 1, with a 0 indicating a
constant rate across the phylogeny, a \eqn{\rho < 0} indicating a slow down
and a \eqn{rho > 0} indicating a speed up of speciation. In contrast to the
Gamma statistic, Pigot's rho is not sensitive to tree size.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
pigot_rho(simulated_tree) # should be around 0.
ddd_tree <- DDD::dd_sim(pars = c(1, 0, 10), age = 7)$tes
pigot_rho(ddd_tree) # because of diversity dependence, should be < 0
}
\references{
Alex L. Pigot, Albert B. Phillimore, Ian P. F. Owens,
C. David L. Orme, The Shape and Temporal Dynamics of Phylogenetic Trees
Arising from Geographic Speciation, Systematic Biology, Volume 59, Issue 6,
December 2010, Pages 660–673, https://doi.org/10.1093/sysbio/syq058
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{phylo_to_l}
\alias{phylo_to_l}
\title{Function to generate an ltable from a phy object.}
\usage{
phylo_to_l(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
ltable (see description)
}
\description{
This function is a C++
implementation of the function DDD::phylo2L. An L table summarises a
phylogeny in a table with four columns, being: 1) time at which a species
is born, 2) label of the parent of the species, where positive and negative
numbers indicate whether the species belongs to the left or right crown
lineage, 3) label of the daughter species itself (again positive or
negative depending on left or right crown lineage), and the last column 4)
indicates the time of extinction of a species, or -1 if the species is
extant.
}
\examples{
simulated_tree <- ape::rphylo(n = 4, birth = 1, death = 0)
ltable <- phylo_to_l(simulated_tree)
reconstructed_tree <- DDD::L2phylo(ltable)
old_par <- par()
par(mfrow = c(1, 2))
# trees should be more or less similar, although labels may not match, and
# rotations might cause (initial) visual mismatches
plot(simulated_tree)
plot(reconstructed_tree)
par(old_par)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sackin.R
\name{sackin}
\alias{sackin}
\title{Sackin index of (im)balance.}
\usage{
sackin(phy, normalization = "none")
}
\arguments{
\item{phy}{phylogeny or ltable}

\item{normalization}{normalization, either 'none' (default), "yule" or "pda".}
}
\value{
Sackin index
}
\description{
The Sackin index is calculated as the sum of ancestors for each
of the tips. Higher values indicate higher imbalance. Two normalizations
are available, where a correction is made for tree size, under either a
Yule expectation, or a pda expectation.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
sackin(balanced_tree)
sackin(unbalanced_tree) # should be much higher
}
\references{
M. J. Sackin (1972). "Good" and "Bad" Phenograms.
Systematic Biology. 21:225-226.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minmax_adj.R
\name{minmax_adj}
\alias{minmax_adj}
\title{Adjancency Matrix properties}
\usage{
minmax_adj(phy, use_rspectra = TRUE, force_min = FALSE)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{use_rspectra}{boolean to indicate whether the helping package RSpectra
should be used, in which case only the minimum and maximum values are
returned}

\item{force_min}{boolean to indicate whether or not to also calculate
the minimum positive eigenvalue if the tree is larger than 23170 tips.
Default is FALSE, as this can be very slow, but in some cases might be
of interest.}
}
\value{
List with the minimum and maximum eigenvalues
}
\description{
Calculates the eigenvalues of the Adjancency Matrix, where the
Adjacency matrix is a square matrix indicate whether pairs of vertices
are adjacent or not on a graph - here, entries in the matrix indicate
connections between nodes (and betweens nodes and tips). Entries in the
adjacency matrix are weighted by branch length. Then, using the adjacency
matrix, we calculate the spectral properties of the matrix, e.g. the
minimum positive and maximum positive eigenvalues of the matrix.
When the R package \pkg{RSpectra} is available,
a faster calculation can be used, which does not calculate all eigenvalues,
but only the maximum and minimum. As such, when using this option,
the vector of all eigenvalues is not returned.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." Plos one 16.12 (2021): e0259877.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/b1.R
\name{b1}
\alias{b1}
\title{B1 metric}
\usage{
b1(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the resulting
statistic is divided by the number of tips in the tree, as a crude way of
normalization.}
}
\value{
B1 statistic
}
\description{
Balance metric (in the case of a binary tree), which measures
the sum across all internal nodes of one over the maximum depth of all
attached tips to that node.
When calculated on a binary tree, the b1 statistic is a true balance
statistic (meaning that the caterpillar tree is the most extreme, unbalanced
tree). However, the treestats package also supports calculation on
non-binary trees.
}
\references{
K.-T. Shao and R. R. Sokal. Tree Balance.
Systematic Zoology, 39(3):266, 1990. doi: 10.2307/2992186.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{check_binary}
\alias{check_binary}
\title{Faster than the ape version}
\usage{
check_binary(phy)
}
\arguments{
\item{phy}{phylo object}
}
\description{
Fast Rcpp function to check if a tree is binary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tot_coph.R
\name{tot_coph}
\alias{tot_coph}
\title{Total cophenetic index.}
\usage{
tot_coph(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "yule", when "yule" is chosen, the statistic
is divided by the Yule expectation}
}
\value{
Total cophenetic index
}
\description{
The total cophenetic index is the sum of the depth of the last
common ancestor of all pairs of leaves.
}
\references{
A. Mir, F. Rosselló, and L. Rotger. A new balance index for
phylogenetic trees. Mathematical Bio-sciences, 241(1):125-136, 2013.
doi: 10.1016/j.mbs.2012.10.005.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean_pair_dist.R
\name{mean_pair_dist}
\alias{mean_pair_dist}
\title{Mean Pairwise distance}
\usage{
mean_pair_dist(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the obtained mean
pairwise distance is normalized by the factor 2log(n),
where n is the number of tips.}
}
\value{
Mean pairwise distance
}
\description{
Fast function using C++ to calculate the mean pairwise distance,
using the fast algorithm by Constantinos, Sandel & Cheliotis (2012).
}
\references{
Webb, C., D. Ackerly, M. McPeek, and M. Donoghue. 2002.
Phylogenies and community ecology. Annual Review of Ecology and
Systematics 33:475-505.

Tsirogiannis, Constantinos, Brody Sandel, and Dimitris Cheliotis.
"Efficient computation of popular phylogenetic tree measures." Algorithms in
Bioinformatics: 12th International Workshop, WABI 2012, Ljubljana, Slovenia,
September 10-12, 2012. Proceedings 12. Springer Berlin Heidelberg, 2012.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blum.R
\name{blum}
\alias{blum}
\title{Blum index of (im)balance.}
\usage{
blum(phy, normalization = FALSE)
}
\arguments{
\item{phy}{phylogeny or ltable}

\item{normalization}{because the Blum index sums over all nodes,
the resulting statistic tends to be correlated with the number of extant
tips. Normalization can be performed by dividing by the number of extant
tips.}
}
\value{
Blum index of imbalance
}
\description{
The Blum index of imbalance (also known as the s-shape
statistic (see \link{sshape})) calculates the sum of \eqn{log(N-1)} over all
internal nodes, where N represents the total number of extant tips connected
to that node. An alternative implementation can be found in the Castor R
package.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
  balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
  unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
  blum(balanced_tree)
  blum(unbalanced_tree) # should be higher
}
\references{
M. G. B. Blum and O. Francois (2006). Which random processes
describe the Tree of Life? A large-scale study of phylogenetic tree
imbalance. Systematic Biology. 55:685-691.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/L2phylo.R
\name{ltable_to_newick}
\alias{ltable_to_newick}
\title{Convert an L table to newick string}
\usage{
ltable_to_newick(ltab, drop_extinct = TRUE)
}
\arguments{
\item{ltab}{ltable}

\item{drop_extinct}{should extinct species be dropped from the phylogeny?}
}
\value{
phylo object
}
\description{
Convert an L table to newick string
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean_branch_length.R
\name{mean_branch_length_int}
\alias{mean_branch_length_int}
\title{Mean length of internal branches of a tree}
\usage{
mean_branch_length_int(phy)
}
\arguments{
\item{phy}{phylo object or Ltable}
}
\value{
mean of internal branch lengths
}
\description{
Mean length of internal branches of a tree, e.g. of
branches not leading to a tip.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colless.R
\name{colless}
\alias{colless}
\title{Colless index of (im)balance.}
\usage{
colless(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{A character string equals to "none" (default) for no
normalization or one of "pda" or "yule".}
}
\value{
colless index
}
\description{
The Colless index is calculated as the sum of
\eqn{abs(L - R)} over all nodes, where L (or R) is the number of extant tips
associated with the L (or R) daughter branch at that node.  Higher values
indicate higher imbalance. Two normalizations are available,
where a correction is made for tree size, under either a yule expectation,
or a pda expectation.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
colless(balanced_tree)
colless(unbalanced_tree) # should be higher
}
\references{
Colless D H. 1982. Review of: Phylogenetics: The Theory and
Practice of Phylogenetic Systematics. Systematic Zoology 31:100-104.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rebase_ltable.R
\name{rebase_ltable}
\alias{rebase_ltable}
\title{rebase ltable}
\usage{
rebase_ltable(ltable)
}
\arguments{
\item{ltable}{ltable}
}
\value{
modified ltable
}
\description{
a function to modify an ltable,
such that the longest path in the phylogeny is a crown lineage.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ILnumber.R
\name{ILnumber}
\alias{ILnumber}
\title{ILnumber}
\usage{
ILnumber(input_obj, normalization = "none")
}
\arguments{
\item{input_obj}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the result is normalized
by dividing by N - 2, where N is the number of tips.}
}
\value{
ILnumber
}
\description{
The ILnumber is the number of internal nodes with a
single tip child. Higher values typically indicate a tree that
is more unbalanced.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inv_branch_dist.R
\name{inv_branch_dist}
\alias{inv_branch_dist}
\title{Sum of Inverse branch length}
\usage{
inv_branch_dist(phy, add_one = FALSE)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{add_one}{should we take 1 / bl or 1 / (1 + bl) ? if TRUE, then
1 / (1 + bl) is used. Default: FALSE.}
}
\description{
This statistic calculates per tip, the sum of the inverse of
all branch lengths of the shortest path between the root and the tip.
}
\references{
Williams, P. H., Alonso-Alonso, P., Arbetman, M., Françoso,
E., Ghisbain, G., Huang, J., Orr, M. C., Ren, Z.-X., Streinzer, M., T
hanoosing, C., Vandame, R., Waite, M., & Brace, S. (2026). Evolutionary
Tree for All Bumblebee Species World-Wide Estimated by Combining
Information from Fast-Evolving Genes, Slow-Evolving Genes, and Genomic
Data (Apidae, Bombus). Insects, 17(6), 540.
https://doi.org/10.3390/insects17060540
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean_branch_length.R
\name{mean_branch_length}
\alias{mean_branch_length}
\title{Mean branch length of a tree}
\usage{
mean_branch_length(phy)
}
\arguments{
\item{phy}{phylo object or Ltable}
}
\value{
mean branch length
}
\description{
Mean branch length of a tree, including extinct branches.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mntd.R
\name{mntd}
\alias{mntd}
\title{Mean Nearest Taxon distance}
\usage{
mntd(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
Mean Nearest Taxon Distance.
}
\description{
Per tip, evaluates the shortest distance to another tip,
then takes the average across all tips.
}
\references{
Webb, C., D. Ackerly, M. McPeek, and M. Donoghue. 2002.
Phylogenies and community ecology. Annual Review of Ecology and
Systematics 33:475-505.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branching_times.R
\name{branching_times}
\alias{branching_times}
\title{Branching times of a tree}
\usage{
branching_times(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
vector of branching times
}
\description{
C++ based alternative to
\link[ape:branching.times]{ape::branching.times}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/average_leaf_depth.R
\name{average_leaf_depth}
\alias{average_leaf_depth}
\title{Average leaf depth statistic.}
\usage{
average_leaf_depth(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "yule", in which case the statistic is
divided by the expectation under the yule model, following Remark 1 in
Coronado et al. 2020.}
}
\value{
average leaf depth statistic
}
\description{
The average leaf depth statistic
is a normalized version of the Sackin index, normalized by the number of
tips.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
average_leaf_depth(simulated_tree)
}
\references{
M. Coronado, T., Mir, A., Rosselló, F. et al. On Sackin’s
original proposal: the variance of the leaves’ depths as a phylogenetic
balance index. BMC Bioinformatics 21, 154 (2020).
https://doi.org/10.1186/s12859-020-3405-1
K.-T. Shao and R. R. Sokal. Tree balance. Systematic Zoology,
39(3):266, 1990. doi: 10.2307/2992186.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blum.R
\name{sshape}
\alias{sshape}
\title{s shape statistic of (im)balance.}
\usage{
sshape(phy, normalization = FALSE)
}
\arguments{
\item{phy}{phylogeny or ltable}

\item{normalization}{because the sshape statistic sums over all nodes,
the resulting statistic tends to be correlated with the number of extant
tips. Normalization can be performed by dividing by the number of extant
tips.}
}
\value{
s shape statistic of imbalance
}
\description{
The s shape statistic of imbalance (also known as the Blum
statistic, see \link{blum}) calculates the sum of \eqn{log(N-1)} over all
internal nodes, where N represents the total number of extant tips connected
to that node. An alternative implementation can be found in the Castor R
package.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
  balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
  unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
  sshape(balanced_tree)
  sshape(unbalanced_tree) # should be higher
}
\references{
M. G. B. Blum and O. Francois (2006). Which random processes
describe the Tree of Life? A large-scale study of phylogenetic tree
imbalance. Systematic Biology. 55:685-691.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_bal_unbal_tree.R
\name{create_fully_balanced_tree}
\alias{create_fully_balanced_tree}
\title{Create a fully balanced tree}
\usage{
create_fully_balanced_tree(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
phylo phylo object
}
\description{
This function takes an input phylogeny, and returns a phylogeny
that is most ideally balanced tree, whilst having the same branching times as
the original input tree. Please note that if the number of tips is not even
or not a power of two, the tree may not have perfect balance, but the most
ideal balance possible.
}
\examples{
phy <- ape::rphylo(n = 16, birth = 1, death = 0)
bal_tree <- treestats::create_fully_balanced_tree(phy)
treestats::colless(phy)
treestats::colless(bal_tree) # much lower
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psv.R
\name{psv}
\alias{psv}
\title{Phylogenetic Species Variability.}
\usage{
psv(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the obtained mean
pairwise distance is normalized by the factor 2log(n), where n is the number
of tips.}
}
\value{
Phylogenetic Species Variability
}
\description{
The phylogenetic species variability is bounded in [0, 1]. The
psv quantifies how phylogenetic relatedness decrease the variance of a
(neutral) trait shared by all species in the tree. As species become more
related, the psv tends to 0.
Please note that the psv is a special case of the Mean Pair Distance (see
appendix of Tucker et al. 2017 for a full derivation), and thus correlates
directly.
}
\references{
Helmus M.R., Bland T.J., Williams C.K. & Ives A.R. (2007)
Phylogenetic measures of biodiversity. American Naturalist, 169, E68-E83

Tucker, Caroline M., et al. "A guide to phylogenetic metrics for
conservation, community ecology and macroecology."
Biological Reviews 92.2 (2017): 698-715.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/avg_vertex_depth.R
\name{avg_vert_depth}
\alias{avg_vert_depth}
\title{Average vertex depth metric}
\usage{
avg_vert_depth(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
Average depth (in number of edges)
}
\description{
The average vertex depth metric, measures the average path
(in edges), between the tips and the root.
}
\references{
C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
transmission patterns. Evolution, Medicine, and Public Health,
2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pitchforks.R
\name{pitchforks}
\alias{pitchforks}
\title{Number of pitchforks}
\usage{
pitchforks(input_obj, normalization = "none")
}
\arguments{
\item{input_obj}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the found number of
pitchforks is divided by the expected number.}
}
\value{
number of pitchforks
}
\description{
Pitchforks are a clade with three tips, as introduced in the phyloTop
package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_bal_unbal_tree.R
\name{create_fully_unbalanced_tree}
\alias{create_fully_unbalanced_tree}
\title{Create an unbalanced tree (caterpillar tree)}
\usage{
create_fully_unbalanced_tree(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
phylo phylo object
}
\description{
This function takes an input phylogeny, and returns a phylogeny
that is a perfectly imbalanced tree (e.g. a full caterpillar tree), that
has the same branching times as the original input tree.
}
\examples{
phy <- ape::rphylo(n = 16, birth = 1, death = 0)
bal_tree <- treestats::create_fully_unbalanced_tree(phy)
treestats::colless(phy)
treestats::colless(bal_tree) # much higher
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colless.R
\name{ew_colless}
\alias{ew_colless}
\title{Equal weights Colless index of (im)balance.}
\usage{
ew_colless(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
colless index
}
\description{
The equal weights Colless index is calculated as the sum of
\eqn{abs(L - R) / (L + R - 2)} over all nodes where \eqn{L + R > 2},
where L (or R) is the number of extant tips associated with the L (or R)
daughter branch at that node.  Maximal imbalance is associated with a value
of 1.0. The ew_colless index is not sensitive to tree size.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
ew_colless(balanced_tree)
ew_colless(unbalanced_tree) # should be higher
}
\references{
A. O. Mooers and S. B. Heard. Inferring Evolutionary Process
from Phylogenetic Tree Shape. The Quarterly Review of Biology, 72(1), 1997.
doi: 10.1086/419657.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/b2.R
\name{b2}
\alias{b2}
\title{B2 metric}
\usage{
b2(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "yule", when "yule" is chosen, the statistic
is divided by the Yule expectation, following from theorem 3.7 in Bienvenu
2020.}
}
\value{
b2 statistic
}
\description{
Balance metric that uses the Shannon-Wiener statistic of
information content. The b2 measure is given by the sum over the depths of
all tips, divided by \eqn{2^{d}}, e.g. \eqn{\sum d_i / 2^{d_i}}
}
\references{
K.-T. Shao and R. R. Sokal. Tree Balance.
Systematic Zoology, 39(3):266, 1990. doi: 10.2307/2992186.

Bienvenu, François, Gabriel Cardona, and Celine Scornavacca.
"Revisiting Shao and Sokal’s $$ B_2 $$ B 2 index of phylogenetic balance."
Journal of Mathematical Biology 83.5 (2021): 1-43.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_topology_stats.R
\name{calc_topology_stats}
\alias{calc_topology_stats}
\title{Calculate all topology based statistics for a single tree}
\usage{
calc_topology_stats(phylo, normalize = FALSE)
}
\arguments{
\item{phylo}{phylo object}

\item{normalize}{if set to TRUE, results are normalized (if possible) under
either the  Yule expectation (if available), or the number of tips}
}
\value{
list with statistics
}
\description{
this function calculates all tree statistics based on topology
available in this package for a single tree, being:
\itemize{
  \item{area_per_pair}
  \item{average_leaf_depth}
  \item{avg_ladder}
  \item{avg_vert_depth}
  \item{b1}
  \item{b2}
  \item{beta}
  \item{blum}
  \item{cherries}
  \item{colless}
  \item{colless_corr}
  \item{colless_quad}
  \item{diameter}
  \item{double_cherries}
  \item{eigen_centrality}
  \item{ew_colless}
  \item{four_prong}
  \item{i_stat}
  \item{il_number}
  \item{imbalance_steps}
  \item{j_one}
  \item{max_betweenness}
  \item{max_closeness}
  \item{max_del_width}
  \item{max_depth}
  \item{max_ladder}
  \item{max_width}
  \item{mw_over_md}
  \item{pitchforks}
  \item{rogers}
  \item{root_imbalance}
  \item{rquartet}
  \item{sackin}
  \item{stairs}
  \item{stairs2}
  \item{symmetry_nodes}
  \item{tot_coph}
  \item{tot_internal_path}
  \item{tot_path_length}
  \item{var_depth}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_statistics.R
\name{list_statistics}
\alias{list_statistics}
\title{Provides a list of all available statistics in the package}
\usage{
list_statistics(only_balance_stats = FALSE)
}
\arguments{
\item{only_balance_stats}{only return those statistics associated with
measuring balance of a tree}
}
\value{
vector with names of summary statistics
}
\description{
Provides a list of all available statistics in the package
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tot_internal_path.R
\name{tot_internal_path}
\alias{tot_internal_path}
\title{Total internal path length}
\usage{
tot_internal_path(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
Total internal path length
}
\description{
The total internal path length describes the sums of the depths
of all inner vertices of the tree.
}
\references{
Knuth, Donald E. The Art of Computer Programming:
Fundamental Algorithms, volume 1. Addison-Wesley Professional, 1997.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var_pair_dist.R
\name{var_pair_dist}
\alias{var_pair_dist}
\title{Variance of all pairwise distances.}
\usage{
var_pair_dist(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
Variance in pairwise distance
}
\description{
After calculating all pairwise distances between all tips,
this function takes the variance across these values.
}
\references{
Webb, C., D. Ackerly, M. McPeek, and M. Donoghue. 2002.
Phylogenies and community ecology. Annual Review of Ecology and
Systematics 33:475-505.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tot_path.R
\name{tot_path_length}
\alias{tot_path_length}
\title{Total path length}
\usage{
tot_path_length(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
Total path length
}
\description{
The total path length describes the sums of the depths
of all vertices of the tree. This can also be calculated as the sum of the
total internal path and the sackin index.
}
\references{
C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
transmission patterns. Evolution, Medicine, and Public Health,
2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colless.R
\name{colless_quad}
\alias{colless_quad}
\title{Quadratic Colless index of (im)balance.}
\usage{
colless_quad(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{A character string equals to "none" (default) for no
normalization or "yule"}
}
\value{
quadratic colless index
}
\description{
The Quadratic Colless index is calculated as the sum of
\eqn{(L - R)^2} over all nodes.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
colless_quad(balanced_tree)
colless_quad(unbalanced_tree) # should be higher
}
\references{
Bartoszek, Krzysztof, et al. "Squaring within the Colless index
yields a better balance index." Mathematical Biosciences 331 (2021): 108503.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/root_imbalance.R
\name{root_imbalance}
\alias{root_imbalance}
\title{Root imbalance}
\usage{
root_imbalance(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
Root imbalance
}
\description{
Measures the distribution of tips over the two crown lineages,
e.g. \eqn{n1 / (n1 + n2)}, where n1 is the number of tips connected to crown
lineage 1 and n2 is the number of tips connected to crown lineage 2. We
always take n1 > n2, thus root imbalance is always in [0.5, 1]. Root
imbalance is also calculated for polytomous trees, as long as there is no
polytomy at the root.
}
\references{
Guyer, Craig, and Joseph B. Slowinski. "Adaptive radiation and
the topology of large phylogenies." Evolution 47.1 (1993): 253-263.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_unbalanced_tree.R
\name{make_unbalanced_tree}
\alias{make_unbalanced_tree}
\title{Stepwise increase the imbalance of a tree}
\usage{
make_unbalanced_tree(
  init_tree,
  unbal_steps,
  group_method = "any",
  selection_method = "random"
)
}
\arguments{
\item{init_tree}{starting tree to work with}

\item{unbal_steps}{number of imbalance generating steps}

\item{group_method}{choice of "any" and "terminal"}

\item{selection_method}{choice of "random", "youngest" and "oldest"}
}
\value{
phylo object
}
\description{
the goal of this function is to increasingly imbalance a tree,
by changing the topology, one move at a time. It does so by re-attaching
terminal branches to the root lineage, through the ltable. In effect, this
causes the tree to become increasingly caterpillarlike. When started with
a balanced tree, this allows for exploring the gradient between a fully
balanced tree, and a fully unbalanced tree.
Please note that the algorithm will try to increase imbalance, until a fully
caterpillar like tree is reached, which may occur before unbal_steps is
reached.
Three methods are available: "youngest", reattaches branches in order of age,
starting with the branch originating from the most recent branching event
and working itself through the tree. "Random" picks a random branch to
reattach. "Terminal" also picks a random branch, but only from terminal
branches (e.g. branches that don't have any daughter lineages, which is
maximized in a fully imbalanced tree).
}
\examples{
simulated_tree <- ape::rphylo(n = 16, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
intermediate_tree <- make_unbalanced_tree(balanced_tree, 8)
colless(balanced_tree)
colless(intermediate_tree) # should be intermediate value
colless(unbalanced_tree) # should be highest colless value
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eigen_centrality.R
\name{eigen_centrality}
\alias{eigen_centrality}
\title{Eigen vector centrality}
\usage{
eigen_centrality(phy, weight = TRUE, scale = FALSE, use_rspectra = TRUE)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{weight}{if TRUE, uses branch lengths.}

\item{scale}{if TRUE, the Eigenvector is rescaled}

\item{use_rspectra}{boolean to indicate whether the helping package RSpectra
should be used, which is faster, but returns fewer eigen values.}
}
\value{
List with the Eigen vector and the leading Eigen value
}
\description{
Eigen vector centrality associates with each node \eqn{v}
the positive value \eqn{e(v)}, such that:
\eqn{ \sum_{e}^v w(uv) * e(u) = \lambda * e(v) }. Thus,
\eqn{e(v)} is the Perron-Frobenius eigenvector of the adjacency matrix of the
tree.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." Plos one 16.12 (2021): e0259877.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crown_age.R
\name{tree_height}
\alias{tree_height}
\title{Height of a tree.}
\usage{
tree_height(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
crown age
}
\description{
In a reconstructed tree, obtaining the tree height is fairly
straightforward, and the function
\link[beautier:get_crown_age]{beautier::get_crown_age} does a great
job at it. However, in a non-ultrametric tree, that function no longer works.
Alternatively, taking the maximum value of
\link[adephylo:distRoot]{adephylo::distRoot} will
also yield the tree height (including the root branch), but will typically
perform many superfluous calculations and thus be slow.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeness.R
\name{treeness}
\alias{treeness}
\title{Treeness statistic}
\usage{
treeness(phy)
}
\arguments{
\item{phy}{phylo object or Ltable}
}
\value{
sum of all internal branch lengths (e.g. branches not leading
to a tip) divided by the sum over all branch lengths.
}
\description{
Calculates the fraction of tree length on internal branches,
also known as treeness or stemminess
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean_branch_length.R
\name{var_branch_length}
\alias{var_branch_length}
\title{Variance of branch lengths of a tree}
\usage{
var_branch_length(phy)
}
\arguments{
\item{phy}{phylo object or Ltable}
}
\value{
variance of branch lengths
}
\description{
Variance of branch lengths of a tree,
including extinct branches.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_ladder.R
\name{max_ladder}
\alias{max_ladder}
\title{Maximum ladder index}
\usage{
max_ladder(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
longest ladder in the tree
}
\description{
Calculate the maximum ladder index, from the phyloTop package.
Higher values indicate more unbalanced trees. To calculate the maximum ladder
index, first all potential ladders in the tree are calculated. A ladder is
defined as a sequence of nodes where one of the daughter branches is a
terminal branch, resulting in a 'ladder' like pattern. The maximum ladder
index then represents the longest ladder found among all observed ladders
in the tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nLTT.R
\name{nLTT}
\alias{nLTT}
\title{Normalized LTT statistic}
\usage{
nLTT(phy, ref_tree)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{ref_tree}{reference tree to compare with (should be same type as phy)}
}
\value{
number of lineages
}
\description{
The nLTT statistic calculates the sum of
absolute differences in the number of lineages over time, where both the
number of lineages and the time are normalized. The number of lineages is
normalized by the number of extant tips, whereas the time is normalized by
the crown age. The nLTT can only be calculated for reconstructed trees.
Only use the treestats version if you are very certain
about the input data, and are certain that performing nLTT is valid (e.g.
your tree is ultrametric etc). If you are less certain, use the nLTT function
from the nLTT package.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
reference_tree <- ape::rphylo(n = 10, birth = 0.2, death = 0)
nLTT(simulated_tree, reference_tree)
nLTT(simulated_tree, simulated_tree) # should be zero.
}
\references{
Janzen, T., Höhna, S. and Etienne, R.S. (2015), Approximate
Bayesian Computation of diversification rates from molecular phylogenies:
introducing a new efficient summary statistic, the nLTT. Methods Ecol Evol,
6: 566-575. https://doi.org/10.1111/2041-210X.12350
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/j_one_stat.R
\name{j_one}
\alias{j_one}
\title{\eqn{J^1} index.}
\usage{
j_one(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
\eqn{j^1} index
}
\description{
The \eqn{J^1} index calculates the Shannon Entropy of a tree,
where at each node with two children, the Shannon Entropy is the sum of
\eqn{p_i log_2(p_i)} over the two children \eqn{i}, and \eqn{p_i} is
\eqn{L / (L + R)},
where L and R represent the number of tips connected to the two daughter
branches.
}
\references{
Jeanne Lemant, Cécile Le Sueur, Veselin Manojlović, Robert Noble,
Robust, Universal Tree Balance Indices, Systematic Biology, Volume 71,
Issue 5, September 2022, Pages 1210–1224

https://doi.org/10.1093/sysbio/syac027
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nLTT.R
\name{nLTT_base}
\alias{nLTT_base}
\title{Reference nLTT statistic}
\usage{
nLTT_base(phy)
}
\arguments{
\item{phy}{phylo object}
}
\value{
number of lineages
}
\description{
The base nLTT statistic can be used as a semi stand-alone
statistic for phylogenetic trees. However, please note that although this
provides a nice way of checking the power of the nLTT statistic without
directly comparing two trees, the nLTT_base statistic is not a substitute
for directly comparing two phylogenetic trees. E.g. one would perhaps
naively assume that \eqn{nLTT(A, B) = |nLTT(A, base) - nLTT(B, base)}.
Indeed, in some cases this may hold true (when, for instance, all normalized
lineages of A are less than all normalized lineages of B), but once the
nLTT curve of A intersects the nLTT curve of B, this no longer applies.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
nLTT_base(simulated_tree)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw_over_md.R
\name{mw_over_md}
\alias{mw_over_md}
\title{Maximum width of branch depths divided by the maximum depth}
\usage{
mw_over_md(phy)
}
\arguments{
\item{phy}{phylogeny or ltable}
}
\description{
Calculates the maximum width divided by the maximum depth.
}
\references{
C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
transmission patterns. Evolution, Medicine, and Public Health,
2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rogers.R
\name{rogers}
\alias{rogers}
\title{Rogers J index of (im)balance.}
\usage{
rogers(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the resulting statistic
is divided by the number of tips - 2 (e.g. the maximum value of the rogers
index for a tree).}
}
\value{
Rogers index
}
\description{
The Rogers index is calculated as the total number of internal
nodes that are unbalanced, e.g. for which both daughter nodes lead to a
different number of extant tips. in other words, the number of nodes where
\eqn{L  !=  R} (where L(R) is the number of extant tips of the Left (Right)
daughter node).
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
rogers(balanced_tree)
rogers(unbalanced_tree) # should be higher
}
\references{
J. S. Rogers. Central Moments and Probability Distributions of
Three Measures of Phylogenetic Tree Imbalance. Systematic Biology,
45(1):99-110, 1996. doi: 10.1093/sysbio/45.1.99.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cherries.R
\name{cherries}
\alias{cherries}
\title{Cherry index}
\usage{
cherries(input_obj, normalization = "none")
}
\arguments{
\item{input_obj}{phylo object or ltable}

\item{normalization}{"none", "yule", or "pda", the found number of
cherries is divided by the expected number, following
McKenzie & Steel 2000.}
}
\value{
number of cherries
}
\description{
Calculate the number of cherries, from the phyloTop package.
A cherry is a pair of sister tips.
}
\references{
McKenzie, Andy, and Mike Steel. "Distributions of cherries for
two models of trees." Mathematical biosciences 164.1 (2000): 81-92.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamma.R
\name{gamma_statistic}
\alias{gamma_statistic}
\title{Gamma statistic}
\usage{
gamma_statistic(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
gamma statistic
}
\description{
The gamma statistic measures the relative position of
internal nodes within a reconstructed phylogeny. Under the Yule process,
the gamma values of a reconstructed tree follow a standard normal
distribution. If gamma > 0, the nodes are located more towards the tips of
the tree, and if gamma < 0, the nodes are located more towards the root of
the tree. Only available for ultrametric trees.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
gamma_statistic(simulated_tree) # should be around 0.
if (requireNamespace("DDD")) {
  ddd_tree <- DDD::dd_sim(pars = c(1, 0, 10), age = 7)$tes
  gamma_statistic(ddd_tree) # because of diversity dependence, should be < 0
}
}
\references{
Pybus, O. G. and Harvey, P. H. (2000) Testing macro-evolutionary
models using incomplete molecular phylogenies. Proceedings of the Royal
Society of London. Series B. Biological Sciences, 267, 2267–2272.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_del_width.R
\name{max_del_width}
\alias{max_del_width}
\title{Maximum difference of widths of a phylogenetic tree}
\usage{
max_del_width(phy, normalization = "none")
}
\arguments{
\item{phy}{phylogeny or ltable}

\item{normalization}{"none" or "tips", in which case the resulting statistic
is divided by the number of tips in the tree.}
}
\value{
maximum difference of widths
}
\description{
Calculates the maximum difference of widths of a phylogenetic
tree. First, the widths are calculated by collecting the depth of each node
and tip across the entire tree, where the depth represents the distance
(in nodes) to the root. Then, the width represents the number of occurrences
of each possible depth. Then, we take the difference between each consecutive
width, starting with the first width. The maximum difference is then
returned - whereas the original statistic designed by Colijn and Gardy used
the absolute maximum difference, we here use the modified version as
introduced in Fischer 2023: this returns the maximum value, without
absoluting negative widths. This ensures that this metric is a proper
(im)balance metric, following Fischer 2023.
}
\references{
C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
transmission patterns. Evolution, Medicine, and Public Health,
2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.

Fischer, M., Herbst, L., Kersting, S., Kühn, A. L., & Wicke, K. (2023).
Tree Balance Indices: A Comprehensive Survey.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mean_branch_length.R
\name{mean_branch_length_ext}
\alias{mean_branch_length_ext}
\title{Mean length of external branch lengths of a tree}
\usage{
mean_branch_length_ext(phy)
}
\arguments{
\item{phy}{phylo object or Ltable}
}
\value{
mean of external branch lengths
}
\description{
Mean length of external branch lengths of a tree, e.g. of
branches leading to a tip.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diameter.R
\name{diameter}
\alias{diameter}
\title{Diameter statistic}
\usage{
diameter(phy, weight = FALSE)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{weight}{if TRUE, uses branch lengths.}
}
\value{
Diameter
}
\description{
The Diameter of a tree is defined as the maximum length of a
shortest path. When taking branch lengths into account, this is equal to
twice the crown age.
When the tree is unrooted, we add 1 to the unweighted diameter, to reflect
traversing the (virtual) root.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." PloS one 16.12 (2021): e0259877.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rpanda.R
\name{laplacian_spectrum}
\alias{laplacian_spectrum}
\title{Laplacian spectrum statistics, from RPANDA}
\usage{
laplacian_spectrum(phy)
}
\arguments{
\item{phy}{phy}
}
\value{
list with five components: 1) eigenvalues the vector of eigen
values, 2) principal_eigenvalue the largest eigenvalue of the spectral
density distribution 3) asymmetry the skewness of the spectral density
distribution 4) peak_height the largest y-axis value of the spectral
density distribution and 5) eigengap the position of the largest
difference between eigenvalues, giving the number of modalities in the tree.
}
\description{
Computes the distribution of eigenvalues for the modified graph
Laplacian of a phylogenetic tree, and several summary statistics of this
distribution. The modified graph Laplacian of a phylogeny is given by the
difference between its' distance matrix (e.g. all pairwise distances between
all nodes), and the degree matrix (e.g. the diagonal matrix where each
diagonal element represents the sum of branch lengths to all other nodes).
Each row of the modified graph Laplacian sums to zero. For a tree with n
tips, there are at most \eqn{N = 2n-1} nodes, and hence the modified graph
Laplacian is represented by a \eqn{N x N} matrix. Where RPANDA relies on the
package igraph to calculate the modified graph Laplacian, the treestats
package uses C++ to directly calculate the different entries in the matrix.
This makes the treestats implementation slightly faster, although the bulk of
computation occurs in estimating the eigen values, where RcppArmadillo
generates a bit of speed up.
}
\references{
Eric Lewitus, Helene Morlon, Characterizing and Comparing
Phylogenies from their Laplacian Spectrum, Systematic Biology, Volume 65,
Issue 3, May 2016, Pages 495–507, https://doi.org/10.1093/sysbio/syv116
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sym_nodes.R
\name{sym_nodes}
\alias{sym_nodes}
\title{Symmetry nodes metric}
\usage{
sym_nodes(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the resulting statistic
is divided by the number of tips - 2 (e.g. the maximum value of the symmetry
nodes index for a tree).}
}
\value{
Maximum depth (in number of edges)
}
\description{
Balance metric that returns the total number of internal nodes
that are not-symmetric (confusingly enough). A node is considered symmetric
when both daughter trees have the same topology, measured as having the
same sum of depths, where depth is measured as the distance from the root
to the node/tip.
}
\references{
S. J. Kersting and M. Fischer. Measuring tree balance using
symmetry nodes — A new balance index and its extremal properties.
Mathematical Biosciences, page 108690, 2021. ISSN 0025-5564.
doi:https://doi.org/10.1016/j.mbs.2021.108690
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wiener.R
\name{wiener}
\alias{wiener}
\title{Wiener index}
\usage{
wiener(phy, normalization = FALSE, weight = TRUE)
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{if TRUE, the Wiener index is normalized by the number of
nodes, e.g. by choose(n, 2), where n is the number of nodes.}

\item{weight}{if TRUE, branch lenghts are used.}
}
\value{
Wiener index
}
\description{
The Wiener index is defined as the sum of all shortest path
lengths between pairs of nodes in a tree.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." Plos one 16.12 (2021): e0259877.
Mohar, B., Pisanski, T. How to compute the Wiener index of a graph.
J Math Chem 2, 267–277 (1988)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_width.R
\name{max_width}
\alias{max_width}
\title{Maximum width of branch depths.}
\usage{
max_width(phy, normalization = "none")
}
\arguments{
\item{phy}{phylogeny or ltable}

\item{normalization}{"none" or "tips", in which case the resulting statistic
is divided by the number of tips in the tree.}
}
\value{
maximum width
}
\description{
Calculates the maximum width, this is calculated by first
collecting the depth of each node and tip across the entire tree, where the
depth represents the distance (in nodes) to the root. Then, the width
represents the number of occurrences of each possible depth. The maximal
width then returns the maximum number of such occurences.
}
\references{
C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
transmission patterns. Evolution, Medicine, and Public Health,
2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stairs.R
\name{stairs}
\alias{stairs}
\title{Stairs index}
\usage{
stairs(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
number of stairs
}
\description{
Calculates the staircase-ness measure, from the phyloTop
package. The staircase-ness reflects the number of subtrees that are
imbalanced, e.g. subtrees where the left child has more extant tips than the
right child, or vice versa.
}
\references{
Norström, Melissa M., et al. "Phylotempo: a set of r scripts for
assessing and visualizing temporal clustering in genealogies inferred from
serially sampled viral sequences." Evolutionary Bioinformatics 8 (2012):
EBO-S9738.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_brts_stats.R
\name{calc_brts_stats}
\alias{calc_brts_stats}
\title{Apply all tree statistics related to branching times to a single tree.}
\usage{
calc_brts_stats(phylo)
}
\arguments{
\item{phylo}{phylo object}
}
\value{
list with statistics
}
\description{
this function applies all tree statistics based on
branching times to a single tree (more or less ignoring topology), being:
\itemize{
  \item{gamma}
  \item{pigot's rho}
  \item{mean branch length}
  \item{nLTT with empty tree}
  \item{treeness}
  \item{var branch length}
  \item{mean internal branch length}
  \item{mean external branch length}
  \item{var internal branch length}
  \item{var external branch length}
  \item{mean of the sum of inverse branch lengths}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/J_stat.R
\name{entropy_j}
\alias{entropy_j}
\title{Intensive quadratic entropy statistic J.}
\usage{
entropy_j(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
intensive quadratic entropy statistic J
}
\description{
The intensive quadratic entropy statistic J is given by the
average distance between two randomly chosen species, thus given by the
sum of all pairwise distances, divided by \eqn{S^2}, where S is the number
of tips of the tree.
}
\references{
Izsák, János, and Laszlo Papp. "A link between ecological
diversity indices and measures of biodiversity."
Ecological Modelling 130.1-3 (2000): 151-156.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/L2phylo.R
\name{l_to_phylo}
\alias{l_to_phylo}
\title{Convert an L table to phylo object}
\usage{
l_to_phylo(ltab, drop_extinct = TRUE)
}
\arguments{
\item{ltab}{ltable}

\item{drop_extinct}{should extinct species be dropped from the phylogeny?}
}
\value{
phylo object
}
\description{
Convert an L table to phylo object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/four_prong.R
\name{four_prong}
\alias{four_prong}
\title{Four prong index}
\usage{
four_prong(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
number of 4-tip caterpillars
}
\description{
Calculate the number of 4-tip caterpillars. The function does
accept non-binary trees, but four-prongs are only calculated on
non-polytomous sub-trees with 4 tips, e.g. the 4-tip part has to be a
caterpillar.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." PloS one 16.12 (2021): e0259877.

Rosenberg, Noah A. "The mean and variance of the numbers of r-pronged nodes
and r-caterpillars in Yule-generated genealogical trees."
Annals of Combinatorics 10 (2006): 129-146.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crown_age.R
\name{crown_age}
\alias{crown_age}
\title{Crown age of a tree.}
\usage{
crown_age(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
crown age
}
\description{
In a reconstructed tree, obtaining the crown age is
fairly straightforward, and the function
\link[beautier:get_crown_age]{beautier::get_crown_age}  does
a great job at it. However, in a non-ultrametric tree, that function no
longer works. This function provides a functioning alternative.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/double_cherries.R
\name{double_cherries}
\alias{double_cherries}
\title{Double Cherry index}
\usage{
double_cherries(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
number of cherries
}
\description{
Calculate the number of double cherries, where a single cherry
is a node connected to two tips, and a double cherry is a node connected
to two cherry nodes.
}
\references{
Chindelevitch, Leonid, et al. "Network science inspires novel
tree shape statistics." PloS one 16.12 (2021): e0259877.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/avgLadder.R
\name{avg_ladder}
\alias{avg_ladder}
\title{Average ladder index}
\usage{
avg_ladder(input_obj)
}
\arguments{
\item{input_obj}{phylo object or ltable}
}
\value{
average number of ladders
}
\description{
Calculate the avgLadder index, from the phyloTop package.
Higher values indicate more unbalanced trees. To calculate the average ladder
index, first all potential ladders in the tree are calculated. A ladder is
defined as a sequence of nodes where one of the daughter branches is a
terminal branch, resulting in a 'ladder' like pattern. The average ladder
index then represents the average length across all observed ladders in the
tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Ibased.R
\name{mean_i}
\alias{mean_i}
\title{Mean I statistic.}
\usage{
mean_i(phy)
}
\arguments{
\item{phy}{phylo object or ltable}
}
\value{
average I value across all nodes
}
\description{
The mean I value is defined for all nodes with at least 4 tips
connected, such that different topologies can be formed. Then, for each node,
\eqn{I = (nm - nt/2) / (nt - 1 - nt/2)}, where nt is the total number of tips
descending from that node, nm is the daughter branch leading to most tips,
and \eqn{nt/2} is the minimum size of the maximum branch, rounded up.
Following Purvis et al 2002, we perform a correction on I, where we correct
I for odd nt, such that \eqn{I' = I * (nt - 1) / nt}. This correction ensures
that I is independent of nt. We report the mean value across all I' (again,
following Purvis et al. 2002).
}
\references{
G. Fusco and Q. C. Cronk. A new method for evaluating the shape
of large phylogenies. Journal of Theoretical Biology, 1995.
doi: 10.1006/jtbi.1995.0136.

A. Purvis, A. Katzourakis, and P.-M. Agapow. Evaluating Phylogenetic Tree
Shape: Two Modifications to Fusco & Cronks Method. Journal of Theoretical
Biology, 2002. doi: 10.1006/jtbi.2001.2443.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var_leaf_depth.R
\name{var_leaf_depth}
\alias{var_leaf_depth}
\title{Variance of leaf depth statistic}
\usage{
var_leaf_depth(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "yule", when "yule" is chosen, the statistic
is divided by the Yule expectation}
}
\value{
Variance of leaf depths
}
\description{
The variance of leaf depth statistic returns the variance
of depths across all tips, where depth of a tip indicates the distance of
the tip to the root.
}
\references{
T. M. Coronado, A. Mir, F. Rosselló, and L. Rotger.
On Sackin's original proposal: the variance of the leaves' depths as a
phylogenetic balance index. BMC Bioinformatics, 21(1), 2020.
doi: 10.1186/s12859-020-3405-1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colless.R
\name{colless_corr}
\alias{colless_corr}
\title{Corrected Colless index of (im)balance.}
\usage{
colless_corr(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{A character string equals to "none" (default) for no
normalization or "yule", in which case the obtained index is divided by
the Yule expectation.}
}
\value{
corrected colless index
}
\description{
The Corrected Colless index is calculated as the sum of
\eqn{abs(L - R)} over all nodes, corrected for tree size by dividing over
\eqn{(n-1) * (n-2)}, where n is the number of nodes.
}
\examples{
simulated_tree <- ape::rphylo(n = 10, birth = 1, death = 0)
balanced_tree <- treestats::create_fully_balanced_tree(simulated_tree)
unbalanced_tree <- treestats::create_fully_unbalanced_tree(simulated_tree)
colless_corr(balanced_tree)
colless_corr(unbalanced_tree) # should be higher
}
\references{
Heard, Stephen B. "Patterns in tree balance among cladistic,
phenetic, and randomly generated phylogenetic trees." Evolution 46.6 (1992):
1818-1826.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_depth.R
\name{max_depth}
\alias{max_depth}
\title{Maximum depth metric}
\usage{
max_depth(phy, normalization = "none")
}
\arguments{
\item{phy}{phylo object or ltable}

\item{normalization}{"none" or "tips", in which case the resulting statistic
is divided by the number of tips in the tree.}
}
\value{
Maximum depth (in number of edges)
}
\description{
The maximum depth metric, measures the maximal path (in edges),
between the tips and the root.
}
\references{
C. Colijn and J. Gardy. Phylogenetic tree shapes resolve disease
transmission patterns. Evolution, Medicine, and Public Health,
2014(1):96-108, 2014. ISSN 2050-6201. doi: 10.1093/emph/eou018.
}
